/* -*- Mode: C; tab-width: 2; indent-tabs-mode: t; c-basic-offset: 2 -*-
 *
 * LASER
 * Copyright (C) 2006 Benjamin Canou <bookeldor@gmail.com>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <gconf/gconf-client.h>
#include <cairo.h>
#include <librsvg/rsvg.h>
#include <librsvg/rsvg-cairo.h>
#include <gtk/gtk.h>
#include <glade/glade.h>
#include <gconf/gconf-client.h>
#include <libgnomeui/libgnomeui.h>
#include <libgnomeui/gnome-help.h>

#include "config.h"
#include "laser-common.h"
#include "laser-config.h"
#include "laser-parser.h"
#include "laser-board.h"

#define MAX_UNDO_STEPS 200

struct undo {
	item**       map;
	gint         materials[MATERIALS];
	gint         width, height, real_height;
};

struct ray {
	guint8       colour_id;
	guint8       hidden;
};

struct board {
	gboolean     loaded, dragged, pressed, *modified;
	LaserConfig* config;
	theme_cache  *cache;
	GtkWidget*   widget;
	board_mode   mode;
	guint16      type;
	guint8       colour;
	item**       map;
  level*       level;
  struct ray** rays[8];
	guint8**     synthetised;
  gboolean     selected;
  gint         select_x, select_y;
	gint         last_drag_x, last_drag_y;
	gint         actions;
	void         (*update_score_cb) (gpointer, gint, gint, gint, gboolean, gboolean);
	gpointer     update_score_data;
	struct undo  undo_stack[MAX_UNDO_STEPS + 1];
	gint         undo_max, undo_pointer;
};

static void save_for_undo (board_t *board) {
	*board->modified = TRUE;

	/* free redo stack */
	if (board->undo_pointer < board->undo_max) {
		for (gint u = board->undo_pointer + 1;u <= board->undo_max;u++) {
			for (gint i = 0;i < board->undo_stack[u].height;i++) {
				g_free (board->undo_stack[u].map[i]);
			}
			g_free (board->undo_stack[u].map);
		}
		board->undo_max = board->undo_pointer;
	}
	
	/* if stack full */
	if (board->undo_max == MAX_UNDO_STEPS) {
		/* free first */
		for (gint i = 0;i < board->undo_stack[0].height;i++) {
			g_free (board->undo_stack[0].map[i]);
		}
		g_free (board->undo_stack[0].map);

		/* shift */
		for (int i = 0; i < MAX_UNDO_STEPS;i++) {
			board->undo_stack[i].map = board->undo_stack[i + 1].map;
			board->undo_stack[i].width = board->undo_stack[i + 1].width;
			board->undo_stack[i].height = board->undo_stack[i + 1].height;
			board->undo_stack[i].real_height = board->undo_stack[i + 1].real_height;
			for (gint j = 0;j < MATERIALS;j++) {
				board->undo_stack[i].materials[j] = board->undo_stack[i + 1].materials[j];
			}
		}
	} else {
		board->undo_max++;
		board->undo_pointer++;
	}

	/* set new undo item */
	board->undo_stack[board->undo_max].width = board->level->width;
	board->undo_stack[board->undo_max].height = board->level->height;
	for (int i = 0;i < MATERIALS;i++) {
		board->undo_stack[board->undo_max].materials[i] = board->level->materials[i];
	}
	board->undo_stack[board->undo_max].map = g_malloc (board->level->height * sizeof (struct item*));
	for (int i = 0;i < board->level->height;i++) {
		board->undo_stack[board->undo_max].map[i] = g_malloc (board->level->width * sizeof (struct item));
		memcpy (board->undo_stack[board->undo_max].map[i],
						board->map[i],
						board->level->width * sizeof (struct item));
	}
}

void update_state (board_t *board) {
  if (board->loaded) {
		/* compute number of targets and of reached ones */
    gint total_targets = 0;
    gint reached_targets = 0;
    for (gint y = 0;y < board->level->height;y++) {
      for (gint x = 0;x < board->level->width;x++) {
				if (board->map[y][x].type == MAP_TARGET) {
					total_targets++;
					if (board->rays[0][y + 1][x + 1 + 1].colour_id        & (1 << board->map[y][x].colour)
							|| board->rays[1][y + 1 + 1][x + 1 + 1].colour_id & (1 << board->map[y][x].colour)
							|| board->rays[2][y + 1 + 1][x + 1].colour_id     & (1 << board->map[y][x].colour)
							|| board->rays[3][y + 1 + 1][x - 1 + 1].colour_id & (1 << board->map[y][x].colour)
							|| board->rays[4][y + 1][x - 1 + 1].colour_id     & (1 << board->map[y][x].colour)
							|| board->rays[5][y - 1 + 1][x - 1 + 1].colour_id & (1 << board->map[y][x].colour)
							|| board->rays[6][y - 1 + 1][x + 1].colour_id     & (1 << board->map[y][x].colour)
							|| board->rays[7][y - 1 + 1][x + 1 + 1].colour_id & (1 << board->map[y][x].colour)) {
						reached_targets++;
					}
				}
      }
    }
		board->update_score_cb (board->update_score_data,
														total_targets, reached_targets, board->actions,
														board->undo_pointer > 1,
														board->undo_pointer < board->undo_max);
	}
}

static void get_sizes (board_t *board, gint *ox, gint *oy, gint *d) {
	*ox = (board->widget->allocation.width % board->level->width) / 2;
	*oy = (board->widget->allocation.height % board->level->height) / 2;
	gint w = (board->widget->allocation.width) - *ox;
	gint h = (board->widget->allocation.height) - *oy;
	gint dx = w / board->level->width;
	gint dy = h / board->level->height;
	if (dx > dy) {
		*d = dy;
		*ox = (w - *d * board->level->width) / 2;
	} else {
		*d = dx;
		*oy = (h - *d * board->level->height) /2;
	}
}

static void draw_sprite (board_t *board, cairo_t *cr, item *item, int x, int y, int s) {
	sprite_cache_set (board->cache, s);
	cairo_surface_t *surface = sprite_cache_get (board->cache, item);
	cairo_save (cr);
  cairo_translate (cr, x, y);
  cairo_set_source_surface (cr, surface, 0, 0);
  cairo_rectangle (cr, 0, 0, s, s);
  cairo_fill (cr);
  cairo_restore (cr);
}

static gboolean paint_cb (GtkWidget *widget, GdkEventExpose *ev, board_t *board) {
	static int nx[8] = {-1,-1, 0, 1, 1, 1, 0,-1};
	static int ny[8] = { 0,-1,-1,-1, 0, 1, 1, 1};
	colour_t colour;
	laser_config_get_colour (board->config, &colour);

	/* check if the widget is well initialised */
  if (widget == NULL
      || widget->window == NULL
      || widget->allocation.width < 10
      || widget->allocation.height < 10) {
    return FALSE;
  }

	/* then draw */
  if (board->loaded) {
		/* if game loaded, draw game */

		/* compute sizes */
		gint ox, oy, d;
		get_sizes (board, &ox, &oy, &d);
		/* create a cairo context */
    cairo_surface_t *buffer =
      cairo_image_surface_create (CAIRO_FORMAT_ARGB32,
																	widget->allocation.width,
																	widget->allocation.height);
    if (buffer == NULL) {
      return FALSE;
    }
    cairo_t *cr = cairo_create(buffer);
    if (cr == NULL) {
      return FALSE;
    }

		/* fill with background colour */
    cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
    cairo_set_source_rgba (cr, colour.r, colour.g, colour.b, colour.a);
    cairo_paint (cr);

		/* draw background slots to materialise the map size */
    cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
    for (int y = 0;y < board->level->height;y++) {
      for (int x = 0;x < board->level->width;x++) {
				int sx = x * d + ox;
				int sy = y * d + oy;
				if (board->map[y][x].type == NONE) {
					draw_sprite (board, cr, &(item){NONE, NONE,0}, sx, sy, d);
				}
				if (board->selected && x == board->select_x && y == board->select_y) {
					GdkColor *color = &gtk_rc_get_style (GTK_WIDGET(widget))->bg[GTK_STATE_SELECTED];
					cairo_set_source_rgba (cr,
																 ((double) color->red) / ((double) 0xFFFF),
																 ((double) color->green) / ((double) 0xFFFF),
																 ((double) color->blue) / ((double) 0xFFFF),
																 1.0);
					cairo_rectangle (cr, sx, sy, d, d);
					cairo_fill (cr);
				}
      }
    }

		/* draw rays */
    for (int y = -1;y < board->level->height + 1;y++) {
      for (int x = -1;x < board->level->width + 1;x++) {
				for (int i = 0;i < 8;i++) {
					/* compute number of rays taking account of surrounding incoming rays */
					int nb_rays = 0;
					for (int c = 1; c <= COLOURS;c++) {
						if ((board->rays[i][y + 1][x + 1].colour_id & (1 << c)
								 && ! (board->rays[i][y + 1][x + 1].hidden & (1 << c)))
								|| (y >= 0 && y < board->level->height
										&& x >= 0 && x < board->level->width
										&& board->rays[(i + 4) % 8][y + 1 + ny[i]][x + 1 + nx[i]].colour_id & (1 << c)
										&& ! (board->rays[(i + 4) % 8][y + 1 + ny[i]][x + 1 + nx[i]].hidden & (1 << c)))) {
							nb_rays++;
						}
					}
					if (nb_rays != 0) {
						/* compute size of a sub ray (recall rays can hold mutiple colours) */
						double norm = sqrt (nx[(2 + i) % 8] * nx[(2 + i) % 8]
																+ ny[(2 + i) % 8] * ny[(2 + i) % 8]);
						double dd = (double) d / (8.0 * (double) nb_rays) / norm;
						int n_ray = 0;
						for (int c = 1; c <= COLOURS;c++) {
							/* draw ray */
							if (board->rays[i][y + 1][x + 1].colour_id & (1 << c)
									&& !(board->rays[i][y + 1][x + 1].hidden & (1 << c))) {
								cairo_set_line_width (cr, (double) d / (8.0 * (double) nb_rays));
								colour_t rcolour;
								laser_config_colour_from_id (c, &rcolour);
								cairo_save (cr);
								cairo_translate (cr, x * d  + d/2 + ox, y * d  + d/2 + oy);
								cairo_rotate (cr, G_PI / 4 * ((gdouble) i + 2));
								cairo_set_source_rgba (cr, rcolour.r, rcolour.g, rcolour.b, rcolour.a);
								if (x == -1 || y == -1 || x == board->level->width || y == board->level->height) {
									/* if infinite, step until the border is reached and draw all */
									cairo_move_to (cr, dd * (double) n_ray - dd * ((double) nb_rays - 1) / 2, 0);
									cairo_rel_line_to (cr, 0, widget->allocation.width + widget->allocation.height);
									cairo_stroke (cr);
								} else {
									/* else draw a single step */
									cairo_move_to (cr, dd * (double) n_ray - dd * ((double) nb_rays - 1) / 2, 0);
									cairo_rel_line_to (cr, 0, d * norm);
									cairo_stroke (cr);
									if (laser_config_get_arrows (board->config)) {
										cairo_move_to (cr, - (double) d / 8, d * norm / 2 - (nb_rays * d / 8) / 2 + n_ray * d / 8);
										cairo_rel_line_to (cr, - (double) d / 16, - d / 8);
										cairo_rel_line_to (cr, 2 * (double) d / 16, 0);
										cairo_fill (cr);
									}
								}
								n_ray++;
								cairo_restore (cr);
							}
						}
					}
				}
      }
    }

		/* draw objects */
    for (int y = 0;y < board->level->height;y++) {
      for (int x = 0;x < board->level->width;x++) {
				if (board->map[y][x].type != NONE) {
					draw_sprite (board, cr, &board->map[y][x], x * d + ox, y * d + oy, d);
				}
      }
    }

    cairo_destroy (cr);

		/* blit to widget */
    cr = gdk_cairo_create(widget->window);
    if (cr == NULL) {
      return FALSE;
    }
    cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
    cairo_set_source_surface (cr, buffer, 0, 0);
    cairo_paint (cr);
    cairo_destroy (cr);
    cairo_surface_destroy (buffer);

    return TRUE;
  } else {
		/* else just fill with background colour */
    cairo_t * cr = gdk_cairo_create(widget->window);
    if (cr == NULL) {
      return FALSE;
    }
    cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
    cairo_set_source_rgba (cr, colour.r, colour.g, colour.b, colour.a);
    cairo_paint (cr);
    cairo_destroy (cr);

    return TRUE;
  }
}

/**
 * colour_sum:
 * @colour_1: First colour to sum
 * @colour_2: Second colour to sum
 *
 * Returns the RGB sum of two colours
 */
static guint8 colour_sum (guint8 colour_1, guint8 colour_2) {
	guint8 result;
	switch (colour_1) {
	case COLOUR_RED :
		switch (colour_2) {
		case COLOUR_GREEN :
			result = COLOUR_YELLOW;
			break;
		case COLOUR_BLUE :
			result = COLOUR_VIOLET;
			break;
		case COLOUR_CYAN :
			result = COLOUR_WHITE;
			break;
		default :
			result = colour_1;
		}
		break;
	case COLOUR_GREEN :
		switch (colour_2) {
		case COLOUR_RED :
			result = COLOUR_YELLOW;
			break;
		case COLOUR_BLUE :
			result = COLOUR_CYAN;
			break;
		case COLOUR_VIOLET :
			result = COLOUR_WHITE;
			break;
		default :
			result = colour_1;
		}
		break;
	case COLOUR_BLUE :
		switch (colour_2) {
		case COLOUR_RED :
			result = COLOUR_VIOLET;
			break;
		case COLOUR_GREEN :
			result = COLOUR_CYAN;
			break;
		case COLOUR_YELLOW :
			result = COLOUR_WHITE;
			break;
		default :
			result = colour_1;
		}
		break;
	case COLOUR_VIOLET :
		switch (colour_2) {
		case COLOUR_GREEN :
		case COLOUR_YELLOW :
		case COLOUR_CYAN :
			result = COLOUR_WHITE;
			break;
		default :
			result = colour_1;
		}
		break;
	case COLOUR_YELLOW :
		switch (colour_2) {
		case COLOUR_BLUE :
		case COLOUR_VIOLET :
		case COLOUR_CYAN :
			result = COLOUR_WHITE;
			break;
		default :
			result = colour_1;
		}
		break;
	case COLOUR_CYAN :
		switch (colour_2) {
		case COLOUR_RED :
		case COLOUR_VIOLET :
		case COLOUR_YELLOW :
			result = COLOUR_WHITE;
			break;
		default :
			result = colour_1;
		}
		break;
	case COLOUR_WHITE :
		result = COLOUR_WHITE;
		break;
	default :
		result = colour_2;
	}
	
	return result;
}

/**
 * compute_ray:
 * @x: Start horizontal coordinate of the ray
 * @y: Start vertical coordinate of the ray
 * @direction: Direction of the ray (from 0 to 7)
 * @colour: Colour index of the ray
 * @fixpoint_reached: Put to false when a synthetised light has changed
 *
 * Computes a ray by checking colisions and reflections at each step
 * in direction, and marks the ray[direction][y][x] with (1 << colour)
 * (because a ray can hold several colour. The board->rays array
 * contains two rows and two columns surrounding the map where
 * infinite rays are stored.
 */
static void compute_ray (board_t *board, int x, int y, guint8 direction, guint8 colour, gboolean hidden, gboolean *fixpoint_reached) {
	/* δx and  δy of a step, indexed by direction */
  int dx[] = {-1,-1, 0, 1, 1, 1, 0,-1};
  int dy[] = { 0, 1, 1, 1, 0,-1,-1,-1};

	/* stop recursion if ray already exists*/
	if (board->rays [(8 - (int) direction) % 8][y + 1][x + 1].colour_id & (1 << colour)) {
		return;
	}

	/* compute path step by step */
  int xc = x, yc = y;
  do {
    board->rays [(8 - (int) direction) % 8][yc + 1][xc + 1].colour_id |= 1 << colour;
		if (hidden) {
			board->rays [(8 - (int) direction) % 8][yc + 1][xc + 1].hidden |= 1 << colour;
		} else {
			board->rays [(8 - (int) direction) % 8][yc + 1][xc + 1].hidden &= ~ (1 << colour);
		}
    xc += dx[(int) direction];
    yc += dy[(int) direction];
  } while (xc >= 0 && xc < board->level->width
					 && yc >= 0 && yc < board->level->height
					 && board->map[yc][xc].type == NONE);
	
  if (xc >= 0 && xc < board->level->width
      && yc >= 0 && yc < board->level->height) {
		/* if the ray ends in the board, perform reflections */
    switch (board->map[yc][xc].type) {
    case MAP_HIDER : /* (un)hide ray */
			compute_ray (board, xc, yc, direction, colour, !hidden, fixpoint_reached);
      break;
    case MAP_DIODE : /* filter ray direction */
      if ((board->map[yc][xc].direction + direction) % 8 == 0) {
				compute_ray (board, xc, yc, direction, colour, hidden, fixpoint_reached);
			}
      break;
    case MAP_TARGET : /* target  */
			compute_ray (board, xc, yc, direction, colour, hidden, fixpoint_reached);
      break;
    case MAT_MIR : /* one sided mirror  */
      if ((board->map[yc][xc].direction + 4 + direction) % 8 == 1) {
				compute_ray (board, xc, yc, (direction + 2) % 8, colour, hidden, fixpoint_reached);
      }
      if ((board->map[yc][xc].direction + 4 + direction) % 8 == 7) {
				compute_ray (board, xc, yc, (direction + 6) % 8, colour, hidden, fixpoint_reached);
      }
      break;
    case MAT_CROSS : /* two sided mirror  */
      if ((board->map[yc][xc].direction + 4 + direction) % 4 == 1) {
				compute_ray (board, xc, yc, (direction + 2) % 8, colour, hidden, fixpoint_reached);
      }
      if ((board->map[yc][xc].direction + 4 + direction) % 4 == 3) {
				compute_ray (board, xc, yc, (direction + 6) % 8, colour, hidden, fixpoint_reached);
      }
      break;
    case MAT_DUP : /* wedge of a tight prism */
      if ((board->map[yc][xc].direction + 4 + direction) % 8 == 0) {
				compute_ray (board, xc, yc, (direction + 1) % 8, colour, hidden, fixpoint_reached);
				compute_ray (board, xc, yc, (direction + 7) % 8, colour, hidden, fixpoint_reached);
      }
      if ((board->map[yc][xc].direction + 4 + direction) % 8 == 3) {
				compute_ray (board, xc, yc, (direction + 1) % 8, colour, hidden, fixpoint_reached);
      }
      if ((board->map[yc][xc].direction + 4 + direction) % 8 == 5) {
				compute_ray (board, xc, yc, (direction + 7) % 8, colour, hidden, fixpoint_reached);
      }
      break;
    case MAT_TEE : /* wedge of a prism */
      if ((board->map[yc][xc].direction + 4 + direction) % 8 == 0) {
				compute_ray (board, xc, yc, (direction + 2) % 8, colour, hidden, fixpoint_reached);
				compute_ray (board, xc, yc, (direction + 6) % 8, colour, hidden, fixpoint_reached);
      }
      if ((board->map[yc][xc].direction + 4 + direction) % 8 == 2) {
				compute_ray (board, xc, yc, (direction + 2) % 8, colour, hidden, fixpoint_reached);
      }
      if ((board->map[yc][xc].direction + 4 + direction) % 8 == 6) {
				compute_ray (board, xc, yc, (direction + 6) % 8, colour, hidden, fixpoint_reached);
      }
      break;
    case MAT_MUX : /* RGB Muxer/Demuxer */
      if ((board->map[yc][xc].direction + 4 + direction) % 8 == 0) {
				if (colour == COLOUR_WHITE || colour == COLOUR_RED
						|| colour == COLOUR_VIOLET || colour == COLOUR_YELLOW) {
					compute_ray (board, xc, yc, (direction + 1) % 8, COLOUR_RED, hidden, fixpoint_reached);
				}
				if (colour == COLOUR_WHITE || colour == COLOUR_GREEN
						|| colour == COLOUR_CYAN || colour == COLOUR_YELLOW) {
					compute_ray (board, xc, yc, (direction) % 8, COLOUR_GREEN, hidden, fixpoint_reached);
				}
				if (colour == COLOUR_WHITE || colour == COLOUR_BLUE
						|| colour == COLOUR_VIOLET || colour == COLOUR_CYAN) {
					compute_ray (board, xc, yc, (direction + 7) % 8, COLOUR_BLUE, hidden, fixpoint_reached);
				}
      }
      if ((board->map[yc][xc].direction + 4 + direction) % 8 == 3) {
				if (colour == COLOUR_BLUE || colour == COLOUR_VIOLET || colour == COLOUR_CYAN) {
					guint8 sum = colour_sum (board->synthetised[yc][xc], COLOUR_BLUE);
					if (sum != board->synthetised[yc][xc]) {
						*fixpoint_reached = FALSE;
						board->synthetised[yc][xc] = sum;
					}
				}
      }
      if ((board->map[yc][xc].direction + 4 + direction) % 8 == 4) {
				if (colour == COLOUR_GREEN || colour == COLOUR_YELLOW || colour == COLOUR_CYAN) {
					guint8 sum = colour_sum (board->synthetised[yc][xc], COLOUR_GREEN);
					if (sum != board->synthetised[yc][xc]) {
						*fixpoint_reached = FALSE;
						board->synthetised[yc][xc] = sum;
					}
				}
      }
      if ((board->map[yc][xc].direction + 4 + direction) % 8 == 5) {
				if (colour == COLOUR_RED || colour == COLOUR_VIOLET || colour == COLOUR_YELLOW) {
					guint8 sum = colour_sum (board->synthetised[yc][xc], COLOUR_RED);
					if (sum != board->synthetised[yc][xc]) {
						*fixpoint_reached = FALSE;
						board->synthetised[yc][xc] = sum;
					}
				}
			}
      break;
    default : /* for others, no reflection */
      break;
    }
  } else {
    /* else an infinite ray is computed */
    board->rays [(8 - (int) direction) % 8][yc + 1][xc + 1].colour_id |= 1 << colour;
		if (hidden) {
			board->rays [(8 - (int) direction) % 8][yc + 1][xc + 1].hidden |= 1 << colour;
		} else {
			board->rays [(8 - (int) direction) % 8][yc + 1][xc + 1].hidden &= ~ (1 << colour);
		}
  }
}

/**
 * compute_rays:
 *
 * Computes the ray of each beamer with compute_ray.
 */
static void compute_rays (board_t *board) {
	/* clear old synthetized beams */
	for (int y = 0;y < board->level->height;y++) {
		for (int x = 0;x < board->level->width;x++) {
			board->synthetised[y][x] = 0;
		}
	}

	/* start a ray for each beamer and each synthetised source, until fixpoint */
	gboolean fixpoint_reached = FALSE;
	while (!fixpoint_reached) {
		/* clear last iteration */
		for (int c = 0;c < 8;c++) {
			for (int y = 0;y < board->level->height + 2;y++) {
				for (int x = 0;x < board->level->width + 2;x++) {
					board->rays[c][y][x].colour_id = 0;
					board->rays[c][y][x].hidden = 0;
				}
			}
		}
		fixpoint_reached = TRUE; /* supposed fixpoint reached, comput_ray may say if not */
		for (int y = 0;y < board->level->height;y++) {
			for (int x = 0;x < board->level->width;x++) {
				if (board->map[y][x].type == MAP_BEAMER) {
					compute_ray (board, x, y, board->map[y][x].direction, board->map[y][x].colour, FALSE, &fixpoint_reached);
				}
				if (board->map[y][x].type == MAT_MUX) {
					if (board->synthetised[y][x] != 0) {
						compute_ray (board, x, y, (8 - board->map[y][x].direction) % 8,
												 board->synthetised[y][x], FALSE, &fixpoint_reached);
					}
				}
			}

		}
	}
}

static void set (board_t *board, int x, int y) {
	struct item **map = board->level->map;

	if (y < 0 || y >= board->level->height
			|| x < 0 || x >= board->level->width
			|| (map[y][x].type == board->type && map[y][x].colour == board->colour)
			|| y == board->level->real_height
			|| (y > board->level->real_height && !IS_MATERIAL (board->type))) {
		return;
	}

	if (IS_MATERIAL (map[y][x].type)) {
		board->level->materials[map[y][x].type - 1]--;
	}

	/* if adding an object */
	if (IS_MATERIAL (board->type)) {
		board->level->materials[board->type - 1]++;
	}

	map[y][x].type = board->type;
	map[y][x].colour = board->colour;

	/* check if new lines are to add */
	int nb_materials = 0;
	for (int i = 0;i < MATERIALS;i++) {
		nb_materials += board->level->materials[i];
	}
	if (nb_materials / board->level->width + ((nb_materials % board->level->width == 0) ? 0 : 1)
		> board->level->height - board->level->real_height - 1) {
		/* one line step can be assumed */
		map = g_realloc (map, (board->level->height + 1) * sizeof (struct item*));
		map[board->level->height] = g_malloc (board->level->width * sizeof (struct item));
		for (int j = 0;j < board->level->width;j++) {
			map[board->level->height][j].type = NONE;
			map[board->level->height][j].direction = 0;
		}
		board->level->map = map;
		board->map = map;
		for (int i = 0;i < 8;i++) {
			board->rays[i] = g_realloc (board->rays[i], (board->level->height + 3) * sizeof (struct ray*));
			board->rays[i][board->level->height + 2] = g_malloc ((board->level->width + 2) * sizeof (struct ray));
		}
		board->synthetised = g_realloc (board->synthetised, (board->level->height + 3) * sizeof (guint8*));
		board->synthetised[board->level->height + 2] = g_malloc ((board->level->width + 2) * sizeof (guint8));
		board->level->height++;
	}
	/* ... or lines to remove */
	if (nb_materials / board->level->width + ((nb_materials % board->level->width == 0) ? 0 : 1)
		< board->level->height - board->level->real_height - 1) {
		for (int i = 0;i < board->level->width;i++) {
			if (map[board->level->height - 1][i].type != NONE) {
				gboolean found = FALSE;
				for (int ry = board->level->real_height + 1;ry < board->level->height - 1 && !found;ry++) {
					for (int rx = 0;rx < board->level->width && !found;rx++) {
						if (map[ry][rx].type == NONE) {
							found = TRUE;
							map[ry][rx].type = map[board->level->height - 1][i].type;
							map[ry][rx].colour = map[board->level->height - 1][i].colour;
							map[ry][rx].direction = map[board->level->height - 1][i].direction;
						}
					}
				}
			}
		}
		g_free (map[board->level->height - 1]);
		map = g_realloc (map, (board->level->height - 1) * sizeof (struct item*));
		board->level->map = map;
		board->map = map;
		for (int i = 0;i < 8;i++) {
			g_free (board->rays[i][board->level->height + 1]);
			board->rays[i] = g_realloc (board->rays[i], (board->level->height + 1) * sizeof (struct ray*));
		}
		g_free (board->synthetised[board->level->height + 1]);
		board->synthetised = g_realloc (board->synthetised, (board->level->height + 1) * sizeof (guint8*));
		board->level->height--;
	}
	compute_rays (board);
	paint_cb (board->widget, NULL, board);
	save_for_undo (board);
}

/**
 * button_release_cb:
 * @widget: The board
 * @event: The mouse event
 * @data: unused
 *
 * Calculates map coordinates and callc laser_board_click ()
 */
static gboolean button_release_cb (GtkWidget *widget, GdkEventButton *event, board_t *board) {
	/* only if game loaded */
  if (board->loaded) {
		board->pressed = FALSE;
		if (board->dragged) {
			board->actions++;
			save_for_undo (board);
			return TRUE;
		}
		if (board->mode != SET_MODE) {
			/* compute sizes */
			gint ox, oy, d;
			get_sizes (board, &ox, &oy, &d);
			/* click with calculated map coordinates */
			int new_x = (event->x - ox) / d;
			int new_y = (event->y - oy) / d;
			if (board->selected	&& board->select_x == new_x && board->select_y == new_y) {
				switch (event->button) {
				case 1 :
					board->map[new_y][new_x].direction = (board->map[new_y][new_x].direction + 1) % 8;
					break;
				case 3 :
					board->map[new_y][new_x].direction = (board->map[new_y][new_x].direction + 7) % 8;
					break;
				}
				board->actions++;
				save_for_undo (board);
			} else {
				if (new_x >= 0 && new_x < board->level->width && new_y >= 0 && new_y < board->level->height) {
					if (board->selected && board->map[new_y][new_x].type == NONE
							&& (IS_MATERIAL (board->map[board->select_y][board->select_x].type)
									|| (board->mode != PLAY_MODE
											&& EDIT_POSITION (board->map[board->select_y][board->select_x].type)
											&& new_y < board->level->real_height
											&& board->select_y < board->level->real_height))) {
						board->map[new_y][new_x] = board->map[board->select_y][board->select_x];
						board->map[board->select_y][board->select_x].type = NONE;
					}
					if (IS_MATERIAL (board->map[new_y][new_x].type)
							|| (board->mode != PLAY_MODE
									&& EDIT_POSITION (board->map[new_y][new_x].type)
									&& new_y < board->level->real_height)) {
						board->selected = TRUE;
						board->select_x = new_x;
						board->select_y = new_y;
					}
				}
			}
			compute_rays(board);
			paint_cb (widget, NULL, board);
			update_state (board);
		}

		return TRUE;
	} else {
		return FALSE;
	}
}

/**
 * button_press_cb:
 * @widget: The board
 * @event: The mouse event
 * @data: unused
 *
 * Calculates map coordinates and start "drag"
 */
static gboolean button_press_cb (GtkWidget *widget, GdkEventButton *event, board_t *board) {
	/* only if game loaded */
  if (board->loaded) {
		/* compute sizes */
		gint ox, oy, d;
		get_sizes (board, &ox, &oy, &d);

		board->last_drag_x = (event->x - ox) / d;
		board->last_drag_y = (event->y - oy) / d;

		board->dragged = FALSE;
		board->pressed = TRUE;

		if (board->mode == SET_MODE) {
			set (board, board->last_drag_x, board->last_drag_y);
			compute_rays(board);
			paint_cb (widget, NULL, board);
			update_state (board);
		}

		return TRUE;
	} else {
		return FALSE;
	}
}

/**
 * mouse_move_cb:
 * @widget: The board
 * @event: The mouse event
 * @data: unused
 *
 * Calculates map coordinates and performs "drag"
 */
static gboolean mouse_move_cb (GtkWidget *widget, GdkEventMotion *event, board_t *board) {
	if (board->loaded && board->pressed) {
		/* compute sizes */
		gint ox, oy, d;
		get_sizes (board, &ox, &oy, &d);
		/* click with calculated map coordinates if changed */
		int new_x = (event->x - ox) / d;
		int new_y = (event->y - oy) / d;
		
		if (new_x >= 0 && new_x < board->level->width	&& new_y >= 0 && new_y < board->level->height) {
			if (board->mode == SET_MODE) {
				set (board, board->last_drag_x, board->last_drag_y);
				board->last_drag_x = new_x;
				board->last_drag_y = new_y;
			} else {
				if (board->last_drag_x != new_x	|| board->last_drag_y != new_y) {
					if (board->map[new_y][new_x].type == NONE
							&& (IS_MATERIAL (board->map[board->last_drag_y][board->last_drag_x].type)
									|| (board->mode != PLAY_MODE
											&& EDIT_POSITION (board->map[board->last_drag_y][board->last_drag_x].type)
											&& board->last_drag_y < board->level->real_height
											&& new_y < board->level->real_height))) {
						board->dragged = TRUE;
						board->map[new_y][new_x] = board->map[board->last_drag_y][board->last_drag_x];
						board->map[board->last_drag_y][board->last_drag_x].type = NONE;
						board->last_drag_x = new_x;
						board->last_drag_y = new_y;
						board->selected = TRUE;
						board->select_x = new_x;
						board->select_y = new_y;
						compute_rays(board);
						paint_cb (widget, NULL, board);
						update_state (board);
					}
				}
			}
		}
		return TRUE;
	} else {
		return FALSE;
	}
}

/**
 * scroll_cb:
 * @widget: The board
 * @event: The scroll event
 * @data: unused
 *
 * Calculates map coordinates and rotates the object under the mouse
 */
static gboolean scroll_cb (GtkWidget *widget, GdkEventScroll *event, board_t *board) {
	if (board->loaded) {
		/* compute sizes */
		gint ox, oy, d;
		get_sizes (board, &ox, &oy, &d);
		/* click with calculated map coordinates if changed */
		int x = (event->x - ox) / d;
		int y = (event->y - oy) / d;
		
		if ((x >= 0 && x < board->level->width	&& y >= 0 && y < board->level->height)
				&& (IS_MATERIAL (board->map[y][x].type)
						|| (board->mode != PLAY_MODE && EDIT_ROTATION (board->map[y][x].type)))) {
			board->selected = TRUE;
			board->select_x = x;
			board->select_y = y;
			switch (event->direction) {
			case GDK_SCROLL_UP :
			case GDK_SCROLL_LEFT :
				board->map[y][x].direction = (board->map[y][x].direction + 7) % 8;
				board->actions++;
				break;
			case GDK_SCROLL_DOWN :
			case GDK_SCROLL_RIGHT :
				board->map[y][x].direction = (board->map[y][x].direction + 1) % 8;
				board->actions++;
				break;
			}
			save_for_undo (board);
			compute_rays(board);
			paint_cb (widget, NULL, board);
			update_state (board);

			return TRUE;
		}
	}
	return FALSE;		
}

board_t* laser_board_attach (GtkWidget *widget,
														 board_mode mode,
														 LaserConfig* config, 
														 void (*update_score_cb) (gpointer, gint, gint, gint, gboolean, gboolean),
														 gboolean *modified,
														 gpointer update_score_data) {
	board_t *board = g_malloc (sizeof (struct board));
	
	board->loaded = FALSE;
	board->config = config;
	board->widget = widget;
	board->mode = mode;
	board->update_score_cb = update_score_cb;
	board->update_score_data = update_score_data;
	board->cache = sprite_cache_new (config, 2);
	board->modified = modified;

	g_signal_connect (board->widget, "expose-event",         G_CALLBACK (paint_cb), board);
  g_signal_connect (board->widget, "button-release-event", G_CALLBACK (button_release_cb), board);
  g_signal_connect (board->widget, "button-press-event",   G_CALLBACK (button_press_cb), board);
  g_signal_connect (board->widget, "motion-notify-event",  G_CALLBACK (mouse_move_cb), board);
  g_signal_connect (board->widget, "scroll-event",         G_CALLBACK (scroll_cb), board);

	return board;
}

void laser_board_set_level (board_t* board, level *level) {
	board->undo_pointer = 0;
	board->undo_max = 0;
	board->selected = FALSE;
	board->actions = 0;
	board->level = level;
	board->loaded = TRUE;
	if (board->mode == PLAY_MODE) {
		board->map = malloc (board->level->height * sizeof (struct item*));
		for (gint i = 0;i < board->level->height;i++) {	
			board->map[i] = malloc (board->level->width * sizeof (struct item));
			memcpy (board->map[i], board->level->map[i], board->level->width * sizeof (struct item));
		}
	} else {
		board->map = board->level->map;
	}
	for (gint d = 0;d < 8;d++) {
		board->rays[d] = malloc ((board->level->height + 2) * sizeof (struct ray*));
		for (gint i = 0;i < board->level->height + 2;i++) {	
			board->rays[d][i] = malloc ((board->level->width + 2) * sizeof (struct ray));
		}
	}
	board->synthetised = malloc ((board->level->height + 2) * sizeof (guint8*));
	for (gint i = 0;i < board->level->height + 2;i++) {	
		board->synthetised[i] = malloc ((board->level->width + 2) * sizeof (guint8));
	}
	save_for_undo (board);
	compute_rays (board);
	paint_cb (board->widget, NULL, board);
	update_state (board);
}

void laser_board_paint (board_t *board) {
	paint_cb (board->widget, NULL, board);
}

void laser_board_set_actions (board_t *board, gint actions) {
	board->actions = actions;
	update_state (board);
}

void laser_board_set_mode (board_t *board, board_mode mode, guint16 type, guint8 colour) {
	board->mode = mode;
	board->type = type;
	board->colour = colour;
}

void laser_board_reset (board_t *board) {
	if (board->loaded) {
		board->selected = FALSE;
		int ny = board->level->real_height + 1;
		int nx = 0;
		for (int y = 0;y < board->level->real_height;y++) {
			for (int x = 0;x < board->level->width;x++) {
				if (board->map[y][x].type > NONE && board->map[y][x].type <= MATERIALS) {
					while (board->map[ny][nx].type != NONE) {
						nx++;
						if (nx >= board->level->width) {
							nx = 0;
							ny++;
						}
					}
					board->map[ny][nx].type = board->map[y][x].type;
					board->map[ny][nx].colour = board->map[y][x].colour;
					board->map[ny][nx].direction = 0;
					board->map[y][x].type = NONE;
				}
			}
		}
		board->actions = 0;
		compute_rays (board);
		paint_cb (board->widget, NULL, board);
		update_state (board);
	save_for_undo (board);
	}
}

void laser_board_undo (board_t *board) {
	if (board->undo_pointer > 1) {
		board->selected = FALSE;
		board->undo_pointer--;
		for (int i = 0;i < board->level->height;i++) {
			g_free (board->map[i]);
		}
		g_free (board->map);

		struct item **map = g_malloc (board->undo_stack[board->undo_pointer].height * sizeof (struct item*));
		for (int i = 0;i < board->undo_stack[board->undo_pointer].height;i++) {
			map[i] = g_malloc (board->undo_stack[board->undo_pointer].width * sizeof (struct item));
			memcpy (map[i],
							board->undo_stack[board->undo_pointer].map[i],
							board->undo_stack[board->undo_pointer].width * sizeof (struct item));
		}
		board->map = map;

		if (board->mode != PLAY_MODE) {
			board->level->map = board->map;
			board->level->width = board->undo_stack[board->undo_pointer].width;
			board->level->height = board->undo_stack[board->undo_pointer].height;
			board->level->real_height = board->undo_stack[board->undo_pointer].real_height;
			for (int i = 0;i < MATERIALS;i++) {
				board->level->materials[i] = board->undo_stack[board->undo_pointer].materials[i];
			}
		}
		
		for (gint d = 0;d < 8;d++) {
			board->rays[d] = malloc ((board->level->height + 2) * sizeof (struct ray*));
			for (gint i = 0;i < board->level->height + 2;i++) {	
				board->rays[d][i] = malloc ((board->level->width + 2) * sizeof (struct ray));
			}
		}
		board->synthetised = malloc ((board->level->height + 2) * sizeof (guint8*));
		for (gint i = 0;i < board->level->height + 2;i++) {	
			board->synthetised[i] = malloc ((board->level->width + 2) * sizeof (guint8));
		}
		compute_rays (board);
		paint_cb (board->widget, NULL, board);
		update_state (board);
	}
}

void laser_board_redo (board_t *board) {
	if (board->undo_pointer < board->undo_max) {
		board->selected = FALSE;
		board->undo_pointer++;
		for (int i = 0;i < board->level->height;i++) {
			g_free (board->map[i]);
		}
		g_free (board->map);

		struct item **map = g_malloc (board->undo_stack[board->undo_pointer].height * sizeof (struct item*));
		for (int i = 0;i < board->undo_stack[board->undo_pointer].height;i++) {
			map[i] = g_malloc (board->undo_stack[board->undo_pointer].width * sizeof (struct item));
			memcpy (map[i],
							board->undo_stack[board->undo_pointer].map[i],
							board->undo_stack[board->undo_pointer].width * sizeof (struct item));
		}
		board->map = map;

		if (board->mode != PLAY_MODE) {
			board->level->map = board->map;
			board->level->width = board->undo_stack[board->undo_pointer].width;
			board->level->height = board->undo_stack[board->undo_pointer].height;
			board->level->real_height = board->undo_stack[board->undo_pointer].real_height;
			for (int i = 0;i < MATERIALS;i++) {
				board->level->materials[i] = board->undo_stack[board->undo_pointer].materials[i];
			}
		}
		
		for (gint d = 0;d < 8;d++) {
			board->rays[d] = malloc ((board->level->height + 2) * sizeof (struct ray*));
			for (gint i = 0;i < board->level->height + 2;i++) {	
				board->rays[d][i] = malloc ((board->level->width + 2) * sizeof (struct ray));
			}
		}
		board->synthetised = malloc ((board->level->height + 2) * sizeof (guint8*));
		for (gint i = 0;i < board->level->height + 2;i++) {	
			board->synthetised[i] = malloc ((board->level->width + 2) * sizeof (guint8));
		}
		compute_rays (board);
		paint_cb (board->widget, NULL, board);
		update_state (board);	
	}
}

void laser_board_resize (board_t *board, guint grow_or_shrink, guint direction) {
	guint8 type_save = board->type;

	board->type = NONE;

	/* hacky :D */
	if (grow_or_shrink == -1) {
		switch (direction) {
		case 2 :
			for (int i = 0;i < board->level->real_height;i++) {
				set (board, 0, i);
			}
			break;
		case 3 :
			for (int i = 0;i < board->level->real_height;i++) {
				set (board, board->level->width - 1, i);
			}
			break;
		case 0 :
			for (int i = 0;i < board->level->width;i++) {
				set (board, i, 0);
			}
			break;
		case 1 :
			for (int i = 0;i < board->level->width;i++) {
				set (board, i, board->level->real_height - 1);
			}
			break;
		}
	}

	board->type = type_save;

	switch (direction) {
	case 0 : case 1 :
		board->level->height += grow_or_shrink;
		board->level->real_height += grow_or_shrink;
		break;
	default :
		board->level->width = board->level->width + grow_or_shrink;
	}

	struct item **map;

	map = malloc (board->level->height * sizeof (struct item*));
	for (gint i = 0;i < board->level->height;i++) {	
		map[i] = g_malloc0 (board->level->width * sizeof (struct item));
	}
	for (gint i = 0;i < board->level->height;i++) {
		switch (direction) {
		case 0 :
			if (grow_or_shrink == 1) {
				if (i < board->level->height - 1) {
					memcpy (map[i + 1], board->map[i], board->level->width * sizeof (struct item));
				}
			} else {
				memcpy (map[i], board->map[i + 1], board->level->width * sizeof (struct item));
			}
			break;
		case 1 :
			if (grow_or_shrink == -1) {
				if (i < board->level->real_height) {
					memcpy (map[i], board->map[i], board->level->width * sizeof (struct item));
				} else {
					if (i > board->level->real_height) {
						memcpy (map[i], board->map[i + 1], board->level->width * sizeof (struct item));
					}
				}
			} else {
				if (i < board->level->real_height - 1) {
					memcpy (map[i], board->map[i], board->level->width * sizeof (struct item));
				} else {
					if (i > board->level->real_height) {
						memcpy (map[i], board->map[i - 1], board->level->width * sizeof (struct item));
					}
				}
			}
			break;
		case 2 :
			if (grow_or_shrink == -1) {
				memcpy (map[i], &board->map[i][1], (board->level->width) * sizeof (struct item));
			} else {
				memcpy (&map[i][1], board->map[i], (board->level->width - 1) * sizeof (struct item));
			}
			break;
		case 3 :
			if (grow_or_shrink == -1) {
				memcpy (map[i], board->map[i], (board->level->width) * sizeof (struct item));
			} else {
				memcpy (map[i], board->map[i], (board->level->width - 1) * sizeof (struct item));
			}
			break;
		}
	}

	for (int i = 0;i < board->level->width;i++) {
		map[board->level->real_height][i].type = MAP_WALL;
	}

	board->map = map;
	board->level->map = map;
	
	for (gint d = 0;d < 8;d++) {
		board->rays[d] = malloc ((board->level->height + 2) * sizeof (struct ray*));
		for (gint i = 0;i < board->level->height + 2;i++) {	
			board->rays[d][i] = malloc ((board->level->width + 2) * sizeof (struct ray));
		}
	}
	board->synthetised = malloc ((board->level->height + 2) * sizeof (guint8*));
	for (gint i = 0;i < board->level->height + 2;i++) {	
		board->synthetised[i] = malloc ((board->level->width + 2) * sizeof (guint8));
	}
	compute_rays (board);
	paint_cb (board->widget, NULL, board);
	update_state (board);
	save_for_undo (board);
}

gchar* laser_board_get_cheat (board_t *board) {
	if (board->loaded) {
		int nb_materials = 0;
		for (int m = 0;m < MATERIALS;m++) {
			nb_materials += board->level->materials[m];
		}
		gchar code[nb_materials * (sizeof (int) * 8 + 1)];
		int s = 0;
		for (int y = 0;y < board->level->height;y++) {
			for (int x = 0;x < board->level->width;x++) {
				if (board->map[y][x].type > NONE && board->map[y][x].type <= MATERIALS) {
					guint32 item = ((board->level->width * y + x) * 8 + board->map[y][x].direction) * 20 + board->map[y][x].type;
					if (s == 0) {
						s += sprintf (&code[s], "%X", item);
					} else {
						s += sprintf (&code[s], "-%X", item);
					}
				}
			}
		}
		return g_strdup (code);
	} else {
		return NULL;
	}
}

void   laser_board_set_cheat (board_t *board, gchar* code) {
	if (board->loaded) {
		int read_materials[MATERIALS];
		for (int m = 0;m < MATERIALS;m++) {
			read_materials[m] = 0;
		}
		for (int y = 0;y < board->level->height;y++) {
			for (int x = 0;x < board->level->width;x++) {
				if (board->map[y][x].type > NONE && board->map[y][x].type <= MATERIALS) {
					board->map[y][x].type = NONE;
				}
			}
		}
		gchar *s = code, *s2 = code;
		while (TRUE) {
			while (*s2 != '\0' && *s2 != '-') {
				s2++;
			}
			guint32 item = 0;
			sscanf (s, "%X", &item);
			guint16 itype = item % 20;
			item /= 20;
			guint8 idirection = item % 8;
			item /= 8;
			int ix = item % board->level->width;
			item /= board->level->width;
			int iy = item;
			if (iy < board->level->height && iy != board->level->real_height) {
				if (board->map[iy][ix].type == NONE && read_materials[itype - 1] < board->level->materials[itype - 1]) {
					read_materials[itype - 1] ++;
					board->map[iy][ix].type = itype;
					board->map[iy][ix].direction = idirection;
				}
			}
			if (*s2 == '\0') {
				break;
			} else {
				s = ++s2;
			}
		}
		int x = 0, y = board->level->real_height + 1;
		for (int m = 0;m < MATERIALS;m++) {
			for (int i = read_materials[m];i < board->level->materials[m];i++) {
				board->map[y][x].type = m + 1;
				board->map[y][x].direction = 0;
				x++;
				if (x == board->level->width) {
					x = 0;
					y++;
				}
			}
		}
		board->selected = FALSE;
		compute_rays (board);
		laser_board_paint (board);
		update_state (board);
		save_for_undo (board);
	}
}
