/* -*- Mode: C; tab-width: 2; indent-tabs-mode: t; c-basic-offset: 2 -*-
 *
 * LASER
 * Copyright (C) 2006 Benjamin Canou <bookeldor@gmail.com>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gdk/gdk.h>
#include <string.h>
#include <gtk/gtk.h>
#include <glade/glade.h>
#include <gconf/gconf-client.h>
#include <libgnomeui/libgnomeui.h>
#include <libgnomeui/gnome-help.h>
#include "config.h"
#include "laser-common.h"

struct materials_asoc materials_assoc[MATERIALS] = {
	{"Mir", MAT_MIR - 1}, {"Cross", MAT_CROSS - 1},
	{"Dup", MAT_DUP - 1},	{"Mux", MAT_MUX - 1},
	{"Tee", MAT_TEE - 1}
};

void colour_from_string (colour_t *colour, gchar str[10]) {
	memcpy (colour->str, str, 10);
	gboolean format_correct = TRUE;
	
	/* check if format is correct : #RRGGBBAA (hexa) */
	if (colour->str [0] == '#') {
		for (int i = 1;i < 9;i++) {
			if (! ((colour->str[i] >= '0' && colour->str[i] <= '9')
						 || (colour->str[i] >= 'a' && colour->str[i] <= 'f')
						 || (colour->str[i] >= 'A' && colour->str[i] <= 'F'))) {
				format_correct = FALSE;
			}
		}
	} else {
		format_correct = FALSE;
	}
	
  if (format_correct) {
    /* if format correct, parse */ 
    int ir, ig, ib, ia;
    sscanf (colour->str, "#%02x%02x%02x%02x", &ir, &ig, &ib, &ia);
    colour->r = (double) ir / 255.0;
    colour->g = (double) ig / 255.0;
    colour->b = (double) ib / 255.0;
    colour->a = (double) ia / 255.0;

		colour->gdk_colour.red = colour->r * 0xFFFF;
		colour->gdk_colour.green = colour->g * 0xFFFF;
		colour->gdk_colour.blue = colour->b * 0xFFFF;
		colour->gdk_alpha = colour->a * 0xFFFF;
  } else {
    /* else reset to default values */
		colour_from_string (colour, LASER_CONFIG_DEFAULT_COLOUR);
	}
}

void colour_from_gdk (colour_t *colour, GdkColor gdk_colour, guint16 gdk_alpha) {
	colour->gdk_colour = gdk_colour;
	colour->gdk_alpha = gdk_alpha;

	colour->r = ((gdouble) gdk_colour.red) / ((gdouble) 0xFFFF);
	colour->g = ((gdouble) gdk_colour.green) / ((gdouble) 0xFFFF);
	colour->b = ((gdouble) gdk_colour.blue) / ((gdouble) 0xFFFF);
	colour->a = ((gdouble) gdk_alpha) / ((gdouble) 0xFFFF);

	snprintf (colour->str, 10, "#%02X%02X%02X%02X",
						((gint) (colour->r * 255.0) % 256),
						((gint) (colour->g * 255.0) % 256),
						((gint) (colour->b * 255.0) % 256),
						((gint) (colour->a * 255.0) % 256));
}

void colour_from_cairo (colour_t *colour, gdouble r, gdouble g, gdouble b, gdouble a) {
	colour->r = r;
	colour->g = g;
	colour->b = b;
	colour->a = a;

	colour->gdk_colour.red = r * 0xFFFF;
	colour->gdk_colour.green = g * 0xFFFF;
	colour->gdk_colour.blue = b * 0xFFFF;
	colour->gdk_alpha = a * 0xFFFF;

	snprintf (colour->str, 10, "#%02X%02X%02X%02X",
						((gint) (r * 255.0) % 256),
						((gint) (g * 255.0) % 256),
						((gint) (b * 255.0) % 256),
						((gint) (a * 255.0) % 256));
}

colour_t* colour_new_from_string (gchar str[10]) {
	colour_t *colour = g_malloc (sizeof (struct colour_t));

	colour_from_string (colour, str);

	return colour;
}

colour_t* colour_new_from_gdk (GdkColor gdk_colour, guint16 gdk_alpha) {
	colour_t *colour = g_malloc (sizeof (struct colour_t));
	
	colour_from_gdk (colour, gdk_colour, gdk_alpha);
	
	return colour;
}

colour_t* colour_new_from_cairo (gdouble r, gdouble g, gdouble b, gdouble a) {
	colour_t *colour = g_malloc (sizeof (struct colour_t));

	colour_from_cairo (colour, r, g, b, a);

	return colour;
}

level_set* level_set_new () {
  level_set *level_set;

  if ((level_set = g_malloc (sizeof (struct level_set))) == NULL) {
    return NULL;
  }

  level_set->name_i18n = NULL;
  level_set->name = NULL;
  level_set->file = g_strdup ("");
  level_set->note_i18n = NULL;
  level_set->note = NULL;
  level_set->levels = NULL;
  level_set->nb_levels = 0;

  return level_set;
}

/**
 * free_i18n:
 *
 */
void free_i18n (GSList *string_i18n) {
	if (string_i18n != NULL) {
		for (GSList *l = string_i18n;l != NULL;l = g_slist_next (l)) {
			g_free (((gchar**)l->data)[0]);
			g_free (((gchar**)l->data)[1]);
			g_free (l->data);
		}
		g_slist_free (string_i18n);
	}
}

/**
 * dup_i18n:
 * @string_i18n: the internationalized string to duplicate
 * @localised_orig: the original localised string (pointer to an element of string_i18n)
 * @localised_dup: a place to store a pointer to the localised copy, or NULL
 *
 * Returns a deep copy of a i18n string or NULL if failed
 */
GSList* dup_i18n (GSList *string_i18n, gchar* localised_orig, gchar** localised_dup) {
	GSList *result = NULL;
	
	/* dup */
	for (GSList *l = string_i18n;l != NULL;l = g_slist_next (l)) {
		gchar **pair = g_malloc (2 * sizeof (gchar*));
		if (pair == NULL) {
			for (GSList *l = result;l != NULL;l = g_slist_next (l)) {
				g_slist_free (l);
			}
			return NULL;
		}
		pair[0] = NULL;
		pair[1] = NULL;
		result = g_slist_prepend (result, pair);
	}

	/* copy */
	for (GSList *l = string_i18n, *l2 = result;
			 l != NULL && l2 != NULL;
			 l = g_slist_next (l), l2 = g_slist_next (l2)) {
		gchar **pair = l->data;
		gchar **pair2 = l2->data;
		pair2[0] = g_strdup (pair[0]);
		pair2[1] = g_strdup (pair[1]);
		if (pair2[0] == NULL || pair2[1] == NULL) {
			if (localised_dup != NULL) {
				*localised_dup = NULL;
			}
			GSList *old;
			for (GSList *l = result;l != NULL;l = g_slist_next (old)) {
				old = l;
				if (((gchar**)l->data)[0] != NULL) {
					g_free (((gchar**)l->data)[0]);
				}
				if (((gchar**)l->data)[1] != NULL) {
					g_free (((gchar**)l->data)[1]);
				}
				g_slist_free (l);
			}
			return NULL;
		}
		if (localised_dup != NULL && localised_orig == pair[1]) {
			*localised_dup = pair2[1];
		}
		result = g_slist_prepend (result, pair);
	}

	return result;
}

/**
 * level_set_free:
 *
 */
void level_set_free (level_set *level_set) {
  free_i18n (level_set->name_i18n);
  free_i18n (level_set->note_i18n);
	if (level_set->file != NULL) {
		g_free (level_set->file);
	}
  for (int i = 0;i < level_set->nb_levels;i++) {
    free_i18n (level_set->levels[i].name_i18n);
    free_i18n (level_set->levels[i].comment_i18n);
    if (level_set->levels[i].solution != NULL) {
			free (level_set->levels[i].solution);
		}
		if (level_set->levels[i].map != NULL) {
			for (int y = 0;y < level_set->levels[i].height;y++) {
				if (level_set->levels[i].map[y] != NULL) {
					g_free (level_set->levels[i].map[y]);
				}
			}
			g_free (level_set->levels[i].map);
		}
  }
}

/**
 * dup_level_set:
 * @level_set: The level set to copy
 *
 * Returns a deep copy of a level set or NULL if failed.
 */
level_set* dup_level_set (level_set *level_set) {
	struct level_set *result = level_set_new ();
	gboolean error = FALSE;

	if (result == NULL) {
		return NULL;
	}

	result->file = g_strdup (level_set->file);
  dup_i18n (level_set->name_i18n, level_set->name, &result->name);
  dup_i18n (level_set->note_i18n, level_set->note, &result->note);
	result->nb_levels = level_set->nb_levels;
	result->levels = g_malloc (level_set->nb_levels * sizeof (struct level));
	if (result->levels != NULL) {
		for (int i = 0;i < level_set->nb_levels;i++) {
			result->levels[i].width = level_set->levels[i].width;
			result->levels[i].height = level_set->levels[i].height;
			result->levels[i].real_height = level_set->levels[i].real_height;
			for (int j = 0;j < MATERIALS;j++) {
				result->levels[i].materials[j] = level_set->levels[i].materials[j];
			}
			result->levels[i].name_i18n =
				dup_i18n (level_set->levels[i].name_i18n,
									level_set->levels[i].name, &result->levels[i].name);
			result->levels[i].comment_i18n =
				dup_i18n (level_set->levels[i].comment_i18n,
									level_set->levels[i].comment, &result->levels[i].comment);
			result->levels[i].map = g_malloc (result->levels[i].height * sizeof (struct item*));
			if (level_set->levels[i].name == NULL || level_set->levels[i].comment == NULL) {
				error = TRUE;
			}
			if (result->levels[i].map != NULL) {
				for (int y = 0;y < level_set->levels[i].height;y++) {
					level_set->levels[i].map[y] = g_malloc (level_set->levels[i].width * sizeof (struct item));
					if (level_set->levels[i].map[y] != NULL) {
						memcpy (result->levels[i].map[y], level_set->levels[i].map[y],
										level_set->levels[i].width * sizeof (struct item));
					} else {
						error = TRUE;
					}
				}
			} else {
				error = TRUE;
			}
		}
  } else {
		error = TRUE;
	}
	
	if (error) {
		level_set_free (result);
		return NULL;
	} else {
		return result;
	}
}

gboolean file_error_dialog(gchar *file, gchar *details) {
	GladeXML *ui = glade_xml_new (LASER_CONFIG_SHARE "file-error.glade", NULL, NULL);
	GtkWidget *dlg = glade_xml_get_widget (ui, "dialog");
	gtk_label_set_text (GTK_LABEL (glade_xml_get_widget (ui, "label_file")), file);
	gtk_label_set_markup (GTK_LABEL (glade_xml_get_widget (ui, "label_details")), details);
	gint r = (gtk_dialog_run (GTK_DIALOG (dlg)) == -3);
	gtk_widget_destroy (dlg);
	return r;
}
