/* -*- Mode: C; tab-width: 2; indent-tabs-mode: t; c-basic-offset: 2 -*-
 *
 * LASER
 * Copyright (C) 2006 Benjamin Canou <bookeldor@gmail.com>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <stdlib.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <gconf/gconf-client.h>
#include <cairo.h>
#include <librsvg/rsvg.h>
#include <librsvg/rsvg-cairo.h>
#include <gtk/gtk.h>
#include "config.h"
#include "laser-common.h"
#include "laser-parser.h"
#include "laser-config.h"

#define LASER_CONFIG_GET_PRIVATE(o) G_TYPE_INSTANCE_GET_PRIVATE (o, LASER_CONFIG_TYPE, LaserConfigPrivate)

#define CACHE_SIZE 23

struct theme_cache {
	LaserConfig*         config;
	guint                size;
	cairo_surface_t*     surface[CACHE_SIZE][8];
};

struct {
	guchar               type;
  guchar               colour;
  gchar*               file;
} theme_cache_static [CACHE_SIZE] = {
  {NONE,       NONE,          "background.svg"},
  {MAT_DUP,    NONE,          "duplicator.svg"},
  {MAT_TEE,    NONE,          "tee.svg"},
  {MAT_MIR,    NONE,          "mirror.svg"},
  {MAT_CROSS,  NONE,          "cross.svg"},
  {MAT_MUX,    NONE,          "muxer.svg"},
  {MAP_WALL,   NONE,          "wall.svg"},
  {MAP_HIDER,  NONE,          "hider.svg"},
  {MAP_DIODE,  NONE,          "diode.svg"},
  {MAP_BEAMER, COLOUR_RED,    "beamer_red.svg"},
  {MAP_BEAMER, COLOUR_GREEN,  "beamer_green.svg"},
  {MAP_BEAMER, COLOUR_BLUE,   "beamer_blue.svg"},
  {MAP_BEAMER, COLOUR_VIOLET, "beamer_violet.svg"},
  {MAP_BEAMER, COLOUR_YELLOW, "beamer_yellow.svg"},
  {MAP_BEAMER, COLOUR_CYAN,   "beamer_cyan.svg"},
  {MAP_BEAMER, COLOUR_WHITE,  "beamer_white.svg"},
  {MAP_TARGET, COLOUR_RED,    "target_red.svg"},
  {MAP_TARGET, COLOUR_GREEN,  "target_green.svg"},
  {MAP_TARGET, COLOUR_BLUE,   "target_blue.svg"},
  {MAP_TARGET, COLOUR_VIOLET, "target_violet.svg"},
  {MAP_TARGET, COLOUR_YELLOW, "target_yellow.svg"},
  {MAP_TARGET, COLOUR_CYAN,   "target_cyan.svg"},
  {MAP_TARGET, COLOUR_WHITE,  "target_white.svg"}
};

struct _LaserConfigPrivate {
  GConfClient*         gconf_client;
  gboolean             restore;
  gboolean             arrows;
  gboolean             show_stats;
  gchar*               theme;
	gchar*               level_set;
  guint                level;
  colour_t             colour;
	guint                width, height;
	guint                editor_width, editor_height;
	gint                 actions;
	gint                 clock;
	gchar*               cheat;
	GSList*              caches;
	guint                nb_caches;
	RsvgHandle*          svg_hnd[CACHE_SIZE];
};

G_DEFINE_TYPE (LaserConfig, laser_config, G_TYPE_OBJECT)

enum {
	COLOUR_CHANGED,
	RESTORE_CHANGED,
	ARROWS_CHANGED,
	SHOW_STATS_CHANGED,
	THEME_CHANGED,
	LAST_SIGNAL
};

static guint signals [LAST_SIGNAL] = { 0, };

static void     gconf_update_cb          (GConfClient *client,guint cnxn_id, GConfEntry *entry, gpointer data);
static void     laser_config_init        (LaserConfig *self);
static void     laser_config_finalize    (GObject *object);
static void     laser_config_class_init  (LaserConfigClass *klass);
static void     verify_level             (LaserConfig *self, gchar *level_set, guint level);
static void     verify_theme             (LaserConfig *self, gchar *theme);

/* gobject functions **********************************************************/
static void laser_config_init (LaserConfig *self) {
  self->priv = LASER_CONFIG_GET_PRIVATE (self);

	self->priv->caches = NULL;
	for (int i = 0; i < CACHE_SIZE; i++) {
		self->priv->svg_hnd[i] = NULL;
	}

	gchar *path;
	path = g_strconcat (g_get_user_config_dir (), "/laser/levels", NULL);
	if (!g_file_test (path, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR)) {
		if (g_mkdir_with_parents (path, 0755) == -1) {
			printf ("Error : cannot create directory \"%s\".\n", path);
		}
	}
	g_free (path);
	path = g_strconcat (g_get_user_config_dir (), "/laser/themes", NULL);
	if (!g_file_test (path, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR)) {
		if (g_mkdir_with_parents (path, 0755) == -1) {
			printf ("Error : cannot create directory \"%s\".\n", path);
		}
	}
	g_free (path);

	self->priv->gconf_client = gconf_client_get_default();
  gconf_client_add_dir(self->priv->gconf_client, LASER_GCONF_PATH, GCONF_CLIENT_PRELOAD_NONE, NULL);

  GConfValue *value;
  GError *err = NULL;

	/* theme */
  value = gconf_client_get (self->priv->gconf_client, LASER_GCONF_KEY_THEME, &err);
	gchar *theme = g_strdup (LASER_CONFIG_DEFAULT_THEME);
  if (err != NULL) {
    g_print (_("Error : %s\n"), err->message);
    g_error_free (err);
    err = NULL;
  } else {
    if (value != NULL) {
			g_free (theme);
			theme = (gchar*) gconf_value_get_string (value);
    }
  }
	verify_theme (self, theme);
	g_free (theme);
	/* colour */
	colour_from_string (&self->priv->colour, LASER_CONFIG_DEFAULT_COLOUR);

  value = gconf_client_get (self->priv->gconf_client, LASER_GCONF_KEY_COLOUR, &err);
  if (err != NULL) {
    g_print (_("Error : %s\n"), err->message);
    g_error_free (err);
    err = NULL;
  } else {
    if (value != NULL) {
			colour_from_string (&self->priv->colour, (gchar*) gconf_value_get_string (value));
    }
  }
	/* restore */
	self->priv->restore = LASER_CONFIG_DEFAULT_RESTORE;
  value = gconf_client_get (self->priv->gconf_client, LASER_GCONF_KEY_RESTORE, &err);
  if (err != NULL) {
    g_print (_("Error : %s\n"), err->message);
    g_error_free (err);
    err = NULL;
  } else {
    if (value != NULL) {
      self->priv->restore = gconf_value_get_bool (value);
    }
  }
	/* arrows */
	self->priv->arrows = LASER_CONFIG_DEFAULT_ARROWS;
  value = gconf_client_get (self->priv->gconf_client, LASER_GCONF_KEY_ARROWS, &err);
  if (err != NULL) {
    g_print (_("Error : %s\n"), err->message);
    g_error_free (err);
    err = NULL;
  } else {
    if (value != NULL) {
      self->priv->arrows = gconf_value_get_bool (value);
    }
  }
	/* show_stats */
	self->priv->show_stats = LASER_CONFIG_DEFAULT_SHOW_STATS;
  value = gconf_client_get (self->priv->gconf_client, LASER_GCONF_KEY_SHOW_STATS, &err);
  if (err != NULL) {
    g_print (_("Error : %s\n"), err->message);
    g_error_free (err);
    err = NULL;
  } else {
    if (value != NULL) {
      self->priv->show_stats = gconf_value_get_bool (value);
    }
  }
	/* window size */
	self->priv->width = LASER_CONFIG_DEFAULT_WIDTH;
  value = gconf_client_get (self->priv->gconf_client, LASER_GCONF_KEY_WIDTH, &err);
  if (err != NULL) {
    g_print (_("Error : %s\n"), err->message);
    g_error_free (err);
    err = NULL;
  } else {
    if (value != NULL) {
      self->priv->width = gconf_value_get_int (value);
    }
  }
	self->priv->height = LASER_CONFIG_DEFAULT_HEIGHT;
  value = gconf_client_get (self->priv->gconf_client, LASER_GCONF_KEY_HEIGHT, &err);
  if (err != NULL) {
    g_print (_("Error : %s\n"), err->message);
    g_error_free (err);
    err = NULL;
  } else {
    if (value != NULL) {
      self->priv->height = gconf_value_get_int (value);
    }
  }
	self->priv->editor_width = LASER_CONFIG_DEFAULT_EDITOR_WIDTH;
  value = gconf_client_get (self->priv->gconf_client, LASER_GCONF_KEY_EDITOR_WIDTH, &err);
  if (err != NULL) {
    g_print (_("Error : %s\n"), err->message);
    g_error_free (err);
    err = NULL;
  } else {
    if (value != NULL) {
      self->priv->editor_width = gconf_value_get_int (value);
    }
  }
	self->priv->editor_height = LASER_CONFIG_DEFAULT_EDITOR_HEIGHT;
  value = gconf_client_get (self->priv->gconf_client, LASER_GCONF_KEY_EDITOR_HEIGHT, &err);
  if (err != NULL) {
    g_print (_("Error : %s\n"), err->message);
    g_error_free (err);
    err = NULL;
  } else {
    if (value != NULL) {
      self->priv->editor_height = gconf_value_get_int (value);
    }
  }
	/* actions */
	self->priv->actions = LASER_CONFIG_DEFAULT_ACTIONS;
  value = gconf_client_get (self->priv->gconf_client, LASER_GCONF_KEY_ACTIONS, &err);
  if (err != NULL) {
    g_print (_("Error : %s\n"), err->message);
    g_error_free (err);
    err = NULL;
  } else {
    if (value != NULL) {
      self->priv->actions = gconf_value_get_int (value);
    }
  }
	/* clock */
	self->priv->clock = LASER_CONFIG_DEFAULT_CLOCK;
  value = gconf_client_get (self->priv->gconf_client, LASER_GCONF_KEY_CLOCK, &err);
  if (err != NULL) {
    g_print (_("Error : %s\n"), err->message);
    g_error_free (err);
    err = NULL;
  } else {
    if (value != NULL) {
      self->priv->clock = gconf_value_get_int (value);
    }
  }
	/* level*/
	gchar *level_set = g_strdup (LASER_CONFIG_DEFAULT_LEVEL_SET);
	gint level = 0;
  value = gconf_client_get (self->priv->gconf_client, LASER_GCONF_KEY_LEVEL, &err);
  if (err != NULL) {
    g_print (_("Error : %s\n"), err->message);
    g_error_free (err);
    err = NULL;
  } else {
    if (value != NULL) {
      level = gconf_value_get_int (value);
    }
  }
  value = gconf_client_get (self->priv->gconf_client, LASER_GCONF_KEY_LEVEL_SET, &err);
  if (err != NULL) {
    g_print (_("Error : %s\n"), err->message);
    g_error_free (err);
    err = NULL;
  } else {
    if (value != NULL) {
			g_free (level_set);
      level_set = g_strdup (gconf_value_get_string (value));
    }
  }
	self->priv->level_set = NULL;
	verify_level (self, level_set, level);
	g_free (level_set);
	/* cheat */
	self->priv->cheat = NULL;
  value = gconf_client_get (self->priv->gconf_client, LASER_GCONF_KEY_CHEAT, &err);
  if (err != NULL) {
    g_print (_("Error : %s\n"), err->message);
    g_error_free (err);
    err = NULL;
  } else {
    if (value != NULL) {
			self->priv->cheat = g_strdup ((gchar*) gconf_value_get_string (value));
    }
  }

  gconf_client_notify_add  (self->priv->gconf_client, LASER_GCONF_PATH,
														gconf_update_cb, self, NULL, NULL);

}

static void laser_config_finalize (GObject *object) {
	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_LASER_CONFIG (object));
	LaserConfig *self = LASER_CONFIG (object);

	g_return_if_fail (self->priv != NULL);
	g_free (self->priv->theme);
	g_free (self->priv->level_set);
	G_OBJECT_CLASS (laser_config_parent_class)->finalize (object);
}

static void laser_config_class_init (LaserConfigClass *klass) {
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	object_class->finalize = laser_config_finalize;

  g_type_class_add_private (klass, sizeof (LaserConfigPrivate));

	signals [COLOUR_CHANGED] =
		g_signal_new ("colour-changed",
									G_TYPE_FROM_CLASS (object_class),
									G_SIGNAL_RUN_LAST,
									G_STRUCT_OFFSET (LaserConfigClass, laser_config_colour_changed),
									NULL,
									NULL,
									laser_marshal_VOID__POINTER,
									G_TYPE_NONE, 1, G_TYPE_POINTER);
	signals [THEME_CHANGED] =
		g_signal_new ("theme-changed",
									G_TYPE_FROM_CLASS (object_class),
									G_SIGNAL_RUN_LAST,
									G_STRUCT_OFFSET (LaserConfigClass, laser_config_theme_changed),
									NULL,
									NULL,
									laser_marshal_VOID__VOID,
									G_TYPE_NONE, 0);
	signals [RESTORE_CHANGED] =
		g_signal_new ("restore-changed",
									G_TYPE_FROM_CLASS (object_class),
									G_SIGNAL_RUN_LAST,
									G_STRUCT_OFFSET (LaserConfigClass, laser_config_restore_changed),
									NULL,
									NULL,
									laser_marshal_VOID__BOOLEAN,
									G_TYPE_NONE, 1, G_TYPE_BOOLEAN);
	signals [ARROWS_CHANGED] =
		g_signal_new ("arrows-changed",
									G_TYPE_FROM_CLASS (object_class),
									G_SIGNAL_RUN_LAST,
									G_STRUCT_OFFSET (LaserConfigClass, laser_config_arrows_changed),
									NULL,
									NULL,
									laser_marshal_VOID__BOOLEAN,
									G_TYPE_NONE, 1, G_TYPE_BOOLEAN);
	signals [SHOW_STATS_CHANGED] =
		g_signal_new ("show_stats-changed",
									G_TYPE_FROM_CLASS (object_class),
									G_SIGNAL_RUN_LAST,
									G_STRUCT_OFFSET (LaserConfigClass, laser_config_show_stats_changed),
									NULL,
									NULL,
									laser_marshal_VOID__BOOLEAN,
									G_TYPE_NONE, 1, G_TYPE_BOOLEAN);
}

LaserConfig* laser_config_new (void) {
	return g_object_new (LASER_CONFIG_TYPE, NULL);
}

void laser_config_destroy (LaserConfig *config) {
	g_object_unref (config);
}

/**
 * gconf_update_cb:
 * @client: the gconf client
 * @cnxn_id: unused
 * @entry: the entry containing the key/value pair
 * @data: unused
 *
 * Callback called when a listened gconf entry has changed, used to
 * update configuration on the fly.
 */
void gconf_update_cb (GConfClient *client,guint cnxn_id, GConfEntry *entry, gpointer data) {
	LaserConfig *self = (LaserConfig*) data;
	
	if (strcmp(entry->key, LASER_GCONF_KEY_THEME) == 0) {
		verify_theme (self, (gchar*) gconf_value_get_string (entry->value));
		g_signal_emit (self, signals [THEME_CHANGED], 0);
	}
	if (strcmp(entry->key, LASER_GCONF_KEY_COLOUR) == 0) {
		colour_from_string (&self->priv->colour, (gchar*) gconf_value_get_string (entry->value));
		g_signal_emit (self, signals [COLOUR_CHANGED], 0, self->priv->colour);
	}
	if (strcmp(entry->key, LASER_GCONF_KEY_RESTORE) == 0) {
		self->priv->restore = gconf_value_get_bool (entry->value);
		g_signal_emit (self, signals [RESTORE_CHANGED], 0, self->priv->restore);
	}
	if (strcmp(entry->key, LASER_GCONF_KEY_ARROWS) == 0) {
		self->priv->arrows = gconf_value_get_bool (entry->value);
		g_signal_emit (self, signals [ARROWS_CHANGED], 0, self->priv->arrows);
	}
	if (strcmp(entry->key, LASER_GCONF_KEY_SHOW_STATS) == 0) {
		self->priv->show_stats = gconf_value_get_bool (entry->value);
		g_signal_emit (self, signals [SHOW_STATS_CHANGED], 0, self->priv->show_stats);
	}
}

/* levels *********************************************************************/

static void verify_level (LaserConfig *self, gchar *level_set, guint level) {
	gint n_levels;
	GError *err = NULL;

	self->priv->level = level;
	self->priv->level_set = g_strdup (level_set);
	parse_level_set_check (self->priv->level_set, &err, &n_levels);
  if (err != NULL) {
		if (err != NULL) {
			g_print (_("Error : %s\n"), err->message);
			g_error_free (err);
			err = NULL;
		} else {
			g_print (_("Error : %s\n"), _("level out of bounds"));
		}
		level = LASER_CONFIG_DEFAULT_LEVEL;
		g_free (self->priv->level_set);
		self->priv->level_set = g_strdup (LASER_CONFIG_DEFAULT_LEVEL_SET);
		parse_level_set_check (self->priv->level_set, &err, &n_levels);
		if (err != NULL) {
			g_print (_("Error : Cannot get default level \"%s\" : %s\n"), self->priv->level_set,err->message);
			exit (1);
		}
		if (self->priv->level < 0 || self->priv->level >= n_levels) {
			self->priv->level = 0;
		}
  }
	if (self->priv->level < 0 || self->priv->level >= n_levels) {
		self->priv->level = 0;
	}
}

void laser_config_set_level (LaserConfig *self, gchar *set, gint level) {
	verify_level (self, set, level);
}

void laser_config_get_level (LaserConfig *self, gchar **set, gint *level) {
	*set = self->priv->level_set;
	*level = self->priv->level;
}

GSList* get_available_level_sets (void) {
	GSList *sets = NULL;

	/* construct path */
  gchar *path[] = {
		g_strdup (LASER_CONFIG_SHARE "levels"),
		g_strconcat (g_get_user_config_dir (),
								 G_DIR_SEPARATOR_S "laser" G_DIR_SEPARATOR_S "levels", NULL),
		NULL
	};

	/* search for levels */
	for (gchar **p = path;*p != NULL;p++) {
		GError *err = NULL;
		GDir *dir = g_dir_open (*p, 0, &err);
		gchar *file;
		if (err != NULL) {
			printf (_("Error : %s\n"), err->message);
			g_error_free (err);
			return NULL;
		}
		while ((file = (gchar*) g_dir_read_name (dir)) != NULL) {
			gchar *file_path = g_strconcat (*p, G_DIR_SEPARATOR_S, file, NULL);
			level_set* level_set = level_set_new ();
			if (level_set != NULL) {
				GError *err = NULL;
				GSList *name_i18n;
				gchar *name;
				GSList *note_i18n;
				gchar *note;
				parse_level_set_meta (file_path, &name_i18n, &name, &note_i18n, &note, &err);
				if (err != NULL) {
					printf (_("Error: in \"%s\" : %s\n"), file_path, err->message);
					g_error_free (err);
					g_free (file_path);
				} else {
					gpointer *tuple = g_malloc (4 * sizeof (gpointer));
					tuple[0] = file_path;
					tuple[1] = name_i18n;
					tuple[2] = name;
					tuple[3] = note_i18n;
					tuple[4] = note;
					sets = g_slist_prepend (sets, tuple);
				}
			}
		}
		g_dir_close (dir);
		g_free (*p);
	}

	if (sets == NULL) {
		printf ("Error: no level set\n");
		return NULL;
	}

	return g_slist_reverse (sets);
}

void free_available_level_sets (GSList *sets) {
  for (GSList *l = sets;l != NULL;l = g_slist_next(l)) {
		gpointer *tuple = l->data;
		g_free (tuple[0]);
		free_i18n ((GSList*) tuple[1]);
		free_i18n ((GSList*) tuple[3]);
	}
	g_slist_free (sets);
}

GSList* get_set_names (gchar *uri) {
	GError *err = NULL;
	GSList *names;
	parse_level_set_levels  (uri, &names, &err);
	if (err != NULL) {
		printf (_("Error: in \"%s\" : %s\n"), uri, err->message);
		g_error_free (err);
		return NULL;
	}
	return names;
}

void free_set_names (GSList* names) {
  for (GSList *l = names;l != NULL;l = g_slist_next(l)) {
		gpointer *pair = l->data;
		free_i18n ((GSList*) pair[0]);
		g_free (pair[1]);
	}
	g_slist_free (names);
}

/* themes *********************************************************************/

gchar* laser_config_get_theme (LaserConfig *self) {
	return g_strdup (self->priv->theme);
}

GSList* get_available_themes (void) {
	GSList *themes = NULL;

	/* construct path */
  gchar *path[] = {
		g_strdup (LASER_CONFIG_SHARE "themes"),
		g_strconcat (g_get_user_config_dir (),
								 G_DIR_SEPARATOR_S "laser" G_DIR_SEPARATOR_S "themes", NULL),
		NULL
	};
	/* search for themes */
	for (gchar **p = path;*p != NULL;p++) {
		GError *err = NULL;
		GDir *dir = g_dir_open (*p, 0, &err);
		gchar *file;
		if (err != NULL) {
			printf ("Error: %s\n", err->message);
			g_error_free (err);
			return NULL;
		}
		while ((file = (gchar*) g_dir_read_name (dir)) != NULL) {
			/* get translation from i18n file */
			gchar *i18n = g_strconcat (*p, G_DIR_SEPARATOR_S, file, "/i18n", NULL);
			FILE *fp = fopen (i18n, "r");
			g_free (i18n);
			if (fp != NULL) {
				const gchar* const * langs = g_get_language_names ();
				gchar name[1001] = ""; /* truncate name to 1000 chars */
				gboolean found = FALSE;
				int compatibility_level = 10000;
				while (!feof (fp)) {
					/* get locate (first part of the line, before ' ') */
					gchar locale [21];
					gchar c = fgetc (fp);
					int i = 0;
					while (i < 20 && c != ' ' && !feof (fp)) {
						locale[i++] = c;
						c = fgetc (fp);
					}
					locale[i] = '\0';
					/* check if the locale is compatible with the lang */
					for (int i = 0; langs[i] != NULL; i++) {
						/* update the name if locale is more compatible */
						if (strcmp (langs[i], locale) == 0 && (!found || i < compatibility_level)) {
							found = TRUE;
							gchar c = fgetc (fp);
							int j = 0;
							while (j < 1000 && c != '\n' && !feof (fp)) {
								name[j++] = c;
								c = fgetc (fp);
							}
							name[j] = '\0';
							compatibility_level = i;
							break;
						}
					}
				}
				/* if translation found, add the theme */
				if (found) {
					gchar **pair = g_malloc (2 * sizeof (gchar*));
					pair[0] = g_strconcat (*p, G_DIR_SEPARATOR_S, file, NULL);
					pair[1] = g_strdup (name);
					themes = g_slist_prepend (themes, pair);
				}
			}
		}
		g_dir_close (dir);
		g_free (*p);
	}

	if (themes == NULL) {
		printf ("Error: no theme\n");
	}
	return g_slist_reverse (themes);
}

void free_available_themes (GSList *themes) {
  for (GSList *l = themes;l != NULL;l = g_slist_next(l)) {
		gpointer *pair = l->data;
		g_free (pair[0]);
		g_free (pair[1]);
	}
	g_slist_free (themes);
}

static void verify_theme (LaserConfig *self, gchar *theme) {
	gboolean not_default;
	if ((not_default = g_file_test (theme, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR))) {
		self->priv->theme = g_strdup (theme);
	} else {
		self->priv->theme = g_strdup (LASER_CONFIG_DEFAULT_THEME);
	}
	/* clear old caches */
	for (int i = 0; i < CACHE_SIZE; i++) {
		if (self->priv->svg_hnd[i] != NULL) {
			rsvg_handle_free (self->priv->svg_hnd[i]);
			self->priv->svg_hnd[i] = NULL;
		}
	}
  for (GSList *l = self->priv->caches;l != NULL;l = g_slist_next(l)) {
		theme_cache *theme_cache = l->data;
		for (int i = 0; i < CACHE_SIZE; i++) {
			for (int d = 0;d < 8;d++) {
				if (theme_cache->surface[i][d] != NULL) {
					cairo_surface_destroy (theme_cache->surface[i][d]);
					theme_cache->surface[i][d] = NULL;
				}
			}
		}
	}

	/* check for required svgs and meta files */
	for (int i = 0; i < CACHE_SIZE; i++) {
		gchar *file = g_strconcat (self->priv->theme, G_DIR_SEPARATOR_S, theme_cache_static[i].file, NULL);
		GError *err = NULL;
		self->priv->svg_hnd[i] = rsvg_handle_new_from_file (file, &err);
		if (err != NULL) {
			g_print (_("Error loading theme \"%s\" : cannot open file \"%s\"\n"), theme, file);
			if (!not_default) {
				exit (1);
			}
			/* revert to default */
			for (int j = 0; j < i; j++) {
				rsvg_handle_free (self->priv->svg_hnd[j]);
			}
			verify_theme (self, LASER_CONFIG_DEFAULT_THEME);
			return;
		}
		g_free (file);
	}

	/* clear caches */
  for (GSList *l = self->priv->caches;l != NULL;l = g_slist_next(l)) {
		theme_cache *cache = l->data;
		for (int i = 0; i < CACHE_SIZE; i++) {
			for (int d = 0;d < 8;d++) {
				if (cache->surface[i][d] != NULL) {
					cairo_surface_destroy (cache->surface[i][d]);
					cache->surface[i][d] = NULL;
				}
			}
		}
	}

	/* set stock */
	gtk_stock_add ((GtkStockItem[]){{ "laser-beamer-red", _("Red beamer"), 0, 0, NULL}}, 1);

	GtkIconFactory *factory = gtk_icon_factory_new ();
	GtkIconSet *set = gtk_icon_set_new  ();
	GtkIconSource *source = gtk_icon_source_new ();
	gtk_icon_source_set_filename (source,
																g_strconcat (self->priv->theme, G_DIR_SEPARATOR_S, "beamer_red.svg", NULL));
	gtk_icon_set_add_source (set, source);
	//	gtk_icon_source_free(source);
	gtk_icon_factory_add (factory, "laser-beamer-red", set);
	gtk_icon_factory_add_default (factory);

	gtk_icon_theme_prepend_search_path (gtk_icon_theme_get_default (), self->priv->theme);
	gtk_icon_theme_rescan_if_needed (gtk_icon_theme_get_default ());
}

theme_cache* sprite_cache_new (LaserConfig *self, guint size) {
	theme_cache *cache = g_malloc (sizeof (struct theme_cache));
	cache->size = size;
	cache->config = self;
	
	for (int i = 0; i < CACHE_SIZE; i++) {
		for (int d = 0;d < 8;d++) {
			cache->surface[i][d] = NULL;
		}
	}
	self->priv->caches = g_slist_prepend (self->priv->caches, cache);
	return cache;
}

void sprite_cache_set (theme_cache *cache, guint size) {
	GTimer *test = g_timer_new ();
	if (size != cache->size) {
		cache->size = size;
		for (int i = 0; i < CACHE_SIZE; i++) {
			for (int d = 0;d < 8;d++) {
				if (cache->surface[i][d] != NULL) {
					cairo_surface_destroy (cache->surface[i][d]);
					cache->surface[i][d] = NULL;
				}
			}
		}
	}
	if (g_timer_elapsed (test, NULL) > 0.05) {
		printf ("TIME : %g\n", g_timer_elapsed (test, NULL));
	}
}

cairo_surface_t* sprite_cache_get (theme_cache *cache, item *item) {
	int i = 0, d = 0;
	/* special background case */
	for (int j = 0; j < CACHE_SIZE; j++) {
		if (theme_cache_static[j].type == item->type
				&& (theme_cache_static[j].colour == item->colour
						|| theme_cache_static[j].colour == NONE)) {
			if (! EDIT_ROTATION (item->type)) {
				i = j; d = 0; break;
			}
			i = j; d = item->direction; break;
		}
	}
	if (cache->surface[i][d] == NULL) {
		cache->surface[i][d] =
			cairo_image_surface_create (CAIRO_FORMAT_ARGB32, cache->size, cache->size);
		if (cache->surface[i][d] == NULL) {
			return NULL;
		}
		
		cairo_t *cr_surf = cairo_create (cache->surface[i][d]);
		if (cr_surf == NULL) {
			return NULL;
		}
		RsvgDimensionData dim;
		rsvg_handle_get_dimensions(cache->config->priv->svg_hnd[i], &dim);
		cairo_scale (cr_surf, cache->size / (double) dim.width, cache->size / (double) dim.height);
		cairo_translate (cr_surf, (double) dim.width / 2, (double) dim.height / 2);
		cairo_rotate (cr_surf, (double) d * G_PI / 4);
		cairo_translate (cr_surf, (double) - dim.width / 2, (double) - dim.height / 2);
		rsvg_handle_render_cairo (cache->config->priv->svg_hnd[i], cr_surf);
		cairo_destroy (cr_surf);
	}
	return cache->surface[i][d];
}

/* history *******************************************************************/

gchar* laser_config_get_cheat (LaserConfig *self) {
	return g_strdup (self->priv->cheat);
}

void laser_config_set_cheat (LaserConfig *self, gchar *cheat) {
	if (self->priv->cheat != NULL) {
		g_free (self->priv->cheat);
	}
	self->priv->cheat = g_strdup (cheat);
}

void laser_config_save_history (LaserConfig *self) {
	/* saves to gconf keys */
	GError *err = NULL;
	GConfValue *value = gconf_value_new (GCONF_VALUE_STRING);
	if (value == NULL) {
		return;
	}
	gconf_value_set_string (value, self->priv->cheat);
	gconf_client_set (self->priv->gconf_client, LASER_GCONF_KEY_CHEAT, value, &err);
	if (err != NULL) {
		printf ("Error : %s\n", err->message);
		g_error_free (err);
		err = NULL;
	}
	
	gconf_value_set_string (value, self->priv->level_set);
	gconf_client_set (self->priv->gconf_client, LASER_GCONF_KEY_LEVEL_SET, value, &err);
	if (err != NULL) {
		printf ("Error : %s\n", err->message);
		g_error_free (err);
		err = NULL;
	}
	gconf_value_free (value);
	
	value = gconf_value_new (GCONF_VALUE_INT);
	if (value == NULL) {
		return;
	}
	gconf_value_set_int (value, self->priv->level);
	gconf_client_set (self->priv->gconf_client, LASER_GCONF_KEY_LEVEL, value, &err);
	if (err != NULL) {
		printf ("Error : %s\n", err->message);
		g_error_free (err);
		err = NULL;
	}
	gconf_value_set_int (value, self->priv->actions);
	gconf_client_set (self->priv->gconf_client, LASER_GCONF_KEY_ACTIONS, value, &err);
	if (err != NULL) {
		printf ("Error : %s\n", err->message);
		g_error_free (err);
		err = NULL;
	}
	gconf_value_set_int (value, self->priv->clock);
	gconf_client_set (self->priv->gconf_client, LASER_GCONF_KEY_CLOCK, value, &err);
	if (err != NULL) {
		printf ("Error : %s\n", err->message);
		g_error_free (err);
		err = NULL;
	}
	gconf_value_free (value);
}

/* getters and setters ********************************************************/

void laser_config_set_width (LaserConfig *self, gint width) {
	self->priv->width = width;
	GError *err = NULL;
	GConfValue *value = gconf_value_new (GCONF_VALUE_INT);
	gconf_value_set_int (value, self->priv->width);
	gconf_client_set (self->priv->gconf_client, LASER_GCONF_KEY_WIDTH, value, &err);
	if (err != NULL) {
		printf ("Error : %s\n", err->message);
		g_error_free (err);
		err = NULL;
	}
	gconf_value_free (value);
}

gint laser_config_get_width (LaserConfig *self) {
	return self->priv->width;
}

void laser_config_set_height (LaserConfig *self, gint height) {
	self->priv->height = height;
	GError *err = NULL;
	GConfValue *value = gconf_value_new (GCONF_VALUE_INT);
	gconf_value_set_int (value, self->priv->height);
	gconf_client_set (self->priv->gconf_client, LASER_GCONF_KEY_HEIGHT, value, &err);
	if (err != NULL) {
		printf ("Error : %s\n", err->message);
		g_error_free (err);
		err = NULL;
	}
	gconf_value_free (value);
}

gint laser_config_get_height (LaserConfig *self) {
	return self->priv->height;
}

void laser_config_set_editor_width (LaserConfig *self, gint editor_width) {
	self->priv->editor_width = editor_width;
	GError *err = NULL;
	GConfValue *value = gconf_value_new (GCONF_VALUE_INT);
	gconf_value_set_int (value, self->priv->editor_width);
	gconf_client_set (self->priv->gconf_client, LASER_GCONF_KEY_EDITOR_WIDTH, value, &err);
	if (err != NULL) {
		printf ("Error : %s\n", err->message);
		g_error_free (err);
		err = NULL;
	}
	gconf_value_free (value);
}

gint laser_config_get_editor_width (LaserConfig *self) {
	return self->priv->editor_width;
}

void laser_config_set_editor_height (LaserConfig *self, gint editor_height) {
	self->priv->editor_height = editor_height;
	GError *err = NULL;
	GConfValue *value = gconf_value_new (GCONF_VALUE_INT);
	gconf_value_set_int (value, self->priv->editor_height);
	gconf_client_set (self->priv->gconf_client, LASER_GCONF_KEY_EDITOR_HEIGHT, value, &err);
	if (err != NULL) {
		printf ("Error : %s\n", err->message);
		g_error_free (err);
		err = NULL;
	}
	gconf_value_free (value);
}

gint laser_config_get_editor_height (LaserConfig *self) {
	return self->priv->editor_height;
}

void laser_config_set_clock (LaserConfig *self, gint clock) {
	self->priv->clock = clock;
}

gint laser_config_get_clock (LaserConfig *self) {
	return self->priv->clock;
}

void laser_config_set_actions (LaserConfig *self, gint actions) {
	self->priv->actions = actions;
}

gint laser_config_get_actions (LaserConfig *self) {
	return self->priv->actions;
}

void laser_config_get_colour (LaserConfig *self, colour_t *colour) {
	memcpy (colour, &self->priv->colour, sizeof (struct colour_t));
}

void laser_config_set_colour (LaserConfig *self, colour_t *colour) {
	memcpy (&self->priv->colour, colour, sizeof (struct colour_t));

	GError *err = NULL;
	GConfValue *value = gconf_value_new (GCONF_VALUE_STRING);
	if (value == NULL) {
		return;
	}
	gconf_value_set_string (value, colour->str);
	gconf_client_set (self->priv->gconf_client, LASER_GCONF_KEY_COLOUR, value, &err);
	if (err != NULL) {
		printf ("Error : %s\n", err->message);
		g_error_free (err);
	}
	gconf_value_free (value);
}

void laser_config_colour_from_id(guchar id, colour_t *colour) {
	static struct {
		guchar colour;
		gchar str[10];
	} assoc[] = {
		{ NONE,          "#000000FF" },
		{ COLOUR_RED,    "#FF0000FF" },
		{ COLOUR_GREEN,  "#00FF00FF" },
		{ COLOUR_BLUE,   "#0000FFFF" },
		{ COLOUR_VIOLET, "#FF00FFFF" },
		{ COLOUR_YELLOW, "#FFFF00FF" },
		{ COLOUR_WHITE,  "#FFFFFFFF" },
		{ COLOUR_CYAN,   "#00FFFFFF" }
	};

	colour_from_string (colour, assoc[id].str);
}

void laser_config_set_theme (LaserConfig *self, gchar *theme) {
	GError *err = NULL;
	GConfValue *value = gconf_value_new (GCONF_VALUE_STRING);
	if (value == NULL) {
		return;
	}
	gconf_value_set_string (value, theme);
	gconf_client_set (self->priv->gconf_client, LASER_GCONF_KEY_THEME, value, &err);
	if (err != NULL) {
		printf ("Error : %s\n", err->message);
		g_error_free (err);
	}
	gconf_value_free (value);
}

void laser_config_set_restore (LaserConfig *self, gboolean restore) {
	GError *err = NULL;
	GConfValue *value = gconf_value_new (GCONF_VALUE_BOOL);
	if (value == NULL) {
		return;
	}
	gconf_value_set_bool (value, restore);
	gconf_client_set (self->priv->gconf_client, LASER_GCONF_KEY_RESTORE, value, &err);
	if (err != NULL) {
		printf ("Error : %s\n", err->message);
		g_error_free (err);
	}
	gconf_value_free (value);
}

gboolean laser_config_get_restore (LaserConfig *self) {
	return self->priv->restore;
}

void laser_config_set_arrows (LaserConfig *self, gboolean arrows) {
	GError *err = NULL;
	GConfValue *value = gconf_value_new (GCONF_VALUE_BOOL);
	if (value == NULL) {
		return;
	}
	gconf_value_set_bool (value, arrows);
	gconf_client_set (self->priv->gconf_client, LASER_GCONF_KEY_ARROWS, value, &err);
	if (err != NULL) {
		printf ("Error : %s\n", err->message);
		g_error_free (err);
	}
	gconf_value_free (value);
}

gboolean laser_config_get_arrows (LaserConfig *self) {
	return self->priv->arrows;
}

void laser_config_set_show_stats (LaserConfig *self, gboolean show_stats) {
	GError *err = NULL;
	GConfValue *value = gconf_value_new (GCONF_VALUE_BOOL);
	if (value == NULL) {
		return;
	}
	gconf_value_set_bool (value, show_stats);
	gconf_client_set (self->priv->gconf_client, LASER_GCONF_KEY_SHOW_STATS, value, &err);
	if (err != NULL) {
		printf ("Error : %s\n", err->message);
		g_error_free (err);
	}
	gconf_value_free (value);
}

gboolean laser_config_get_show_stats (LaserConfig *self) {
	return self->priv->show_stats;
}
