/* -*- Mode: C; tab-width: 2; indent-tabs-mode: t; c-basic-offset: 2 -*-
 *
 * LASER
 * Copyright (C) 2006 Benjamin Canou <bookeldor@gmail.com>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "config.h"

#include <gtk/gtk.h>
#include <glade/glade.h>
#include <gconf/gconf-client.h>
#include <libgnomeui/libgnomeui.h>
#include <libgnomeui/gnome-help.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib/gi18n.h>
#include <librsvg/rsvg.h>

#include "laser-common.h"
#include "laser-config.h"
#include "laser-parser.h"
#include "laser-writer.h"
#include "laser-board.h"

/* datatypes and macros *******************************************************/

struct tool;
struct tab_data;
typedef struct editor_instance {
	LaserConfig *config;
	gboolean modified;
	board_t *board;
	gboolean supports_alpha;
	GTimer *timer_clock;
	gint timer_offset;
	gboolean loaded;
	level_set* level_set;
	gint level;
	GladeXML *ui;
	GladeXML    *win_main, *win_prefs;
	struct tab_data *tab_translations, *tab_level_translations;
	struct tool *tool;
} editor_instance;

struct tab_data {
	editor_instance   *editor;
	GtkWidget         *tab;
	GSList            **titles;
	GSList            **comments;
	GtkTreeStore      *treestore;
};

struct tool {
	gchar*   name;
	guchar   type;
	guchar   colour;
};

/* ui shortcut macros */
#define GET_WINDOW(w)                GTK_WINDOW (glade_xml_get_widget (editor->win_##w, "window"))
#define GET_DIALOG(w)                GTK_DIALOG (glade_xml_get_widget (editor->win_##w, "dialog"))
#define GET_WIDGET(w, n)             glade_xml_get_widget (editor->win_##w, #n)
#define GET_WIDGET_S(w, n)           glade_xml_get_widget (editor->win_##w, n)
#define GET_LABEL(w, n)              GTK_LABEL (glade_xml_get_widget (editor->win_##w, "label_" #n))
#define GET_BUTTON(w, n)             GTK_BUTTON (glade_xml_get_widget (editor->win_##w, "button_" #n))
#define GET_COMBO(w, n)              GTK_COMBO_BOX (glade_xml_get_widget (editor->win_##w, "combo_" #n))
#define GET_CHECK(w, n)              GTK_CHECK_BUTTON (glade_xml_get_widget (editor->win_##w, "check_" #n))
#define GET_ENTRY(w, n)              GTK_ENTRY (glade_xml_get_widget (editor->win_##w, "entry_" #n))
#define CONNECT(w, n, s, cb)         g_signal_connect (GET_WIDGET (w,n), s, G_CALLBACK (cb), editor)
#define CONNECT_DATA(w, n, s, cb, d) g_signal_connect (GET_WIDGET (w,n), s, G_CALLBACK (cb), d)
#define SET_SENSITIVE(w, n, v)       gtk_widget_set_sensitive (glade_xml_get_widget (editor->win_##w, #n), v)
#define SET_CHECK_ACTIVE(w, n, v)    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (GET_WIDGET (w,check_##n)), v)
#define GET_CHECK_ACTIVE(w, n)       gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (GET_WIDGET (w,check_##n)))

/* globals ********************************************************************/

static struct tool static_tools [] = {
	{ "beamer_red",       MAP_BEAMER,      COLOUR_RED },
	{ "beamer_green",     MAP_BEAMER,      COLOUR_GREEN },
	{ "beamer_blue",      MAP_BEAMER,      COLOUR_BLUE },
	{ "beamer_yellow",    MAP_BEAMER,      COLOUR_YELLOW },
	{ "beamer_violet",    MAP_BEAMER,      COLOUR_VIOLET },
	{ "beamer_cyan",      MAP_BEAMER,      COLOUR_CYAN },
	{ "beamer_white",     MAP_BEAMER,      COLOUR_WHITE },
	{ "target_red",       MAP_TARGET,      COLOUR_RED },
	{ "target_green",     MAP_TARGET,      COLOUR_GREEN },
	{ "target_blue",      MAP_TARGET,      COLOUR_BLUE },
	{ "target_yellow",    MAP_TARGET,      COLOUR_YELLOW },
	{ "target_violet",    MAP_TARGET,      COLOUR_VIOLET },
	{ "target_cyan",      MAP_TARGET,      COLOUR_CYAN },
	{ "target_white",     MAP_TARGET,      COLOUR_WHITE },
	{ "hider",            MAP_HIDER,       NONE },
	{ "diode",            MAP_DIODE,       NONE },
	{ "wall",             MAP_WALL,        NONE },
	{ "cross",            MAT_CROSS,       NONE },
	{ "mirror",           MAT_MIR,         NONE },
	{ "duplicator",       MAT_DUP,         NONE },
	{ "tee",              MAT_TEE,         NONE },
	{ "muxer",            MAT_MUX,         NONE },
	{ NULL, 0, 0}
};

static GnomeProgram *program;

/* translation tables *********************************************************/

static void tab_locale_setter (GtkCellRendererText *renderer, gchar *path,	gchar *new_text, gpointer data) {
	struct tab_data *tab_data = data;
	GtkTreeIter iter;
	GtkTreeIter toplevel;
	gchar *locale;
	gtk_tree_model_get_iter_from_string (GTK_TREE_MODEL (tab_data->treestore), &iter, path);
	gtk_tree_model_get (GTK_TREE_MODEL (tab_data->treestore), &iter, 0, &locale, -1);
	if (strcmp (new_text, "") != 0) {
		if (strcmp (locale, new_text) != 0) {
			gboolean found = FALSE;
			for (GSList *l = *tab_data->titles;l != NULL;l = g_slist_next (l)) {
				gchar **pair = (gchar**)l->data;
				if (strcmp (new_text, pair[0]) == 0) {
					found = TRUE;
				}
			}
			for (GSList *l = *tab_data->comments;l != NULL;l = g_slist_next (l)) {
				gchar **pair = (gchar**)l->data;
				if (strcmp (new_text, pair[0]) == 0) {
					found = TRUE;
				}
			}
			if (!found) {
				gtk_tree_store_set (tab_data->treestore, &iter, 0, new_text, -1);
				
				if (strcmp (locale, "") == 0) {
					gtk_tree_store_prepend(tab_data->treestore, &toplevel, NULL);
					gtk_tree_store_set(tab_data->treestore, &toplevel, 0, locale, 1, "", 2, "", -1);
					gchar **pair = malloc (2 * sizeof (gchar*));
					pair[0] = g_strdup (new_text);
					pair[1] = g_strdup ("");
					*tab_data->titles = g_slist_prepend (*tab_data->titles, pair);
					pair = malloc (2 * sizeof (gchar*));
					pair[0] = g_strdup (new_text);
					pair[1] = g_strdup ("");
					*tab_data->comments = g_slist_prepend (*tab_data->comments, pair);
				} else {
					for (GSList *l = *tab_data->titles;l != NULL;l = g_slist_next (l)) {
						gchar **pair = (gchar**)l->data;
						if (strcmp (locale, pair[0]) == 0) {
							pair[0] = g_strdup (new_text);
						}
					}
					for (GSList *l = *tab_data->comments;l != NULL;l = g_slist_next (l)) {
						gchar **pair = (gchar**)l->data;
						if (strcmp (locale, pair[0]) == 0) {
							pair[0] = g_strdup (new_text);
						}
					}
				}
								tab_data->editor->modified = TRUE;		
			} else {
				printf ("locale already defined\n");
			}
		}
	} else {
		if (strcmp (locale, "") != 0 && strcmp (locale, "C") != 0) {
			gtk_tree_store_remove (tab_data->treestore, &iter);
			for (GSList *l = *tab_data->titles;l != NULL;l = g_slist_next (l)) {
				gchar **pair = (gchar**)l->data;
				if (strcmp (locale, pair[0]) == 0) {
					*tab_data->titles = g_slist_remove_link (*tab_data->titles, l);
					break;
				}
			}
			for (GSList *l = *tab_data->comments;l != NULL;l = g_slist_next (l)) {
				gchar **pair = (gchar**)l->data;
				if (strcmp (locale, pair[0]) == 0) {
					*tab_data->comments = g_slist_remove_link (*tab_data->comments, l);
					break;
				}
			}
			tab_data->editor->modified = TRUE;
		}
	}
}


static void tab_title_setter (GtkCellRendererText *renderer, gchar *path,	gchar *new_text, gpointer data) {
	if (strcmp (new_text, "") != 0) {
		struct tab_data *tab_data = data;
		GtkTreeIter iter;
		gchar *locale;
		gtk_tree_model_get_iter_from_string (GTK_TREE_MODEL (tab_data->treestore), &iter, path);
		gtk_tree_model_get (GTK_TREE_MODEL (tab_data->treestore), &iter, 0, &locale, -1);
		gtk_tree_store_set (tab_data->treestore, &iter, 1, new_text, -1);
		if (strcmp (locale, "") == 0) {
			printf ("define locale first!\n");
		} else {
			for (GSList *l = *tab_data->titles;l != NULL;l = g_slist_next (l)) {
				gchar **pair = (gchar**)l->data;
				if (strcmp (locale, pair[0]) == 0) {
					pair[1] = g_strdup (new_text);
				}
			}
			tab_data->editor->modified = TRUE;
		}
	}
}

static void tab_comment_setter (GtkCellRendererText *renderer, gchar *path,	gchar *new_text, gpointer data) {
	if (strcmp (new_text, "") != 0) {
		struct tab_data *tab_data = data;
		GtkTreeIter iter;
		gchar *locale;
		gtk_tree_model_get_iter_from_string (GTK_TREE_MODEL (tab_data->treestore), &iter, path);
		gtk_tree_model_get (GTK_TREE_MODEL (tab_data->treestore), &iter, 0, &locale, -1);
		gtk_tree_store_set (tab_data->treestore, &iter, 2, new_text, -1);
		if (strcmp (locale, "") == 0) {
			printf ("define locale first!\n");
		} else {
			for (GSList *l = *tab_data->comments;l != NULL;l = g_slist_next (l)) {
				gchar **pair = (gchar**)l->data;
				if (strcmp (locale, pair[0]) == 0) {
					pair[1] = g_strdup (new_text);
				}
			}
			tab_data->editor->modified = TRUE;
		}
	}
}

static struct tab_data* init_tab (editor_instance *editor, GtkWidget *tab) {
	struct tab_data *tab_data = g_malloc (sizeof (struct tab_data));
	tab_data->editor = editor;
	tab_data->tab = tab;

	GtkTreeViewColumn   *col;
  GtkCellRenderer     *renderer;

	col = gtk_tree_view_column_new();
  gtk_tree_view_column_set_title(col, _("Locale"));
	gtk_tree_view_append_column(GTK_TREE_VIEW(tab), col);
	renderer = gtk_cell_renderer_text_new();
	g_object_set(renderer, "editable", TRUE, NULL);
	gtk_tree_view_column_pack_start(col, renderer, TRUE);
	gtk_tree_view_column_add_attribute(col, renderer, "text", 0);
	g_signal_connect (renderer, "edited", G_CALLBACK (tab_locale_setter), tab_data);

	col = gtk_tree_view_column_new();
  gtk_tree_view_column_set_title(col, _("Title"));
	gtk_tree_view_append_column(GTK_TREE_VIEW(tab), col);
	renderer = gtk_cell_renderer_text_new();
	g_object_set(renderer, "editable", TRUE, NULL);
	gtk_tree_view_column_pack_start(col, renderer, TRUE);
	gtk_tree_view_column_add_attribute(col, renderer, "text", 1);
	g_signal_connect (renderer, "edited", G_CALLBACK (tab_title_setter), tab_data);

	col = gtk_tree_view_column_new();
  gtk_tree_view_column_set_title(col, _("Comment"));
	gtk_tree_view_append_column(GTK_TREE_VIEW(tab), col);
	renderer = gtk_cell_renderer_text_new();
	g_object_set(renderer, "editable", TRUE, NULL);
	gtk_tree_view_column_pack_start(col, renderer, TRUE);
	gtk_tree_view_column_add_attribute(col, renderer, "text", 2);
	g_signal_connect (renderer, "edited", G_CALLBACK (tab_comment_setter), tab_data);

	return tab_data;
}

void fill_tab (struct tab_data *tab_data, GSList **titles, GSList **comments) {
	GtkTreeStore  *treestore;
  GtkTreeIter    toplevel;

  treestore = gtk_tree_store_new(3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);

	for (GSList *l = *titles;l != NULL;l = g_slist_next (l)) {
		gchar **pair = (gchar**)l->data;
		gchar *com = NULL;
		for (GSList *l = *comments;l != NULL;l = g_slist_next (l)) {
			gchar **pair_com = (gchar**)l->data;
			if (strcmp(pair_com[0], pair[0]) == 0) {
				com = pair_com[1];
			}
		}
		if (com == NULL) {
			gchar **npair = g_malloc (2 * sizeof (gchar*));
			npair[0] = g_strdup (pair[0]);
			npair[1] = g_strdup ("");
			com = npair[1];
			*comments = g_slist_prepend (*comments, npair);
		}
		gtk_tree_store_append(treestore, &toplevel, NULL);
		gtk_tree_store_set(treestore, &toplevel, 0, pair[0], 1, pair[1], 2, com, -1);
	}

	gtk_tree_store_prepend(treestore, &toplevel, NULL);
	gtk_tree_store_set(treestore, &toplevel, 0, "", 1, "", 2, "", -1);
			
	gtk_tree_view_set_model(GTK_TREE_VIEW(tab_data->tab), GTK_TREE_MODEL(treestore));
	tab_data->treestore = treestore;
	tab_data->titles = titles;
	tab_data->comments = comments;
}

/******************************************************************************/

static void update_states (gpointer data, gint total_targets, gint reached_targets,
													 gint actions, gboolean undo, gboolean redo) {
	editor_instance *editor = data;

	SET_SENSITIVE (main, button_undo, undo);
	SET_SENSITIVE (main, button_redo, redo);
}

/**
 * screen_changed:
 * @widget: the migrated widget
 * @old_screen: unused
 * @data: unused
 *
 * Checks if alpha is supported at each migration.
 */
static void screen_changed (GtkWidget *widget, GdkScreen *old_screen, gpointer data) {
	GdkScreen *screen = gtk_widget_get_screen(widget);
	GdkColormap *colormap = gdk_screen_get_rgba_colormap(screen);
	editor_instance *editor = data;

	/* supports alpha is able to get an rgba color map */
	if (!colormap) {
		colormap = gdk_screen_get_rgb_colormap(screen);
		printf ("Your screen supports alpha channel, translucent background enabled.\n");
		editor->supports_alpha = FALSE;
	} else {
		editor->supports_alpha = TRUE;
	}

	/* sets color map accordingly */
	gtk_widget_set_colormap(widget, colormap);
}



/**
 * load_level:
 * @set: the set folder
 * @level: the level index in the set
 *
 * Loads the given level and forces repaint.
 */
void load_level (editor_instance *editor, gchar *set, gint level) {
	if ((!editor->loaded) || strcmp (editor->level_set->file, set) != 0) {
		if (editor->loaded) {
			level_set_free (editor->level_set);
		}
		SET_SENSITIVE (main, button_add_after, TRUE);
		SET_SENSITIVE (main, button_add_before, TRUE);
		SET_SENSITIVE (main, edit_level_difficulty, TRUE);

		editor->modified = FALSE;
		editor->level_set = level_set_new ();
		GError *err = NULL;
		parse_level_set (set, editor->level_set, &err);
		if (err != NULL) {
			printf (_("Error : \"%s\"\n"), err->message);
			g_error_free (err);
		}
		fill_tab (editor->tab_translations, &editor->level_set->name_i18n, &editor->level_set->note_i18n);
	}
	editor->level = level;
	laser_board_set_level (editor->board, &editor->level_set->levels [editor->level]);
	if (editor->level_set->levels[editor->level].solution != NULL) {
		laser_board_set_cheat (editor->board, editor->level_set->levels[editor->level].solution);
	}
	fill_tab (editor->tab_level_translations,
						&editor->level_set->levels [editor->level].name_i18n,
						&editor->level_set->levels [editor->level].comment_i18n);
	gtk_range_set_value (GTK_RANGE (GET_WIDGET (main, edit_level_difficulty)), editor->level_set->levels[editor->level].difficulty);
	SET_SENSITIVE (main, button_previous, editor->level > 0);
	SET_SENSITIVE (main, button_next, editor->level < editor->level_set->nb_levels - 1);
	SET_SENSITIVE (main, button_delete_level, editor->level_set->nb_levels > 1);
	SET_SENSITIVE (main, button_expand_top, TRUE);
	SET_SENSITIVE (main, button_expand_bottom, TRUE);
	SET_SENSITIVE (main, button_expand_left, TRUE);
	SET_SENSITIVE (main, button_expand_right, TRUE);
	SET_SENSITIVE (main, button_shrink_top, editor->level_set->levels[editor->level].height > 2);
	SET_SENSITIVE (main, button_shrink_bottom, editor->level_set->levels[editor->level].height > 2);
	SET_SENSITIVE (main, button_shrink_left, editor->level_set->levels[editor->level].width > 2);
	SET_SENSITIVE (main, button_shrink_right, editor->level_set->levels[editor->level].width > 2);
	SET_SENSITIVE (main, button_save_solution, TRUE);
	editor->loaded = TRUE;
}


static void colour_changed_cb (LaserConfig *laser_config, gdouble r, gdouble g, gdouble b, gdouble a, gpointer data) {
	editor_instance *editor = data;

	laser_board_paint (editor->board);
}

static void arrows_changed_cb (LaserConfig *laser_config, gboolean arrows, gpointer data) {
	editor_instance *editor = data;

	laser_board_paint (editor->board);
}


static void theme_changed_cb (LaserConfig *laser_config, gpointer data) {
	editor_instance *editor = data;

	laser_board_paint (editor->board);
}

/**
 * dialog_about_cb:
 *
 * displays about dialog
 **/
static void dialog_about_cb (GtkMenuItem *menuitem, gpointer data) {
	editor_instance *editor = data;

	GtkAboutDialog *about;

	GdkPixbuf *logo =
		gtk_icon_theme_load_icon (gtk_icon_theme_get_default (),
					  "laser-editor",
					  128, 0, NULL);

	const gchar *authors[] = {
		_("Programming"),
		" - Benjamin Canou <bookeldor@gmail.com>",
		"",
		_("Levels"),
		" - Benjamin Canou <bookeldor@gmail.com>",
		"",
		_("Distribution"),
		" - Ubuntu GNU/Linux :",
		"   David Valot <ptitdav69@gmail.com>",
		NULL
	};
	const char *documenters [] = {
		" - Benjamin Canou <bookeldor@gmail.com>",
		NULL
	};
	const char *license[] = {
		 N_("Licensed under the GNU General Public License Version 2"),
		 N_("Laser is free software; you can redistribute it and/or\n"
		   "modify it under the terms of the GNU General Public License\n"
		   "as published by the Free Software Foundation; either version 2\n"
		   "of the License, or (at your option) any later version."),
		 N_("Laser is distributed in the hope that it will be useful,\n"
		   "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
		   "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
		   "GNU General Public License for more details."),
		 N_("You should have received a copy of the GNU General Public License\n"
		   "along with this program; if not, write to the Free Software\n"
		   "Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA\n"
		   "02110-1301, USA.")
	};
	const char *translator_credits = NULL;
	char	   *license_trans;

	license_trans = g_strconcat (_(license[0]), "\n\n", _(license[1]), "\n\n",
				     _(license[2]), "\n\n", _(license[3]), "\n", NULL);

	about = (GtkAboutDialog*) gtk_about_dialog_new ();
	gtk_window_set_icon_name (GTK_WINDOW (about), "laser-editor");
	gtk_window_set_transient_for (GTK_WINDOW (about), GET_WINDOW (main));
	gtk_window_set_position (GTK_WINDOW (about), GTK_WIN_POS_CENTER_ON_PARENT);
	gtk_about_dialog_set_name (about, _("Laser level set editor"));
	gtk_about_dialog_set_version (about, VERSION);
	gtk_about_dialog_set_copyright (about, _("Copyright \xc2\xa9 2007 Benjamin Canou"));
	gtk_about_dialog_set_comments (about, _("An enlightened game for the gnome desktop"));
	gtk_about_dialog_set_authors (about, authors);
	gtk_about_dialog_set_documenters (about, documenters);
	gtk_about_dialog_set_translator_credits (about, translator_credits);
	gtk_about_dialog_set_logo (about, logo);
	gtk_about_dialog_set_license (about, license_trans);
	gtk_about_dialog_set_website (about, "http://bookeldor-net.info");

	g_signal_connect (G_OBJECT(about), "response",
			  G_CALLBACK(gtk_widget_destroy), NULL);

	gtk_widget_show (GTK_WIDGET(about));

	g_free (license_trans);
	gdk_pixbuf_unref (logo);
}

static void select_object_cb (GtkMenuItem *menuitem, gpointer data) {
	editor_instance *editor = ((gpointer*) data)[0];
	struct tool *tool = ((gpointer*) data)[1];

	if (editor->loaded) {
		editor->tool = tool;
		gtk_tool_button_set_icon_name (GTK_TOOL_BUTTON (GET_WIDGET (main, button_object)), tool->name);
		laser_board_set_mode (editor->board, SET_MODE, editor->tool->type, editor->tool->colour);
	}
}

static void edit_mode_cb (GtkToggleToolButton *btn, gpointer data) {
	editor_instance *editor = data;
	if (editor->loaded) {
		SET_SENSITIVE (main, button_object, FALSE);
		laser_board_set_mode (editor->board, EDIT_MODE, NONE, NONE);
	}
}

static void erase_mode_cb (GtkToggleToolButton *btn, gpointer data) {
	editor_instance *editor = data;
	if (editor->loaded) {
		SET_SENSITIVE (main, button_object, FALSE);
		laser_board_set_mode (editor->board, SET_MODE, NONE, NONE);
	}
}

static void add_mode_cb (GtkToggleToolButton *btn, gpointer data) {
	editor_instance *editor = data;
	if (editor->loaded) {
		SET_SENSITIVE (main, button_object, TRUE);
		laser_board_set_mode (editor->board, SET_MODE, editor->tool->type, editor->tool->colour);
	}
}

static void objects_menu_cb (GtkToggleToolButton *btn, gpointer data) {
	editor_instance *editor = data;
	if (editor->loaded) {
		gtk_menu_popup (GTK_MENU (GET_WIDGET (main, menu_objects)), NULL, NULL, NULL, NULL, 0, gtk_get_current_event_time());
	}	
}

static void reset_cb (GtkWidget *widget, gpointer data) {
	editor_instance *editor = data;
	if (editor->loaded) {
		laser_board_reset (editor->board);
	}
}

static void redo_cb (GtkWidget *widget, gpointer data) {
	editor_instance *editor = data;
	if (editor->loaded) {
		laser_board_redo (editor->board);
	}
}

static void undo_cb (GtkWidget *widget, gpointer data) {
	editor_instance *editor = data;
	if (editor->loaded) {
		laser_board_undo (editor->board);
	}
}

static void save_solution_cb (GtkWidget *widget, gpointer data) {
	editor_instance *editor = data;
	if (editor->loaded) {
		if (editor->level_set->levels[editor->level].solution != NULL) {
			g_free (editor->level_set->levels[editor->level].solution);
		}
		editor->level_set->levels[editor->level].solution = laser_board_get_cheat (editor->board);
	}
}

/**
 * next_cb:
 * @widget: the next button
 * @data: unused
 *
 * Callback called on click on the next button
 */
static void next_cb (GtkWidget *widget, gpointer data) {
	editor_instance *editor = data;
  if (editor->loaded == TRUE && editor->level < editor->level_set->nb_levels - 1) {
    load_level (editor, editor->level_set->file, ++editor->level);
  }
}

/**
 * previous_cb:
 * @widget: the previous button
 * @data: unused
 *
 * Callback called on click on the previous button
 */
static void previous_cb (GtkWidget *widget, gpointer data) {
	editor_instance *editor = data;
  if (editor->loaded == TRUE && editor->level > 0) {
			load_level (editor, editor->level_set->file, --editor->level);
  }
}

static void delete_level_cb (GtkWidget *widget, gpointer data) {
	editor_instance *editor = data;

	free_i18n (editor->level_set->levels[editor->level].name_i18n);
	free_i18n (editor->level_set->levels[editor->level].comment_i18n);
	if (editor->level_set->levels[editor->level].solution != NULL) {
		g_free (editor->level_set->levels[editor->level].solution);
	}
	for (int i = 0;i < editor->level_set->levels[editor->level].height;i++) {
		g_free (editor->level_set->levels[editor->level].map[i]);
	}
	g_free (editor->level_set->levels[editor->level].map);
	editor->level_set->nb_levels--;
	for (int i = editor->level;i < editor->level_set->nb_levels;i++) {
		memcpy (&editor->level_set->levels[i], &editor->level_set->levels[i + 1], sizeof (struct level));
	}
	if (editor->level == editor->level_set->nb_levels) {
		editor->level--;
	}
	load_level (editor, editor->level_set->file, editor->level);
	SET_SENSITIVE (main, button_delete_level, editor->level_set->nb_levels > 1);
}

static void add_before_cb (GtkWidget *widget, gpointer data) {
	editor_instance *editor = data;
	editor->level_set->nb_levels++;
	if (editor->level_set->nb_levels == 1) {
		editor->level_set->levels = g_malloc (sizeof (struct level));
	} else {
		editor->level_set->levels = g_realloc (editor->level_set->levels, editor->level_set->nb_levels * sizeof (struct level));
	}
	for (int i = editor->level_set->nb_levels - 1;i > editor->level;i--) {
		editor->level_set->levels[i] = editor->level_set->levels[i - 1];
	}
	editor->level_set->levels[editor->level].solution = NULL;
	editor->level_set->levels[editor->level].width = 4;
	editor->level_set->levels[editor->level].height = 5;
	editor->level_set->levels[editor->level].real_height = 4;
	editor->level_set->levels[editor->level].difficulty = 0;
	for (int i = 0;i < MATERIALS;i++) {
		editor->level_set->levels[editor->level].materials[i] = 0;
	}
	editor->level_set->levels[editor->level].map = g_malloc (5 * sizeof (struct item*));
	for (int i = 0;i < 5;i++) {
		editor->level_set->levels[editor->level].map[i] = g_malloc (4 * sizeof (struct item));
		for (int j = 0;j < 4;j++) {
			editor->level_set->levels[editor->level].map[i][j].type = (i == 4) ? MAP_WALL : NONE;
			editor->level_set->levels[editor->level].map[i][j].colour = NONE;
			editor->level_set->levels[editor->level].map[i][j].direction = 0;
		}
	}
	gchar **pair;
	pair = g_malloc (2 * sizeof (gchar*));
	pair[0] = g_strdup ("C");
	pair[1] = g_strdup ("");
	editor->level_set->levels[editor->level].name_i18n = NULL;
	editor->level_set->levels[editor->level].name_i18n =
		g_slist_prepend (editor->level_set->levels[editor->level].name_i18n, pair);
	editor->level_set->levels[editor->level].name = pair[0];
	pair = g_malloc (2 * sizeof (gchar*));
	pair[0] = g_strdup ("C");
	pair[1] = g_strdup ("");
	editor->level_set->levels[editor->level].comment_i18n = NULL;
	editor->level_set->levels[editor->level].comment_i18n =
		g_slist_prepend (editor->level_set->levels[editor->level].comment_i18n, pair);
	editor->level_set->levels[editor->level].comment = pair[0];
	load_level (editor, editor->level_set->file, editor->level);
	SET_SENSITIVE (main, button_delete_level, editor->level_set->nb_levels > 1);
}

static void add_after_cb (GtkWidget *widget, gpointer data) {
	editor_instance *editor = data;
	editor->level++;
	add_before_cb (widget, data);
}

/**
 * open_cb:
 * @menuitem: the menu item activated
 * @data: unused
 * 
 * Shows the level set file open dialog.
 */
static void open_cb (GtkWidget *menuitem, gpointer data) {
	editor_instance *editor = data;
  GladeXML *ui = glade_xml_new (DATADIR "/laser/open.glade", NULL, NULL);
  GtkWidget *w_open = glade_xml_get_widget (ui, "dialog");

	gtk_widget_hide (glade_xml_get_widget (ui, "button_copy"));
	gboolean retry = TRUE;
	while (retry) {
		retry = FALSE;
		gint result = gtk_dialog_run (GTK_DIALOG (w_open));
		if (result == -3) {
			gchar* file_path = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER (w_open));
			GError *err = NULL;
			gint nb = 0;
			parse_level_set_check (file_path, &err, &nb);
			if (file_path != NULL && err == NULL) {
				load_level (editor, file_path, 0);
				SET_SENSITIVE (main, button_save, TRUE);
				SET_SENSITIVE (main, menu_save, TRUE);
				SET_SENSITIVE (main, menu_save_as, TRUE);
				SET_SENSITIVE (main, menu_save_backup, TRUE);
				retry = FALSE;
			} else {
				retry = file_error_dialog (file_path, (err == NULL) ? "NULL" : err->message);
				if (err != NULL) {
					g_error_free (err);
				}
			}
		}
	}
  gtk_widget_destroy (w_open);
}

static gint xfer_cb (GnomeVFSXferProgressInfo *info, gpointer data) {
	GError **err = (GError**) data;
	switch (info->status) {
	case GNOME_VFS_XFER_PROGRESS_STATUS_OK :
		return 1;
	case GNOME_VFS_XFER_PROGRESS_STATUS_VFSERROR :
		*err = g_error_new (0, 0, gnome_vfs_result_to_string (info->vfs_status));
		return GNOME_VFS_XFER_OVERWRITE_ACTION_ABORT;
	case GNOME_VFS_XFER_PROGRESS_STATUS_DUPLICATE : {
		gchar *duplicate_name = g_strdup_printf ("%s (%s %d)", info->duplicate_name, _("backup "), info->duplicate_count);
		g_free (info->duplicate_name);
		info->duplicate_name = duplicate_name;
		return 1;
	}
	default :
		return 0;
	}
}

static void save_as_cb (GtkWidget *menuitem, gpointer data) {
	editor_instance *editor = data;
	GError *err = NULL;

  GladeXML *ui = glade_xml_new (DATADIR "/laser/save.glade", NULL, NULL);
  GtkWidget *w_save = glade_xml_get_widget (ui, "dialog");

	if (editor->level_set->file != NULL) {
		gtk_file_chooser_set_uri (GTK_FILE_CHOOSER (w_save), editor->level_set->file);
	}
	gboolean retry = TRUE;
	while (retry) {
		retry = FALSE;
		gint result = gtk_dialog_run (GTK_DIALOG (w_save));
		if (result == -3 || result == -4) {
			gchar* file_path = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER (w_save));
			if (file_path != NULL) {
				gboolean backup = (result == -4);
				if (backup) {
					GnomeVFSURI *to = gnome_vfs_uri_new  (file_path);
					if (gnome_vfs_uri_exists (to)) {
						GError *err = NULL;
						GnomeVFSResult r = gnome_vfs_xfer_uri (to, to,
																									 GNOME_VFS_XFER_DEFAULT | GNOME_VFS_XFER_USE_UNIQUE_NAMES,
																									 GNOME_VFS_XFER_ERROR_MODE_QUERY,
																									 GNOME_VFS_XFER_OVERWRITE_MODE_ABORT,
																									 xfer_cb, &err);
						if (err != NULL) {
							retry = file_error_dialog (file_path,
																				 g_strconcat (_("Unable to backup :"),
																											err->message,
																											_("\n.Please retry without backup or choose another file."),
																											NULL));
							g_error_free (err);
						} else {
							if (r != GNOME_VFS_OK) {
								retry = file_error_dialog (file_path,
																					 g_strconcat (_("Unable to backup :"),
																												gnome_vfs_result_to_string(r),
																												_("\n.Please retry without backup or choose another file."),
																												NULL));
							}
						}
					}
				}
				if (!retry) {
					write_level_set (editor->level_set, file_path, &err);
					if (err != NULL) {
						retry = file_error_dialog (file_path, (err == NULL) ? "NULL" : err->message);
						g_error_free (err);
					} else {
												editor->modified = FALSE;
						SET_SENSITIVE (main, button_save, TRUE);
						SET_SENSITIVE (main, menu_save, TRUE);
						SET_SENSITIVE (main, menu_save_as, TRUE);
						SET_SENSITIVE (main, menu_save_backup, TRUE);
					}
				}
			}
		}
	}
  gtk_widget_destroy (w_save);
}

static void save_cb (GtkWidget *menuitem, gpointer data) {
	editor_instance *editor = data;
	GError *err = NULL;
	write_level_set (editor->level_set, editor->level_set->file, &err);
	if (err != NULL) {
		if (file_error_dialog (editor->level_set->file, (err == NULL) ? "NULL" : err->message)) {
			save_as_cb (menuitem, data);
		}
		g_error_free (err);
	} else {
				editor->modified = FALSE;
	}
}

static void save_backup_cb (GtkWidget *menuitem, gpointer data) {
	editor_instance *editor = data;
	GError *err = NULL;
	gboolean error = FALSE;
	gboolean retry = FALSE;
	GnomeVFSURI *to = gnome_vfs_uri_new  (editor->level_set->file);
	if (gnome_vfs_uri_exists (to)) {
		GError *err = NULL;
		GnomeVFSResult r = gnome_vfs_xfer_uri (to, to,
																					 GNOME_VFS_XFER_DEFAULT | GNOME_VFS_XFER_USE_UNIQUE_NAMES,
																					 GNOME_VFS_XFER_ERROR_MODE_QUERY,
																					 GNOME_VFS_XFER_OVERWRITE_MODE_ABORT,
																					 xfer_cb, &err);
		if (err != NULL) {
			error = TRUE;
			retry = file_error_dialog (editor->level_set->file,
																 g_strconcat (_("Unable to backup :"),
																							err->message,
																							_("\n.Please retry without backup or choose another file."),
																							NULL));
			g_error_free (err);
		} else {
			if (r != GNOME_VFS_OK) {
				error = TRUE;
				retry = file_error_dialog (editor->level_set->file,
																	 g_strconcat (_("Unable to backup :"),
																								gnome_vfs_result_to_string(r),
																								_("\n.Please retry without backup or choose another file."),
																								NULL));
			}
		}
	}
	if (!error) {
		write_level_set (editor->level_set, editor->level_set->file, &err);
		if (err != NULL) {
			if (file_error_dialog (editor->level_set->file, (err == NULL) ? "NULL" : err->message)) {
				save_as_cb (menuitem, data);
			}
			g_error_free (err);
		} else {
						editor->modified = FALSE;
		}
	} else {
		if (retry) {
			save_as_cb (menuitem, data);
		}
	}
}

static void resize (editor_instance *editor, gint grow_or_shrink, gint direction) {
	laser_board_resize (editor->board, grow_or_shrink, direction);
	editor->modified = TRUE;
	SET_SENSITIVE (main, button_expand_top, TRUE);
	SET_SENSITIVE (main, button_expand_bottom, TRUE);
	SET_SENSITIVE (main, button_expand_left, TRUE);
	SET_SENSITIVE (main, button_expand_right, TRUE);
	SET_SENSITIVE (main, button_shrink_top, editor->level_set->levels[editor->level].height > 2);
	SET_SENSITIVE (main, button_shrink_bottom, editor->level_set->levels[editor->level].height > 2);
	SET_SENSITIVE (main, button_shrink_left, editor->level_set->levels[editor->level].width > 2);
	SET_SENSITIVE (main, button_shrink_right, editor->level_set->levels[editor->level].width > 2);
}

void expand_top_cb (GtkWidget *widget, gpointer data) {
	editor_instance *editor = data;

	resize (editor, 1, 0);
}

void expand_bottom_cb (GtkWidget *widget, gpointer data) {
	editor_instance *editor = data;

	resize (editor, 1, 1);
}

void expand_left_cb (GtkWidget *widget, gpointer data) {
	editor_instance *editor = data;

	resize (editor, 1, 2);
}

void expand_right_cb (GtkWidget *widget, gpointer data) {
	editor_instance *editor = data;

	resize (editor, 1, 3);
}

void shrink_top_cb (GtkWidget *widget, gpointer data) {
	editor_instance *editor = data;

	resize (editor, -1, 0);
}

void shrink_bottom_cb (GtkWidget *widget, gpointer data) {
	editor_instance *editor = data;

	resize (editor, -1, 1);
}

void shrink_left_cb (GtkWidget *widget, gpointer data) {
	editor_instance *editor = data;

	resize (editor, -1, 2);
}

void shrink_right_cb (GtkWidget *widget, gpointer data) {
	editor_instance *editor = data;

	resize (editor, -1, 3);
}

void difficulty_cb (GtkRange *range, gpointer data) {
	editor_instance *editor = data;

	if (editor->level_set->levels[editor->level].difficulty != gtk_range_get_value (range)) {
		editor->level_set->levels[editor->level].difficulty = gtk_range_get_value (range);
			editor->modified = TRUE;
	}
}

static void new_cb (GtkRange *range, gpointer data) {
	editor_instance *editor = data;

	if (editor->modified) {
		GladeXML *ui = glade_xml_new (DATADIR G_DIR_SEPARATOR_S "laser" G_DIR_SEPARATOR_S "question.glade", NULL, NULL);
		GtkWidget *w_question = glade_xml_get_widget (ui, "dialog");
		
		gint r = gtk_dialog_run (GTK_DIALOG (w_question));
		gtk_widget_destroy (w_question);
		switch (r) {
		case -4 : /* continue */
			break;
		case -2 : /* save */
			if (strcmp (editor->level_set->file, "") == 0) {
				save_as_cb (NULL, editor);
			} else {
				save_cb (NULL, editor);
			}
		case -3 : /* cancel */
		default :
			return;
		}
	}
	if (editor->loaded) {
		level_set_free (editor->level_set);
	}
	editor->modified = FALSE;
	editor->level_set = level_set_new ();
	gchar **pair;
	pair = g_malloc (2 * sizeof (gchar*));
	pair[0] = g_strdup ("C");
	pair[1] = g_strdup ("");
	editor->level_set->name_i18n = NULL;
	editor->level_set->name_i18n =
		g_slist_prepend (editor->level_set->name_i18n, pair);
	editor->level_set->name = pair[0];
	pair = g_malloc (2 * sizeof (gchar*));
	pair[0] = g_strdup ("C");
	pair[1] = g_strdup ("");
	editor->level_set->note_i18n = NULL;
	editor->level_set->note_i18n =
		g_slist_prepend (editor->level_set->note_i18n, pair);
	editor->level_set->note = pair[0];
	fill_tab (editor->tab_translations, &editor->level_set->name_i18n, &editor->level_set->note_i18n);
	editor->level_set->nb_levels = 0;
	editor->level = 0;
	editor->level_set->file = g_strdup ("");
	editor->loaded = TRUE;
	add_before_cb (NULL, editor);
	SET_SENSITIVE (main, button_add_after, TRUE);
	SET_SENSITIVE (main, button_add_before, TRUE);
	SET_SENSITIVE (main, menu_save_as, TRUE);
	SET_SENSITIVE (main, edit_level_difficulty, TRUE);
}

/**
 * quit_cb:
 *
 * Quits saving history (called from menu)
 */
static void quit_cb (GtkMenuItem *menuitem, gpointer data) {
	editor_instance *editor = data;

	if (editor->modified) {
		GladeXML *ui = glade_xml_new (DATADIR "/laser/question.glade", NULL, NULL);
		GtkWidget *w_question = glade_xml_get_widget (ui, "dialog");
		
		gint r = gtk_dialog_run (GTK_DIALOG (w_question));
		gtk_widget_destroy (w_question);
		switch (r) {
		case -4 : /* continue */
			break;
		case -2 : /* save */
			if (strcmp (editor->level_set->file, "") == 0) {
				save_as_cb (NULL, editor);
			} else {
				save_cb (NULL, editor);
			}
		case -3 : /* cancel */
		default :
			return;
		}
	}
	gtk_main_quit ();
}

/**
 * delete_cb:
 *
 * Quits saving history (called on window close)
 */
static gboolean delete_cb (GtkWidget *widget, GdkEvent *event, gpointer data) {
	quit_cb (NULL, data);
  return TRUE;
}

/* preferences ****************************************************************/

/**
 * color_set_cb:
 * @widget: the colour button
 * @data: unused
 *
 * sets the background colour from colour button (with gconf)
 */
static void color_set_cb (GtkColorButton *widget, gpointer data) {
	editor_instance *editor = data;

	GdkColor gdk_colour;
	colour_t colour;
	gtk_color_button_get_color (widget, &gdk_colour);
	colour_from_gdk (&colour, gdk_colour, gtk_color_button_get_alpha (widget));
	laser_config_set_colour (editor->config, &colour);
}

/**
 * combo_theme_change_cb:
 * @combo: the theme combo
 * @data: unused
 *
 * Sets theme from combo (with gconf)
 */
static void combo_theme_change_cb (GtkComboBox *combo,	gpointer data) {
	editor_instance *editor = data;
  GtkTreeIter it;
  if (gtk_combo_box_get_active_iter (combo, &it)) {
    GtkTreeModel* model = gtk_combo_box_get_model (combo);
    gchar *name;
    gtk_tree_model_get (model, &it, 1, &name, -1);
		laser_config_set_theme (editor->config, name);
	}
}

/**
 * stats_toggled_cb:
 * @togglebutton: the stats checkbox
 * @data: unused
 *
 * Set statistics display from checkbox (with gconf)
 */
static void stats_toggled_cb (GtkToggleButton *togglebutton, gpointer data) {
	editor_instance *editor = data;

	laser_config_set_show_stats (editor->config, gtk_toggle_button_get_active (togglebutton));
}

/**
 * restore_toggled_cb:
 * @togglebutton: the restore checkbox
 * @data: unused
 *
 * Set restoration from checkbox (with gconf)
 */
static void restore_toggled_cb (GtkToggleButton *togglebutton, gpointer data) {
	editor_instance *editor = data;

	laser_config_set_restore (editor->config, gtk_toggle_button_get_active (togglebutton));
}

/**
 * arrows_toggled_cb:
 * @togglebutton: the arrows checkbox
 * @data: unused
 *
 * Set restoration from checkbox (with gconf)
 */
static void arrows_toggled_cb (GtkToggleButton *togglebutton, gpointer data) {
	editor_instance *editor = data;

	laser_config_set_arrows (editor->config, gtk_toggle_button_get_active (togglebutton));
}

/**
 * preferences_cb:
 * @menuitem: the menu item activated
 * @data: unused
 * 
 * Shows the preferences dialog.
 */
static void preferences_cb (GtkMenuItem *menuitem, gpointer data) {
	editor_instance *editor = data;

  editor->win_prefs = glade_xml_new (LASER_CONFIG_SHARE "editor-preferences.glade", NULL, NULL);
  
	colour_t colour;
	laser_config_get_colour (editor->config, &colour);

  gtk_color_button_set_color (GTK_COLOR_BUTTON (GET_BUTTON (prefs, color)), &colour.gdk_colour);
  gtk_color_button_set_alpha (GTK_COLOR_BUTTON (GET_BUTTON (prefs, color)), colour.gdk_alpha);
  SET_CHECK_ACTIVE (prefs, stats, laser_config_get_show_stats (editor->config));
  SET_CHECK_ACTIVE (prefs, restore, laser_config_get_restore (editor->config));
  SET_CHECK_ACTIVE (prefs, arrows, laser_config_get_arrows (editor->config));

  CONNECT (prefs, button_color, "color-set", color_set_cb);
  CONNECT (prefs, check_stats, "toggled", stats_toggled_cb);
  CONNECT (prefs, check_restore, "toggled",	restore_toggled_cb);
  CONNECT (prefs, check_arrows, "toggled", arrows_toggled_cb);
	
  GtkListStore  *store = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_STRING);
	GSList* themes = get_available_themes ();
  int sel = 0, i = 0;
  for (GSList *l = themes;l != NULL;l = g_slist_next(l)) {
		GtkTreeIter    iter;
    gtk_list_store_append (store, &iter);
    gtk_list_store_set (store, &iter, 0, (gchar*)((gpointer*)l->data)[1], 1, (gchar*)((gpointer*)l->data)[0], -1);
    if (strcmp (((gchar**)l->data)[0], laser_config_get_theme (editor->config)) == 0) {
      sel = i;
    }
    i++;
  }
	gtk_combo_box_set_model (GET_COMBO (prefs, theme), GTK_TREE_MODEL (store));
  SET_SENSITIVE (prefs, combo_theme, TRUE);
  CONNECT (prefs, combo_theme, "changed", combo_theme_change_cb);
	gtk_combo_box_set_active (GET_COMBO (prefs, theme), sel); 
  gtk_dialog_run (GET_DIALOG (prefs));
  gtk_widget_destroy (GET_WIDGET (prefs, dialog));

	free_available_themes (themes);
}

static void resize_window_cb (GtkWidget *window, GtkAllocation *allocation, gpointer data) {
	editor_instance *editor = data;

	laser_config_set_editor_height (editor->config, allocation->height);
	laser_config_set_editor_width (editor->config, allocation->width);
}


/* main ***********************************************************************/

int main (int argc, gchar **argv) {
  gtk_init (&argc, &argv);
  rsvg_init (); 
  gconf_init(argc, argv, NULL);


	GOptionContext *context;
	context = g_option_context_new (_("Laser level set editor"));
	
	bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
	textdomain (GETTEXT_PACKAGE);
	
	program = gnome_program_init (PACKAGE, VERSION, LIBGNOMEUI_MODULE, argc, argv,
																GNOME_PROGRAM_STANDARD_PROPERTIES,
			    									    GNOME_PARAM_GOPTION_CONTEXT, context,
																GNOME_PARAM_NONE);

	
	editor_instance *editor = g_malloc (sizeof (struct editor_instance));
	editor->config = laser_config_new ();
	editor->timer_offset = 0;
	editor->loaded = FALSE;
	editor->supports_alpha = FALSE;
	editor->level_set = NULL;
	editor->level = 0;
	editor->tool = &static_tools[0];
	editor->modified = FALSE;
  editor->win_main = glade_xml_new (LASER_CONFIG_SHARE "editor.glade", NULL, NULL);
	editor->tab_translations = init_tab (editor, GET_WIDGET (main, tab_translations));
	editor->tab_level_translations = init_tab (editor, GET_WIDGET (main, tab_level_translations));
	editor->board = laser_board_attach (GET_WIDGET (main, board), EDIT_MODE, editor->config,
																			update_states, &editor->modified, editor);

  screen_changed (GET_WIDGET (main, window), NULL, editor);

  CONNECT (main, window, "screen-changed", screen_changed);
  CONNECT (main, window, "delete-event", delete_cb);
  CONNECT (main, window, "size-allocate", resize_window_cb);
	CONNECT (main, menu_quit,	"activate", quit_cb);
	CONNECT (main, menu_about,	"activate", dialog_about_cb);
	CONNECT (main, menu_open,	"activate", open_cb);
	CONNECT (main, button_open,	"clicked", open_cb);
	CONNECT (main, menu_new,	"activate", new_cb);
	CONNECT (main, button_new,	"clicked", new_cb);
	CONNECT (main, menu_save,	"activate", save_cb);
	CONNECT (main, button_save,	"clicked", save_cb);
	CONNECT (main, menu_save_as,	"activate", save_as_cb);
	CONNECT (main, menu_save_backup,	"activate", save_backup_cb);
	CONNECT (main, menu_prefs,  "activate", preferences_cb);
	CONNECT (main, edit_level_difficulty,	"value-changed", difficulty_cb);
	CONNECT (main, button_delete_level,	"clicked", delete_level_cb);
	CONNECT (main, button_add_after,	"clicked", add_after_cb);
	CONNECT (main, button_add_before,	"clicked", add_before_cb);
	CONNECT (main, button_expand_top, "clicked", expand_top_cb);
	CONNECT (main, button_expand_bottom, "clicked", expand_bottom_cb);
	CONNECT (main, button_expand_left, "clicked", expand_left_cb);
	CONNECT (main, button_expand_right, "clicked", expand_right_cb);
	CONNECT (main, button_shrink_top, "clicked", shrink_top_cb);
	CONNECT (main, button_shrink_bottom, "clicked", shrink_bottom_cb);
	CONNECT (main, button_shrink_left, "clicked", shrink_left_cb);
	CONNECT (main, button_shrink_right, "clicked", shrink_right_cb);
	CONNECT (main, button_edit, "toggled", edit_mode_cb);
	CONNECT (main, button_erase, "toggled", erase_mode_cb);
	CONNECT (main, button_add, "toggled", add_mode_cb);
	CONNECT (main, button_object, "clicked", objects_menu_cb);
	CONNECT (main, button_redo, "clicked", redo_cb);
	CONNECT (main, button_undo, "clicked", undo_cb);
	CONNECT (main, button_save_solution, "clicked", save_solution_cb);
	CONNECT (main, button_previous, "clicked", previous_cb);
	CONNECT (main, button_next, "clicked", next_cb);
  CONNECT (main, button_reset, "clicked", reset_cb);

	for (int i = 0;static_tools[i].name != NULL;i++) {
		gchar *uic = g_strconcat ("object_menu_", static_tools[i].name, NULL);
		gpointer *pair = malloc (2 * sizeof (gpointer));
		pair[0] = editor;
		pair[1] = &static_tools[i];
		g_signal_connect (GET_WIDGET_S (main, uic),	"activate", G_CALLBACK (select_object_cb), pair);
		g_free (uic);
	}

  g_signal_connect (editor->config, "arrows_changed", G_CALLBACK (arrows_changed_cb), editor);
  g_signal_connect (editor->config, "colour_changed", G_CALLBACK (colour_changed_cb), editor);
  g_signal_connect (editor->config, "theme_changed", G_CALLBACK (theme_changed_cb), editor);

  gtk_window_set_default_size (GET_WINDOW (main),
															 laser_config_get_editor_width (editor->config),
															 laser_config_get_editor_height (editor->config));

  gtk_widget_show_all (GET_WIDGET (main, window));

  gtk_main ();

  return 0;
}
