/* -*- Mode: C; tab-width: 2; indent-tabs-mode: t; c-basic-offset: 2 -*-
 *
 * LASER
 * Copyright (C) 2006 Benjamin Canou <bookeldor@gmail.com>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "config.h"

#include <gtk/gtk.h>
#include <glade/glade.h>
#include <gconf/gconf-client.h>
#include <libgnomeui/libgnomeui.h>
#include <libgnomeui/gnome-help.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib/gi18n.h>
#include <librsvg/rsvg.h>

#include "laser-common.h"
#include "laser-config.h"
#include "laser-parser.h"
#include "laser-board.h"

/* datatypes and macros *******************************************************/

typedef struct game_instance {
	/* config instance */
	LaserConfig *config;
	/* game properties */
	board_t     *board;
	gboolean    loaded;
	gboolean    modified;
	GTimer      *timer_clock;
	gint        timer_offset;
	level_set   *level_set;
	gint        level;
	gboolean    won;
	/* ui */
	gboolean    displayed;
	gboolean    supports_alpha;
  GladeXML    *win_main;
  GladeXML    *win_new;
  GladeXML    *win_cheat;
  GladeXML    *win_prefs;
} game_instance;

/* ui shortcut macros */
#define GET_WINDOW(w)                GTK_WINDOW (glade_xml_get_widget (game->win_##w, "window"))
#define GET_DIALOG(w)                GTK_DIALOG (glade_xml_get_widget (game->win_##w, "dialog"))
#define GET_WIDGET(w, n)             glade_xml_get_widget (game->win_##w, #n)
#define GET_LABEL(w, n)              GTK_LABEL (glade_xml_get_widget (game->win_##w, "label_" #n))
#define GET_BUTTON(w, n)             GTK_BUTTON (glade_xml_get_widget (game->win_##w, "button_" #n))
#define GET_COMBO(w, n)              GTK_COMBO_BOX (glade_xml_get_widget (game->win_##w, "combo_" #n))
#define GET_CHECK(w, n)              GTK_CHECK_BUTTON (glade_xml_get_widget (game->win_##w, "check_" #n))
#define GET_ENTRY(w, n)              GTK_ENTRY (glade_xml_get_widget (game->win_##w, "entry_" #n))
#define CONNECT(w, n, s, cb)         g_signal_connect (GET_WIDGET (w,n), s, G_CALLBACK (cb), game)
#define CONNECT_DATA(w, n, s, cb, d) g_signal_connect (GET_WIDGET (w,n), s, G_CALLBACK (cb), d)
#define SET_SENSITIVE(w, n, v)       gtk_widget_set_sensitive (glade_xml_get_widget (game->win_##w, #n), v)
#define SET_CHECK_ACTIVE(w, n, v)    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (GET_WIDGET (w,check_##n)), v)
#define GET_CHECK_ACTIVE(w, n)       gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (GET_WIDGET (w,check_##n)))

/* globals ********************************************************************/

static GnomeProgram *program;

/* prototypes ************************ *****************************************/

static void     update_states             (gpointer data, gint total, gint reached, gint actions, gboolean undo, gboolean redo);
static void     screen_changed            (GtkWidget *widget, GdkScreen *old_screen, gpointer data);
static void     update_label              (game_instance *game);
static void     load_level                (game_instance *game, gchar *set, gint level);
static void     reset_cb                  (GtkWidget *widget, gpointer data);
static void     redo_cb                   (GtkWidget *widget, gpointer data);
static void     undo_cb                   (GtkWidget *widget, gpointer data);
static void     next_cb                   (GtkWidget *widget, gpointer data);
static void     previous_cb               (GtkWidget *widget, gpointer data);
static gint     xfer_cb                   (GnomeVFSXferProgressInfo *info, gpointer data);
static void     open_cb                   (GtkWidget *menuitem, gpointer data);
static void     combo_level_change_cb     (GtkWidget *combo_level, gpointer data);
static void     combo_level_set_change_cb (GtkWidget *combo_level_set, gpointer data);
static void     new_game_cb               (GtkWidget *menuitem, gpointer data);
static void     color_set_cb              (GtkColorButton *widget, gpointer data);
static void     combo_theme_change_cb     (GtkComboBox *combo,	gpointer data);
static void     stats_toggled_cb          (GtkToggleButton *togglebutton, gpointer data);
static void     restore_toggled_cb        (GtkToggleButton *togglebutton, gpointer data);
static void     arrows_toggled_cb         (GtkToggleButton *togglebutton, gpointer data);
static void     preferences_cb            (GtkMenuItem *menuitem, gpointer data);
static void     resize_window_cb          (GtkWidget *window, GtkAllocation *allocation, gpointer data);
static gboolean clock_display_cb          (gpointer data);
static void     quit_cb                   (GtkMenuItem *menuitem, gpointer data);
static gboolean delete_cb                 (GtkWidget *widget, GdkEvent *event, gpointer data);
static void     dialog_about_cb           (GtkMenuItem *menuitem, gpointer data);
static void     help_cb                   (GtkMenuItem *menuitem, gpointer data);
static void     colour_changed_cb         (LaserConfig *laser_config, colour_t *colour, gpointer data);
static void     arrows_changed_cb         (LaserConfig *laser_config, gboolean arrows, gpointer data);
static void     show_stats_changed_cb     (LaserConfig *laser_config, gboolean show_stats, gpointer data);
static void     theme_changed_cb          (LaserConfig *laser_config, gpointer data);
static void     cheat_copy_cb             (GtkWidget *widget, gpointer data);
static void     cheat_paste_cb            (GtkWidget *widget, gpointer data);
static void     cheat_cb                  (GtkMenuItem *menuitem, gpointer data);


/**
 * update_states:
 *
 * Updates display of state of game (number of targets to reach, or
 * success if all reached). Updates number of actions too.
 */
static void update_states (gpointer data, gint total, gint reached, gint actions, gboolean undo, gboolean redo) {
	game_instance *game = data;
	
	if (total != reached) {
		if (total - reached == 1) {
			gtk_label_set_markup (GET_LABEL (main, state), _("Still <b>one</b> target to reach..."));
		} else {
			gchar *label = g_strdup_printf (_("Still <b>%d</b> targets to reach..."), total - reached);
			gtk_label_set_markup (GET_LABEL (main, state), label);
			g_free (label);
		}
	} else {
		gtk_label_set_markup (GET_LABEL (main, state), _("<b>LEVEL CLEARED</b>"));
		if (game->displayed && !game->won) {
			GladeXML *ui = glade_xml_new (LASER_CONFIG_SHARE "level-finished.glade", NULL, NULL);
			GtkWidget *dlg = glade_xml_get_widget (ui, "dialog");
			if (! (game->level < game->level_set->nb_levels - 1)) {
				gtk_widget_hide (glade_xml_get_widget (ui, "button_next"));
			}
			gint r =gtk_dialog_run (GTK_DIALOG (dlg));
			gtk_widget_destroy (dlg);
			if (r== -3) {
				next_cb (NULL, game);
			}
		}
		game->won = TRUE;
	}
	
	/* update number of actions label */
	gchar *label = g_strdup_printf (_("<b>Actions : </b>%d"), actions);
	gtk_label_set_markup (GET_LABEL (main, actions), label);
	g_free (label);
	laser_config_set_actions (game->config, actions);
	SET_SENSITIVE (main, button_undo, undo);
	SET_SENSITIVE (main, menu_undo, undo);
	SET_SENSITIVE (main, button_redo, redo);
	SET_SENSITIVE (main, menu_redo, redo);
}

/**
 * screen_changed:
 * @widget: the migrated widget
 * @old_screen: unused
 * @data: unused
 *
 * Checks if alpha is supported at each migration.
 */
static void screen_changed (GtkWidget *widget, GdkScreen *old_screen, gpointer data) {
	GdkScreen *screen = gtk_widget_get_screen(widget);
	GdkColormap *colormap = gdk_screen_get_rgba_colormap(screen);
	game_instance *game = data;

	/* supports alpha is able to get an rgba color map */
	if (!colormap) {
		colormap = gdk_screen_get_rgb_colormap(screen);
		printf ("Your screen supports alpha channel, translucent background enabled.\n");
		game->supports_alpha = FALSE;
	} else {
		game->supports_alpha = TRUE;
	}

	/* sets color map accordingly */
	gtk_widget_set_colormap(widget, colormap);
}



/**
 * update_label:
 *
 * Update display of name and comment of current level. Also sets
 * sensitivity of next and previous buttons;
 */
static void update_label (game_instance *game) {
  if (game->loaded) {
    gchar *buf;
    buf = g_strconcat ("<b>",game->level_set->levels[game->level].name,"</b>\n",
											 "<small>",game->level_set->levels[game->level].comment,"</small>", NULL);
    gtk_label_set_markup (GET_LABEL (main, title), buf);
    g_free (buf);
		SET_SENSITIVE (main, button_previous, game->level > 0);
    SET_SENSITIVE (main, button_next, game->level < game->level_set->nb_levels - 1);
  }
}

/**
 * load_level:
 * @set: the set folder
 * @level: the level index in the set
 *
 * Loads the given level and forces repaint.
 */
static void load_level (game_instance *game, gchar *set, gint level) {
	if ((!game->loaded) || strcmp (game->level_set->file, set) != 0) {
		if (game->loaded) {
			level_set_free (game->level_set);
		}
		game->level_set = level_set_new ();
		GError *err = NULL;
		parse_level_set (set, game->level_set, &err);
		if (err != NULL) {
			printf (_("Error : \"%s\"\n"), err->message);
			g_error_free (err);
		}
	}
	game->level = level;
	laser_board_set_level (game->board, &game->level_set->levels [game->level]);
	game->loaded = TRUE;
	update_label (game);
	g_timer_reset (game->timer_clock);
	game->timer_offset = 0;
	game->won = FALSE;
	SET_SENSITIVE (main, button_reset, TRUE);
	SET_SENSITIVE (main, button_cheat, TRUE);
	SET_SENSITIVE (main, menu_reset, TRUE);
	SET_SENSITIVE (main, menu_cheat, TRUE);
}

/**
 * reset_cb:
 * @widget: the reset button
 * @data: unused
 *
 * Callback called on click on the reset button
 */
static void reset_cb (GtkWidget *widget, gpointer data) {
	game_instance *game = data;
  if (game->loaded) {
		laser_board_reset (game->board);
		clock_display_cb (game);
  }
}

static void redo_cb (GtkWidget *widget, gpointer data) {
	game_instance *game = data;
	if (game->loaded) {
		laser_board_redo (game->board);
	}
}

static void undo_cb (GtkWidget *widget, gpointer data) {
	game_instance *game = data;
	if (game->loaded) {
		laser_board_undo (game->board);
	}
}

/**
 * next_cb:
 * @widget: the next button
 * @data: unused
 *
 * Callback called on click on the next button
 */
static void next_cb (GtkWidget *widget, gpointer data) {
	game_instance *game = data;
  if (game->loaded == TRUE && game->level < game->level_set->nb_levels - 1) {
    load_level (game, game->level_set->file, ++game->level);
  }
}

/**
 * previous_cb:
 * @widget: the previous button
 * @data: unused
 *
 * Callback called on click on the previous button
 */
static void previous_cb (GtkWidget *widget, gpointer data) {
	game_instance *game = data;
  if (game->loaded == TRUE && game->level > 0) {
			load_level (game, game->level_set->file, --game->level);
  }
}

static gint xfer_cb (GnomeVFSXferProgressInfo *info, gpointer data) {
	GError **err = (GError**) data;
	switch (info->status) {
	case GNOME_VFS_XFER_PROGRESS_STATUS_OK :
		return 1;
	case GNOME_VFS_XFER_PROGRESS_STATUS_VFSERROR :
		*err = g_error_new (0, 0, gnome_vfs_result_to_string (info->vfs_status));
		return GNOME_VFS_XFER_OVERWRITE_ACTION_ABORT;
	case GNOME_VFS_XFER_PROGRESS_STATUS_DUPLICATE : {
		gchar *duplicate_name = g_strdup_printf ("%s (%d)", info->duplicate_name, info->duplicate_count);
		g_free (info->duplicate_name);
		info->duplicate_name = duplicate_name;
		return 1;
	}
	default :
		return 0;
	}
}

/**
 * open_cb:
 * @menuitem: the menu item activated
 * @data: unused
 * 
 * Shows the level set file open dialog.
 */
static void open_cb (GtkWidget *menuitem, gpointer data) {
	game_instance *win = data;
  GladeXML *ui = glade_xml_new (LASER_CONFIG_SHARE "open.glade", NULL, NULL);
  GtkWidget *w_open = glade_xml_get_widget (ui, "dialog");

	gboolean retry = TRUE;
	while (retry) {
		retry = FALSE;
		gint result = gtk_dialog_run (GTK_DIALOG (w_open));
		if (result == -3) {
			gchar* file_path = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER (w_open));
			GError *err = NULL;
			gint nb = 0;
			parse_level_set_check (file_path, &err, &nb);
			if (file_path != NULL && err == NULL) {
				load_level (win, file_path, 0);
				retry = FALSE;
			} else {
				retry = file_error_dialog (file_path, (err == NULL) ? "NULL" : err->message);
				if (err != NULL) {
					g_error_free (err);
				}
			}
		}
		if (result == -4) {
			gchar* file_uri = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER (w_open));
			if (file_uri != NULL) {
				/* get only name */
				gchar *copy_base = file_uri;
				for (gchar *s = file_uri;*s != '\0';s++) {
					if (*s == GNOME_VFS_URI_PATH_CHR) {
						copy_base = s;
					}
				}
				gchar *copy_path = g_strconcat (g_get_user_config_dir (),
																				G_DIR_SEPARATOR_S "laser" G_DIR_SEPARATOR_S "levels" G_DIR_SEPARATOR_S,
																				copy_base, NULL);
				gchar *copy_uri = gnome_vfs_get_uri_from_local_path (copy_path);
				GnomeVFSURI *from = gnome_vfs_uri_new  (file_uri);
				GnomeVFSURI *to = gnome_vfs_uri_new  (copy_uri);
				GError *err = NULL;
				GnomeVFSResult r = gnome_vfs_xfer_uri (from, to,
																							 GNOME_VFS_XFER_DEFAULT | GNOME_VFS_XFER_USE_UNIQUE_NAMES,
																							 GNOME_VFS_XFER_ERROR_MODE_QUERY,
																							 GNOME_VFS_XFER_OVERWRITE_MODE_ABORT,
																							 xfer_cb, &err);
				if (err != NULL) {
					retry = file_error_dialog (file_uri, (err == NULL) ? "NULL" : err->message);
					g_error_free (err);
				} else {
					if (r != GNOME_VFS_OK) {
						gchar *msg = g_strconcat (_("<b>While copying</b>\n  "), file_uri,
																			_("\n<b>to</b>\n  "), copy_uri,"\n  \n<b>",
																			(gchar*) gnome_vfs_result_to_string(r), "</b>",
																			NULL);
						retry = file_error_dialog (file_uri, msg);
					} else {
						GError *err = NULL;
						gint nb = 0;
						parse_level_set_check (copy_uri, &err, &nb);
						if (err == NULL) {
							load_level (win, file_uri, 0);
							retry = FALSE;
						} else {
							retry = file_error_dialog (file_uri, (err == NULL) ? "NULL" : err->message);
							g_error_free (err);
						}
					}
				}
			}
		}
	}
  gtk_widget_destroy (w_open);
}

/**
 * combo_level_change_cb:
 * @combo_level: the level combo
 * @data: the set combo
 *
 * Callback called when a level is selected, instantly loads it.
 */
static void combo_level_change_cb (GtkWidget *combo_level, gpointer data) {
	game_instance *game = data;

  GtkTreeIter it;
  if (gtk_combo_box_get_active_iter (GET_COMBO (new, level_set), &it)) {
    GtkTreeModel* model =
      gtk_combo_box_get_model (GET_COMBO (new, level_set));
    gchar *set;
    gtk_tree_model_get (model, &it, 1, &set, -1);

    GtkTreeIter it_l;
    if (gtk_combo_box_get_active_iter (GET_COMBO (new, level), &it_l)) {
      GtkTreeModel* model_l =
				gtk_combo_box_get_model (GET_COMBO (new, level));
      int level;
      gtk_tree_model_get (model_l, &it_l, 1, &level, -1);

      if ((!game->loaded) || strcmp (set, game->level_set->file) != 0 || level != game->level) {
				load_level (game, set, level);
			}
			gtk_label_set_text (GET_LABEL (new, comment), game->level_set->levels[game->level].comment);
			gchar dif [30];
			gint d = game->level_set->levels[game->level].difficulty;
			snprintf (dif, 29, "%s%s%s%s%s",
								(d == 1 ? "◐" : (d < 1 ?"○" : "●")),
								(d == 3 ? "◐" : (d < 3 ?"○" : "●")),
								(d == 5 ? "◐" : (d < 5 ?"○" : "●")),
								(d == 7 ? "◐" : (d < 7 ?"○" : "●")),
								(d == 9 ? "◐" : (d < 9 ?"○" : "●")));
			gtk_label_set_text (GET_LABEL (new, difficulty), dif);
    }
  }
}

/**
 * combo_level_set_change_cb:
 * @combo_level: the set combo
 * @data: the level combo
 *
 * Callback called when a set is selected, fills level combo and
 * selects the first level in the set
 */
static void combo_level_set_change_cb (GtkWidget *combo_level_set, gpointer data) {
	game_instance *game = data;
  
  GtkTreeIter it;
  if (gtk_combo_box_get_active_iter (GET_COMBO (new, level_set), &it)) {
    GtkTreeModel* model =
      gtk_combo_box_get_model (GET_COMBO (new, level_set));
    gchar *set_name;
    gtk_tree_model_get (model, &it, 1, &set_name, -1);

		GError *err = NULL;
		GSList *names;
		parse_level_set_levels  (set_name, &names, &err);

		GSList *name_i18n;
		gchar *name;
		GSList *note_i18n;
		gchar *note;

		parse_level_set_meta (set_name, &name_i18n, &name, &note_i18n, &note, &err);
		gtk_label_set_text (GET_LABEL (new, note), note);
		free_i18n (name_i18n);
		free_i18n (note_i18n);

    GtkListStore  *store;
    GtkTreeIter    iter;
    store = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_INT);

		GSList *l = names;
    for (int n = 0;l != NULL;n++) {
      gtk_list_store_append (store, &iter);
      gtk_list_store_set (store, &iter,	0, (gchar*) ((gpointer*)l->data)[1],	1, n,	-1);
			free_i18n ((GSList*) ((gpointer*)l->data)[0]);
			l = g_slist_next (l);
    }

    gtk_combo_box_set_model (GET_COMBO (new, level),
														 GTK_TREE_MODEL (store));
    SET_SENSITIVE (new, combo_level, TRUE);
    CONNECT (new, combo_level, "changed", combo_level_change_cb);

		int sel = 0;
		if (game->loaded && strcmp (set_name, game->level_set->file) == 0) {
			sel = game->level;
		}
    gtk_combo_box_set_active (GET_COMBO (new, level), sel);
  } else {
		SET_SENSITIVE (new, combo_level, FALSE);
  }
}

/**
 * new_game_cb:
 * @menuitem: the menu item activated
 * @data: unused
 * 
 * Shows the new game dialog.
 */
static void new_game_cb (GtkWidget *menuitem, gpointer data) {
	game_instance *game = data;
  game->win_new = glade_xml_new (LASER_CONFIG_SHARE "new-game.glade", NULL, NULL);
  GtkListStore  *store = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_STRING);
	int i = 0, sel = 0;
	GSList* level_sets = get_available_level_sets ();
  for (GSList *l = level_sets;l != NULL;l = g_slist_next(l)) {
		GtkTreeIter iter;		
    gtk_list_store_append (store, &iter);
    gtk_list_store_set (store, &iter,0, (gchar*)((gpointer*)l->data)[2], 1, (gchar*)((gpointer*)l->data)[0],	-1);
		if (game->loaded && strcmp ((gchar*)((gpointer*)l->data)[2], game->level_set->file) == 0) {
			sel = i;
		}
		i++;
  }
  gtk_combo_box_set_model (GET_COMBO (new, level_set), GTK_TREE_MODEL (store));
  SET_SENSITIVE (new, combo_level_set, TRUE);
  CONNECT (new, combo_level_set, "changed", combo_level_set_change_cb);
	gtk_combo_box_set_active (GET_COMBO (new, level_set), sel); 
  gtk_dialog_run (GET_DIALOG (new));
  gtk_widget_destroy (GET_WIDGET (new, dialog));
	free_available_level_sets (level_sets);
}

/* preferences ****************************************************************/

/**
 * color_set_cb:
 * @widget: the colour button
 * @data: unused
 *
 * sets the background colour from colour button (with gconf)
 */
static void color_set_cb (GtkColorButton *widget, gpointer data) {
	game_instance *game = data;

	GdkColor gdk_colour;
	colour_t colour;
	gtk_color_button_get_color (widget, &gdk_colour);
	colour_from_gdk (&colour, gdk_colour, gtk_color_button_get_alpha (widget));
	laser_config_set_colour (game->config, &colour);
}

/**
 * combo_theme_change_cb:
 * @combo: the theme combo
 * @data: unused
 *
 * Sets theme from combo (with gconf)
 */
static void combo_theme_change_cb (GtkComboBox *combo,	gpointer data) {
	game_instance *game = data;
  GtkTreeIter it;
  if (gtk_combo_box_get_active_iter (combo, &it)) {
    GtkTreeModel* model = gtk_combo_box_get_model (combo);
    gchar *name;
    gtk_tree_model_get (model, &it, 1, &name, -1);
		laser_config_set_theme (game->config, name);
	}
}

/**
 * stats_toggled_cb:
 * @togglebutton: the stats checkbox
 * @data: unused
 *
 * Set statistics display from checkbox (with gconf)
 */
static void stats_toggled_cb (GtkToggleButton *togglebutton, gpointer data) {
	game_instance *game = data;

	laser_config_set_show_stats (game->config, gtk_toggle_button_get_active (togglebutton));
}

/**
 * restore_toggled_cb:
 * @togglebutton: the restore checkbox
 * @data: unused
 *
 * Set restoration from checkbox (with gconf)
 */
static void restore_toggled_cb (GtkToggleButton *togglebutton, gpointer data) {
	game_instance *game = data;

	laser_config_set_restore (game->config, gtk_toggle_button_get_active (togglebutton));
}

/**
 * arrows_toggled_cb:
 * @togglebutton: the arrows checkbox
 * @data: unused
 *
 * Set restoration from checkbox (with gconf)
 */
static void arrows_toggled_cb (GtkToggleButton *togglebutton, gpointer data) {
	game_instance *game = data;

	laser_config_set_arrows (game->config, gtk_toggle_button_get_active (togglebutton));
}

/**
 * preferences_cb:
 * @menuitem: the menu item activated
 * @data: unused
 * 
 * Shows the preferences dialog.
 */
static void preferences_cb (GtkMenuItem *menuitem, gpointer data) {
	game_instance *game = data;

  game->win_prefs = glade_xml_new (LASER_CONFIG_SHARE "preferences.glade", NULL, NULL);
  
	colour_t colour;
	laser_config_get_colour (game->config, &colour);

  gtk_color_button_set_color (GTK_COLOR_BUTTON (GET_BUTTON (prefs, color)), &colour.gdk_colour);
  gtk_color_button_set_alpha (GTK_COLOR_BUTTON (GET_BUTTON (prefs, color)), colour.gdk_alpha);
  SET_CHECK_ACTIVE (prefs, stats, laser_config_get_show_stats (game->config));
  SET_CHECK_ACTIVE (prefs, restore, laser_config_get_restore (game->config));
  SET_CHECK_ACTIVE (prefs, arrows, laser_config_get_arrows (game->config));

  CONNECT (prefs, button_color, "color-set", color_set_cb);
  CONNECT (prefs, check_stats, "toggled", stats_toggled_cb);
  CONNECT (prefs, check_restore, "toggled",	restore_toggled_cb);
  CONNECT (prefs, check_arrows, "toggled", arrows_toggled_cb);
	
  GtkListStore  *store = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_STRING);
	GSList* themes = get_available_themes ();
  int sel = 0, i = 0;
  for (GSList *l = themes;l != NULL;l = g_slist_next(l)) {
		GtkTreeIter    iter;
    gtk_list_store_append (store, &iter);
    gtk_list_store_set (store, &iter, 0, (gchar*)((gpointer*)l->data)[1], 1, (gchar*)((gpointer*)l->data)[0], -1);
    if (strcmp (((gchar**)l->data)[0], laser_config_get_theme (game->config)) == 0) {
      sel = i;
    }
    i++;
  }
	gtk_combo_box_set_model (GET_COMBO (prefs, theme), GTK_TREE_MODEL (store));
  SET_SENSITIVE (prefs, combo_theme, TRUE);
  CONNECT (prefs, combo_theme, "changed", combo_theme_change_cb);
	gtk_combo_box_set_active (GET_COMBO (prefs, theme), sel); 
  gtk_dialog_run (GET_DIALOG (prefs));
  gtk_widget_destroy (GET_WIDGET (prefs, dialog));

	free_available_themes (themes);
}

/**
 * resize_window_cb:
 * @window: the main window
 * @allocation: the new size
 * @data: unused
 *
 * saves via gconf the new size
 */
static void resize_window_cb (GtkWidget *window, GtkAllocation *allocation, gpointer data) {
	game_instance *game = data;

	laser_config_set_height (game->config, allocation->height);
	laser_config_set_width (game->config, allocation->width);
}

/**
 * clock_display_cb:
 * @data: ignored
 *
 * GTimer callback to update lock diaplay.
 * Never stops because always returns TRUE.
 */
static gboolean clock_display_cb (gpointer data) {
	game_instance *game = data;

  if (game->loaded && laser_config_get_show_stats (game->config)) {
    gint t = g_timer_elapsed (game->timer_clock, NULL) + game->timer_offset;
		gchar *label =
			g_strdup_printf (_("<b>Clock : </b>%d:%02d:%02d"),
											 (t / 3600) % 60,
											 (t / 60) % 60,
											 t % 60);
    gtk_label_set_markup (GET_LABEL (main, clock), label);
		g_free (label);
  }
  return TRUE; /* never stop timer */
}

/**
 * quit_cb:
 *
 * Quits saving history (called from menu)
 */
static void quit_cb (GtkMenuItem *menuitem, gpointer data) {
	game_instance *game = data;
	if (game->loaded) {
		laser_config_set_clock (game->config, g_timer_elapsed (game->timer_clock, NULL) + game->timer_offset);
		laser_config_set_level (game->config, game->level_set->file, game->level);
		laser_config_set_cheat (game->config, laser_board_get_cheat (game->board));
	}
  laser_config_save_history (game->config);
  gtk_main_quit ();
}

/**
 * delete_cb:
 *
 * Quits saving history (called on window close)
 */
static gboolean delete_cb (GtkWidget *widget, GdkEvent *event, gpointer data) {
	quit_cb (NULL, data);
  return TRUE;
}

/**
 * dialog_about_cb:
 *
 * displays about dialog
 **/
static void dialog_about_cb (GtkMenuItem *menuitem, gpointer data) {
	game_instance *game = data;

	GtkAboutDialog *about;

	GdkPixbuf *logo =
		gtk_icon_theme_load_icon (gtk_icon_theme_get_default (),
					  "laser-game",
					  128, 0, NULL);

	const gchar *authors[] = {
		_("Programming"),
		" - Benjamin Canou <bookeldor@gmail.com>",
		"",
		_("Levels"),
		" - Benjamin Canou <bookeldor@gmail.com>",
		" - Michel Quenault <miq75@free.fr>",
		" - Bertrand Perrier <bertrandperrier@laposte.net>",
		"",
		_("Distribution"),
		" - Ubuntu GNU/Linux :",
		"   David Valot <ptitdav69@gmail.com>",
		NULL
	};
	const char *documenters [] = {
		" - Benjamin Canou <bookeldor@gmail.com>",
		NULL
	};
	const char *license[] = {
		 N_("Licensed under the GNU General Public License Version 2"),
		 N_("Laser is free software; you can redistribute it and/or\n"
		   "modify it under the terms of the GNU General Public License\n"
		   "as published by the Free Software Foundation; either version 2\n"
		   "of the License, or (at your option) any later version."),
		 N_("Laser is distributed in the hope that it will be useful,\n"
		   "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
		   "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
		   "GNU General Public License for more details."),
		 N_("You should have received a copy of the GNU General Public License\n"
		   "along with this program; if not, write to the Free Software\n"
		   "Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA\n"
		   "02110-1301, USA.")
	};
	const char *translator_credits = NULL;
	char	   *license_trans;

	license_trans = g_strconcat (_(license[0]), "\n\n", _(license[1]), "\n\n",
				     _(license[2]), "\n\n", _(license[3]), "\n", NULL);

	about = (GtkAboutDialog*) gtk_about_dialog_new ();
	gtk_window_set_icon_name (GTK_WINDOW (about), "laser-game");
	gtk_window_set_transient_for (GTK_WINDOW (about), GET_WINDOW (main));
	gtk_window_set_position (GTK_WINDOW (about), GTK_WIN_POS_CENTER_ON_PARENT);
	gtk_about_dialog_set_name (about, _("Laser"));
	gtk_about_dialog_set_version (about, VERSION);
	gtk_about_dialog_set_copyright (about, _("Copyright \xc2\xa9 2007 Benjamin Canou"));
	gtk_about_dialog_set_comments (about, _("An enlightened game for the gnome desktop"));
	gtk_about_dialog_set_authors (about, authors);
	gtk_about_dialog_set_documenters (about, documenters);
	gtk_about_dialog_set_translator_credits (about, translator_credits);
	gtk_about_dialog_set_logo (about, logo);
	gtk_about_dialog_set_license (about, license_trans);
	gtk_about_dialog_set_website (about, "http://bookeldor-net.info");

	g_signal_connect (G_OBJECT(about), "response",
			  G_CALLBACK(gtk_widget_destroy), NULL);

	gtk_widget_show (GTK_WIDGET(about));

	g_free (license_trans);
	gdk_pixbuf_unref (logo);
}

/**
 * help_cb:
 *
 * open help
 **/
static void help_cb (GtkMenuItem *menuitem, gpointer data) {
	game_instance *game = data;

	GError *error = NULL;

	gnome_help_display_with_doc_id (program, NULL,"laser.xml",NULL, &error);
	if (error != NULL) {
		GtkWidget *dialog =
			gtk_message_dialog_new (GET_WINDOW (main),
															GTK_DIALOG_DESTROY_WITH_PARENT,
															GTK_MESSAGE_ERROR,
															GTK_BUTTONS_CLOSE,
															error->message);
		gtk_dialog_run (GTK_DIALOG (dialog));
		gtk_widget_destroy (dialog);
		g_error_free (error);
	}
}

static void colour_changed_cb (LaserConfig *laser_config, colour_t *colour, gpointer data) {
	game_instance *game = data;

	laser_board_paint (game->board);
}

static void arrows_changed_cb (LaserConfig *laser_config, gboolean arrows, gpointer data) {
	game_instance *game = data;

	laser_board_paint (game->board);
}

static void show_stats_changed_cb (LaserConfig *laser_config, gboolean show_stats, gpointer data) {
	game_instance *game = data;

	if (show_stats) {
		gtk_widget_show (GET_WIDGET (main, stats));
	} else {
		gtk_widget_hide (GET_WIDGET (main, stats));
	}
}

static void theme_changed_cb (LaserConfig *laser_config, gpointer data) {
	game_instance *game = data;

	laser_board_paint (game->board);
}

static void cheat_show_0_cb (GtkToggleButton *togglebutton, gpointer data) {
	game_instance *game = data;

	if (gtk_toggle_button_get_active (togglebutton)) {
		gtk_widget_show (GET_WIDGET (cheat, check_show_1));
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (GET_WIDGET (cheat, check_show_1)), FALSE);
	} else {
		gtk_widget_hide (GET_WIDGET (cheat, check_show_1));
		gtk_widget_hide (GET_WIDGET (cheat, check_show_2));
		gtk_widget_hide (GET_WIDGET (cheat, check_show_3));
	}
}

static void cheat_show_1_cb (GtkToggleButton *togglebutton, gpointer data) {
	game_instance *game = data;

	if (gtk_toggle_button_get_active (togglebutton)) {
		gtk_widget_show (GET_WIDGET (cheat, check_show_2));
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (GET_WIDGET (cheat, check_show_2)), FALSE);
	} else {
		gtk_widget_hide (GET_WIDGET (cheat, check_show_2));
		gtk_widget_hide (GET_WIDGET (cheat, check_show_3));
	}
}

static void cheat_show_2_cb (GtkToggleButton *togglebutton, gpointer data) {
	game_instance *game = data;

	if (gtk_toggle_button_get_active (togglebutton)) {
		gtk_widget_show (GET_WIDGET (cheat, check_show_3));
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (GET_WIDGET (cheat, check_show_3)), FALSE);
	} else {
		gtk_widget_hide (GET_WIDGET (cheat, check_show_3));
	}
}

static void cheat_show_3_cb (GtkToggleButton *togglebutton, gpointer data) {
	game_instance *game = data;

	if (gtk_toggle_button_get_active (togglebutton)) {
		gtk_entry_set_text (GET_ENTRY (cheat, new), game->level_set->levels[game->level].solution);
	} else {
		gtk_entry_set_text (GET_ENTRY (cheat, new), "");
	}
}

static void cheat_copy_cb (GtkWidget *widget, gpointer data) {
	game_instance *game = data;
	gchar *cheat = (gchar*) gtk_entry_get_text (GET_ENTRY (cheat, current));
	gtk_clipboard_set_text(gtk_clipboard_get_for_display (gdk_display_get_default (),GDK_SELECTION_CLIPBOARD),
												 cheat, strlen (cheat));
}

static void cheat_paste_cb (GtkWidget *widget, gpointer data) {
	game_instance *game = data;
	gtk_entry_set_text (GET_ENTRY (cheat, new),
											gtk_clipboard_wait_for_text (gtk_clipboard_get_for_display (gdk_display_get_default (),
																																									GDK_SELECTION_CLIPBOARD)));
}

static void cheat_cb (GtkMenuItem *menuitem, gpointer data) {
	game_instance *game = data;
	
  game->win_cheat = glade_xml_new (LASER_CONFIG_SHARE "cheat.glade", NULL, NULL);	

	if (game->level_set->levels[game->level].solution != NULL) {
		gtk_widget_show (GET_WIDGET (cheat, label3));
		gtk_widget_show (GET_WIDGET (cheat, alignment3));
		CONNECT (cheat, button_copy, "clicked", cheat_copy_cb);
		CONNECT (cheat, button_paste, "clicked", cheat_paste_cb);
		CONNECT (cheat, check_show_0, "toggled", cheat_show_0_cb);
		CONNECT (cheat, check_show_1, "toggled", cheat_show_1_cb);
		CONNECT (cheat, check_show_2, "toggled", cheat_show_2_cb);
		CONNECT (cheat, check_show_3, "toggled", cheat_show_3_cb);
	}

	gchar *cheat = laser_board_get_cheat (game->board);
	gtk_entry_set_text (GET_ENTRY (cheat, current), cheat);
	g_free (cheat);

	gint r = gtk_dialog_run (GET_DIALOG (cheat));
	gchar *code = g_strdup ((gchar*) gtk_entry_get_text (GET_ENTRY (cheat, new)));
	gtk_widget_destroy (GET_WIDGET (cheat, dialog));
	if (r == -2) {
		laser_board_set_cheat (game->board, code);
	}
	g_free (code);
}


/**
 * main:
 *
 * where all the fun starts
 */
int main (int argc, gchar **argv) {
	game_instance *game = g_malloc (sizeof (struct game_instance));
	GOptionContext *context;

  gtk_init (&argc, &argv);
  rsvg_init (); 
  glade_init ();
  gconf_init(argc, argv, NULL);

	context = g_option_context_new (_("Laser"));
	
	bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
	textdomain (GETTEXT_PACKAGE);
	
	program = gnome_program_init (PACKAGE, VERSION, LIBGNOMEUI_MODULE, argc, argv,
																GNOME_PROGRAM_STANDARD_PROPERTIES,
			    									    GNOME_PARAM_GOPTION_CONTEXT, context,
																GNOME_PARAM_NONE);

	game->config = laser_config_new ();
	game->timer_clock = g_timer_new ();
	game->timer_offset = 0;
	game->loaded = FALSE;
 	game->displayed = FALSE;
	game->supports_alpha = FALSE;
	game->level_set = NULL;
	game->level = 0;
	game->won = TRUE;

  game->win_main = glade_xml_new (LASER_CONFIG_SHARE "main.glade", NULL, NULL);

  screen_changed (GET_WIDGET (main, window), NULL, game);

  CONNECT (main, window, "screen-changed", screen_changed);
  CONNECT (main, window, "delete-event", delete_cb);
  CONNECT (main, window, "size-allocate", resize_window_cb);
  CONNECT (main, menu_quit, "activate", quit_cb);
	CONNECT (main, menu_new, "activate", new_game_cb);
	CONNECT (main, menu_prefs, "activate", preferences_cb);
	CONNECT (main, menu_help, "activate", help_cb);
	CONNECT (main, menu_about, "activate", dialog_about_cb);
	CONNECT (main, menu_open, "activate", open_cb);
	CONNECT (main, button_open, "clicked", open_cb);
	CONNECT (main, button_new, "clicked", new_game_cb);
  CONNECT (main, button_reset, "clicked", reset_cb);
  CONNECT (main, menu_reset, "activate", reset_cb);
  CONNECT (main, button_redo, "clicked", redo_cb);
  CONNECT (main, button_undo, "clicked", undo_cb);
  CONNECT (main, button_next, "clicked", next_cb);
  CONNECT (main, menu_redo, "activate", redo_cb);
  CONNECT (main, menu_undo, "activate", undo_cb);
  CONNECT (main, menu_cheat,	"activate", cheat_cb);
  CONNECT (main, button_cheat,	"clicked", cheat_cb);
  CONNECT (main, button_previous, "clicked", previous_cb);

  gtk_window_set_default_size (GET_WINDOW (main),
															 laser_config_get_width (game->config),
															 laser_config_get_height (game->config));

	game->board =
		laser_board_attach (GET_WIDGET (main, board), PLAY_MODE, game->config, update_states, &game->modified, game);

  g_signal_connect (game->config, "arrows_changed", G_CALLBACK (arrows_changed_cb), game);
  g_signal_connect (game->config, "colour_changed", G_CALLBACK (colour_changed_cb), game);
  g_signal_connect (game->config, "theme_changed", G_CALLBACK (theme_changed_cb), game);
  g_signal_connect (game->config, "show_stats_changed", G_CALLBACK (show_stats_changed_cb), game);

  if (laser_config_get_restore (game->config)) {
		gchar *set;
		int level;
		int a = laser_config_get_actions (game->config);
		gchar *cheat = laser_config_get_cheat (game->config);
		laser_config_get_level (game->config, &set, &level);
		load_level (game, set, level);
		laser_board_set_cheat (game->board, cheat);
		g_free (cheat);
		laser_board_set_actions (game->board, a);
		game->timer_offset = laser_config_get_clock (game->config);
		clock_display_cb (game);
  }

  gtk_widget_show_all (GET_WIDGET (main, window));

  if (laser_config_get_show_stats (game->config)) {
    gtk_widget_show (GET_WIDGET (main, stats));
  }
  g_timeout_add (1000, clock_display_cb, game);

	game->displayed = TRUE;

  gtk_main ();

  return 0;
}
