/* -*- Mode: C; tab-width: 2; indent-tabs-mode: t; c-basic-offset: 2 -*-
 *
 * LASER
 * Copyright (C) 2006 Benjamin Canou <bookeldor@gmail.com>
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <libgnomevfs/gnome-vfs.h>
#include "laser-common.h"
#include "laser-parser.h"

/**
 * stream_new:
 *
 */
stream_t* stream_new (gchar *uri) {
	if (!gnome_vfs_initialized () && !gnome_vfs_init ()) {
		return NULL;
	}
	
	stream_t *stream = g_malloc (sizeof (struct stream));
	if (stream == NULL) {
		return NULL;
	}

	stream->col = 0;
	stream->row = 1;
	stream->i = -1;
	stream->col_i = 0;
	stream->eof = FALSE;
	stream->was_eof = FALSE;
	stream->message = NULL;

	if (gnome_vfs_open (&stream->handle, uri, GNOME_VFS_OPEN_READ) != GNOME_VFS_OK) {
		g_free (stream);
		return NULL;
	}

	return stream;
}


/**
 * stream_destroy:
 *
 */
void stream_destroy (stream_t *stream) {
	if (stream != NULL) {
		gnome_vfs_close (stream->handle);
		if (stream->message != NULL) {
			g_free (stream->message);
		}
		g_free (stream);
	}
}

/**
 * stream_get:
 *
 */
gchar stream_get (stream_t *stream) {
  char c;
	if (stream->i >= 0) {
		c = stream->buffer[stream->i--];
		stream->eof = stream->was_eof;
	} else {
		GnomeVFSFileSize nb;
		switch (gnome_vfs_read (stream->handle, &c, (GnomeVFSFileSize) 1, &nb)) {
		case GNOME_VFS_OK :
			break;
		case GNOME_VFS_ERROR_EOF :
			stream->eof = TRUE;
			break;
		default :
			return '\0';
		}
	}

  if (c == '\n') {
    stream->row++;
		stream->col_hist[stream->col_i] = stream->col;
		stream->col_i++;
    stream->col = 0;
  } else {
    stream->col++;
  }
  return c;
}

/**
 * stream_put:
 *
 */
void stream_put (stream_t *stream, gchar c) {
	if (stream->i < 1000) {
		if (stream->i == -1) {
			stream->was_eof = stream->eof;
		}
		stream->i++;
		stream->buffer[stream->i] = c;
		if (c == '\n') {
			stream->row--;
			stream->col_i--;
			stream->col = stream->col_hist[stream->col_i];
		} else {
			stream->col--;
		}
		stream->eof = FALSE;
	}
}

/**
 * stream_skip_blanks:
 *
 */
void stream_skip_blanks (stream_t *stream) {
	gboolean end = FALSE;
	while (!stream->eof && !end) {
		gchar c = stream_get(stream);
		if (!isspace (c)) {
			stream_put (stream, c);
			end = TRUE;
		}
	}
}

/**
 * stream_read_keyword:
 *
 */
gboolean stream_read_keyword (stream_t *stream, gchar **k) {
	stream_skip_blanks (stream);

	guint k_i = 0, k_size = READ_BLOCK;

	if ((*k = g_malloc ((k_size + 1) * sizeof (gchar))) == NULL) {
		return TRUE;
	}

	gboolean end = FALSE;
	while (!stream->eof && !end) {
		gchar c = stream_get(stream);
		if (!isalnum (c)) {
			stream_put (stream, c);
			end = TRUE;
		} else {
			if (k_i >= k_size) {
				k_size += READ_BLOCK;
				gchar *new_k = g_realloc (*k, (k_size + 1) * sizeof (gchar));
				if (new_k == NULL) {
					free (*k);
					*k = NULL;
					return TRUE;
				}
				*k = new_k;
			}
			(*k)[k_i] = c;
			k_i++;
		}
	}
	(*k)[k_i] = '\0';

	if (k_i > 0) {
		return FALSE;
	} else {
		g_free (*k);
		*k = NULL;
		return TRUE;
	}
}

/**
 * stream_read_string:
 *
 */
gboolean stream_read_string (stream_t *stream, gchar **s, gboolean alloc) {
	stream_skip_blanks (stream);

	guint s_i = 0, s_size = READ_BLOCK;
	if (alloc) {
		if ((*s = malloc ((s_size + 1) * sizeof (gchar))) == NULL) {
			return TRUE;
		}
	}
	
	if (!stream->eof) {
		gchar c = stream_get(stream);
		if (c != '"') {
			if (alloc) {
				g_free (*s);
				*s = NULL;
			}
			return TRUE;
		}
	}

	gboolean end = FALSE;
	while (!stream->eof && !end) {
		gchar c = stream_get(stream);
		if (c == '"') {
			end = TRUE;
			if (alloc) {
				(*s)[s_i] = '\0';
			}
			break;
		}
		if (c == '\\') {
			if (stream->eof) {
				if (alloc) {
					g_free (*s);
					*s = NULL;
				}
				return TRUE;
			} else {
				gchar c2 = stream_get(stream);
				switch (c2) {
				case 'n' :
					c = '\n';
					break;
				case 't' :
					c = '\t';
					break;
				case 'r' :
					c = '\r';
					break;
				default :
					c = c2;
				}
			}
		}
		if (alloc) {
			if (s_i >= s_size) {
				s_size += READ_BLOCK;
				gchar *new_s = realloc (*s, (s_size + 1) * sizeof (gchar));
				if (new_s == NULL) {
					g_free (*s);
					*s = NULL;
					return TRUE;
				}
				*s = new_s;
			}
			(*s)[s_i] = c;
		}
		s_i++;
	}
	if (end) {
		return FALSE;
	} else {
		if (alloc) {
			g_free (*s);
		}
		return TRUE;
	}
}

/**
 * stream_read_string_i18n:
 *
 */
gboolean stream_read_string_i18n (stream_t *stream,	GSList **s_i18n, gchar **s, gboolean alloc) {
	if (stream_read_string (stream, s, alloc)) {
		if (alloc) {
			*s = NULL;
			*s_i18n = NULL;
		}
		return TRUE;
	}

	if (alloc) {
		gchar **pair = g_malloc (2 * sizeof (gchar*));
		if (pair == NULL) {
			g_free (*s);
			*s = NULL;
			*s_i18n = NULL;
			return TRUE;
		}
		pair[0] = g_strdup ("C");
		pair[1] = *s;
		*s_i18n = g_slist_prepend (*s_i18n, pair);
	}

	int compatibility_level = 10000;
	const gchar* const * langs = g_get_language_names ();
	gboolean locale_found = FALSE;

	gboolean end = FALSE;
	while (!end) {
		stream_skip_blanks (stream);
		if (!stream->eof) {
			gchar c = stream_get(stream);
			if (c != '[') {
				stream_put (stream, c);
				return FALSE;
			}
		}
		stream_skip_blanks (stream);
		gchar locale[21] = "";
		int i = 0;
		while (!stream->eof && i < 20) {
			gchar c = stream_get(stream);
			if (isspace(c) || c == ']') {
				stream_put (stream, c);
				break;
			} else{
				locale[i] = c;
				i++;
			}
		}
		locale[i] = '\0';

		stream_skip_blanks (stream);
		if (!stream->eof) {
			gchar c = stream_get(stream);
			if (c != ']') {
				if (alloc) {
					for (GSList *l = *s_i18n;l != NULL;l = g_slist_next (l)) {
						g_free (((gchar**)l->data)[0]);
						g_free (((gchar**)l->data)[1]);
						g_free (l->data);
					}
					g_slist_free (*s_i18n);
					*s_i18n = NULL;
					*s = NULL;
				}
				return TRUE;
			}
		}
		gchar *str;
		if (stream_read_string (stream, &str, alloc)) {
			if (alloc) {
				for (GSList *l = *s_i18n;l != NULL;l = g_slist_next (l)) {
					g_free (((gchar**)l->data)[0]);
					g_free (((gchar**)l->data)[1]);
					g_free (l->data);
				}
				g_slist_free (*s_i18n);
				*s_i18n = NULL;
				*s = NULL;
			}
			return TRUE;
		}
		if (alloc) {
			gchar **pair = g_malloc (2 * sizeof (gchar*));
			if (pair == NULL) {
				g_free (str);
				for (GSList *l = *s_i18n;l != NULL;l = g_slist_next (l)) {
					g_free (((gchar**)l->data)[0]);
					g_free (((gchar**)l->data)[1]);
					g_free (l->data);
				}
				g_slist_free (*s_i18n);
				*s_i18n = NULL;
				*s = NULL;
				return TRUE;
			}
			pair[0] = g_strdup (locale);
			pair[1] = str;
			*s_i18n = g_slist_prepend (*s_i18n, pair);
		}
		/* check if the locale is compatible with the lang */
		for (int i = 0; langs[i] != NULL; i++) {
			/* update the name if locale is more compatible */
			if (strcmp (langs[i], locale) == 0 && (!locale_found || i < compatibility_level)) {
				locale_found = TRUE;
				if (alloc) {
					*s = str;
				}
				compatibility_level = i;
				break;
			}
		}
	}

	return FALSE;
}

static int  parse_map                (stream_t *stream, level *level, gboolean alloc);
static int  parse_level              (stream_t *stream, level_set *level_set,
																			gboolean alloc, gboolean meta, GSList **names);
static void parse_level_set_internal (gchar *uri, GError **gerror, guint8 mode, gpointer data[]);

enum {
	MODE_CHECK,
	MODE_META,
	MODE_LEVELS,
	MODE_ALL
};

/**
 * parse_level_set:
 *
 */
void parse_level_set (gchar *uri, level_set *level_set, GError **gerror) {
	gpointer data[1] = { level_set };
	level_set->file = g_strdup (uri);
	parse_level_set_internal (uri, gerror, MODE_ALL, data);
}

/**
 * parse_level_set_check:
 *
 * gchar *uri, GError **error
 */
void parse_level_set_check (gchar *uri, GError **gerror, gint *n_levels) {
	gpointer data[1] = { n_levels };
	parse_level_set_internal (uri, gerror, MODE_CHECK, data);
}

/**
 * parse_level_set_meta:
 *
 */
void parse_level_set_meta (gchar *uri,
													 GSList **name_i18n, gchar **name, GSList **note_i18n, gchar **note,
													 GError **gerror) {
	gpointer data[4] = { name_i18n, name, note_i18n, note };
	parse_level_set_internal (uri, gerror, MODE_META, data);
}

/**
 * parse_level_set_levels:
 *
 * gchar *uri, GSList **names, GError **error
 */
void parse_level_set_levels (gchar *uri, GSList **names, GError **gerror) {
	gpointer data[1] = { names };
	parse_level_set_internal (uri, gerror, MODE_LEVELS, data);
	*names = g_slist_reverse (*names);
}


/**
 * parse_level_set_internal:
 *
 */
static void parse_level_set_internal (gchar *uri, GError **gerror, guint8 mode, gpointer data[]) {
	/* MODE_ALL */
	level_set *level_set = NULL;
	/* MODE_META */
	GSList **name_i18n = NULL;
	gchar **name = NULL;
	GSList **note_i18n = NULL;
	gchar **note = NULL;
	/* MODE_LEVELS */
	GSList **names = NULL;
	/* MODE_CHECK */
	gint *n_levels = 0;

	static GQuark quark = -1;
	if (quark == -1) {
		quark = g_quark_from_string ((const gchar*) "laser-parser");
	}

	stream_t *stream = stream_new (uri);
	if (stream == NULL) {
		*gerror = g_error_new(quark, 1, _("Cannot open file"));
		return;
	}

	switch (mode) {
	case MODE_ALL :
		level_set = data[0];
		break;
	case MODE_META :
		name_i18n = data[0];
		*name_i18n = NULL;
		name = data[1];
		note_i18n = data[2];
		*note_i18n = NULL;
		note = data[3];
		break;
	case MODE_LEVELS :
		names = data[0];
		*names = NULL;
		break;
	case MODE_CHECK :
		n_levels = data[0];
		*n_levels = 0;
		break;
	default :
		*gerror = g_error_new(quark, 1, _("Internal error"));
		return;
	}

  gboolean end = FALSE, error = FALSE;

  while (!stream->eof && !end && !error) {
		gboolean found = FALSE;
		gchar *key;
		
		if (stream_read_keyword (stream, &key)) {
			error = TRUE;
			break;
		}

    if (strcmp(key, "Set") == 0) {
			found = TRUE;
			switch (mode) {
			case MODE_ALL :
				if (stream_read_string_i18n (stream, &level_set->name_i18n, &level_set->name, TRUE)) {
					error = TRUE;
				}
				break;
			case MODE_META :
				if (stream_read_string_i18n (stream, name_i18n, name, TRUE)) {
					error = TRUE;
				}
				break;
			default :
				if (stream_read_string_i18n (stream, NULL, NULL, FALSE)) {
					error = TRUE;
				}
			}
    }

    if (strcmp(key, "Note") == 0) {
			found = TRUE;
			switch (mode) {
			case MODE_ALL :
				if (stream_read_string_i18n (stream, &level_set->note_i18n, &level_set->note, TRUE)) {
					error = TRUE;
				}
				break;
			case MODE_META :
				if (stream_read_string_i18n (stream, note_i18n, note, TRUE)) {
					error = TRUE;
				}
				break;
			default :
				if (stream_read_string_i18n (stream, NULL, NULL, FALSE)) {
					error = TRUE;
				}
			}
    }

    if (strcmp(key, "Level") == 0) {
			found = TRUE;
			switch (mode) {
			case MODE_ALL :
				if (parse_level(stream, level_set, TRUE, FALSE, NULL) == -1) {
					error = TRUE;
				}
				break;
			case MODE_LEVELS :
				if (parse_level(stream, NULL, FALSE, TRUE, names) == -1) {
					error = TRUE;
				}
				break;
			case MODE_CHECK :
				if (parse_level (stream, NULL, FALSE, FALSE, NULL) == -1) {
					error = TRUE;
				}
				(*n_levels)++;
				break;
			default :
				if (parse_level (stream, NULL, FALSE, FALSE, NULL) == -1) {
					error = TRUE;
				}
			}
		}

    if (strcmp(key, "EndSet") == 0) {
			found = TRUE;
      end = TRUE;
    }

		if (!found) {
			stream->message =
				g_strdup_printf (_("Line %d column %d : unknown keyword %s"),
												 stream->row, stream->col, key);
			error = TRUE;
		}

		g_free (key);
  }

	if (!end && stream->eof) {
		*gerror = g_error_new(quark, 1, "%s", _("File not correctly terminated"));
		switch (mode) {
		case MODE_ALL :
			level_set_free (level_set);
			break;
		case MODE_META :
			free_i18n ((GSList*) name_i18n);
			free_i18n ((GSList*) note_i18n);
			break;
		case MODE_LEVELS :
			for (GSList *l = *names;l != NULL;l = g_slist_next (l)) {
				free_i18n ((GSList*) ((gpointer*)l->data)[0]);
				g_free (l->data);
			}
			g_slist_free (*names);
		}
		stream_destroy (stream);
		return;
	}

	if (error) {
		if (stream->message) {
			*gerror = g_error_new(quark, 1, "%s", stream->message);
		} else {
			*gerror = g_error_new(quark, 1,
									g_strdup_printf (_("Line %d column %d : %s"),
																	 stream->row, stream->col,
																	 _("Internal error")));
		}
		switch (mode) {
		case MODE_ALL :
			level_set_free (level_set);
			break;
		case MODE_META :
			free_i18n ((GSList*) name_i18n);
			free_i18n ((GSList*) note_i18n);
			break;
		case MODE_LEVELS :
			for (GSList *l = *names;l != NULL;l = g_slist_next (l)) {
				free_i18n ((GSList*) ((gpointer*)l->data)[0]);
				g_free (l->data);
			}
			g_slist_free (*names);
		}
		stream_destroy (stream);
		return;
	}

	stream_destroy (stream);
}


/**
 * parse_map:
 *
 * @stream:
 * @level:
 *
 */
static int parse_map (stream_t *stream, level *level, gboolean alloc) {
  level->height = 0;
  level->width = 0;

  /* calc size */
	stream_skip_blanks (stream);
	if (stream_get (stream) != '/') {
		return -2;
	}
  while (!stream->eof) {
		gchar s[4] = "   ";
    s[0] = stream_get (stream);
		if (isspace(s[0])) {
			break;
		}
    s[1] = stream_get (stream);
    s[2] = stream_get (stream);
    if (strcmp (s, "--+") != 0) {
			return -2;
    }
    level->width++;
  }

  /* read lines */
	gboolean end = FALSE;
  while (!stream->eof && !end) {
		if (alloc) {
			/* (re) alloc map */
			item **new_map;
			if (level->map == NULL) {
				new_map = malloc (sizeof(struct item*));
			} else {
				new_map = realloc (level->map, (level->height + 1) * sizeof(struct item*));
			}
			if (new_map == NULL) {
				return -1;
			} else {
				level->map = new_map;
				level->map[level->height] = malloc (level->width * sizeof (struct item));
				if (level->map[level->height] == NULL) {
					return -1;
				}
				level->height++;
			}
		}

		stream_skip_blanks (stream);
    if (stream_get (stream) != '|') {
      return -2;
    }
    int w = 0;
    while (!stream->eof && w < level->width) {
      char c_colour = stream_get (stream);
      char c_type = stream_get (stream);

			if (alloc) {
				level->map[level->height - 1][w].colour = NONE;
				level->map[level->height - 1][w].type = NONE;
				level->map[level->height - 1][w].direction = NONE;
			}
			
      if (c_type != ' ' && c_type != '#' && c_type != 'H') {
				switch (c_type) {
				case '@' :
					if (alloc) {
						level->map[level->height - 1][w].type = MAP_TARGET;
					}
					break;
				case '0' :
				case '<' :
					if (alloc) {
						level->map[level->height - 1][w].type = MAP_BEAMER;
						level->map[level->height - 1][w].direction = DIR_LEFT;
					}
					break;
				case '1' :
					if (alloc) {
						level->map[level->height - 1][w].type = MAP_BEAMER;
						level->map[level->height - 1][w].direction = 1;
					}
					break;
				case '2' :
				case '^' :
					if (alloc) {
						level->map[level->height - 1][w].type = MAP_BEAMER;
						level->map[level->height - 1][w].direction = DIR_UP;
					}
					break;
				case '3' :
					if (alloc) {
						level->map[level->height - 1][w].type = MAP_BEAMER;
						level->map[level->height - 1][w].direction = 3;
					}
					break;
				case '4' :
				case '>' :
					if (alloc) {
						level->map[level->height - 1][w].type = MAP_BEAMER;
						level->map[level->height - 1][w].direction = DIR_RIGHT;
					}
					break;
				case '5' :
					if (alloc) {
						level->map[level->height - 1][w].type = MAP_BEAMER;
						level->map[level->height - 1][w].direction = 5;
					}
					break;
				case '6' :
				case 'v' :
					if (alloc) {
						level->map[level->height - 1][w].type = MAP_BEAMER;
						level->map[level->height - 1][w].direction = DIR_DOWN;
					}
					break;
				case '7' :
					if (alloc) {
						level->map[level->height - 1][w].type = MAP_BEAMER;
						level->map[level->height - 1][w].direction = 7;
					}
					break;
				default :
					return -2;
				}
      }

      switch (c_colour) {
			case 'D' :
				if (alloc) {
					level->map[level->height - 1][w].type = MAP_DIODE;
				}
				break;
      case 'r' :
				if (alloc) {
					level->map[level->height - 1][w].colour = COLOUR_RED;
				}
				break;
      case 'g' :
				if (alloc) {
					level->map[level->height - 1][w].colour = COLOUR_GREEN;
				}
				break;
      case 'b' :
				if (alloc) {
					level->map[level->height - 1][w].colour = COLOUR_BLUE;
				}
				break;
      case 'v' :
				if (alloc) {
					level->map[level->height - 1][w].colour = COLOUR_VIOLET;
				}
				break;
      case 'w' :
				if (alloc) {
					level->map[level->height - 1][w].colour = COLOUR_WHITE;
				}
				break;
      case 'y' :
				if (alloc) {
					level->map[level->height - 1][w].colour = COLOUR_YELLOW;
				}
				break;
      case 'c' :
				if (alloc) {
					level->map[level->height - 1][w].colour = COLOUR_CYAN;
				}
				break;
      case ' ' :
				if (c_type != ' ') {
					return -2;
				}
				break;
      case '#' :
				if (c_type != '#') {
					return -2;
				} else {
					if (alloc) {
						level->map[level->height - 1][w].type = MAP_WALL;
					}
				}
				break;
      case 'H' :
				if (c_type != 'H') {
					return -2;
				} else {
					if (alloc) {
						level->map[level->height - 1][w].type = MAP_HIDER;
					}
				}
				break;
      default :
				return -2;
      }

      if (stream_get (stream) != '|') {
				return -2;
      }
      w++;
    }

		stream_skip_blanks (stream);
    if (stream_get (stream) != '+') {
      return -2;
    }

    w = 0;
    while (!stream->eof) {
			gchar s[4] = "   ";
			s[0] = stream_get (stream);
			if (isspace(s[0])) {
				break;
			}
			s[1] = stream_get (stream);
			s[2] = stream_get (stream);
			if (strcmp (s, "--/") == 0) {
				end = TRUE;
			} else {
				if (strcmp (s, "--+") != 0) {
					return -2;
				}
			}
      w++;
    }
    if (w != level->width) {
      return -2;
    }
  }

  return 0;
}

/**
 * parse_level:
 *
 * @stream: @
 * level_set:
 *
 */
static int parse_level (stream_t *stream, level_set *level_set, gboolean alloc, gboolean meta, GSList **names) {
  level level, *new_levels;
  gboolean end = FALSE, error = FALSE;

	for (int j = 0;j < MATERIALS;j++)  {
		level.materials[j] = 0;
	}
	
	level.comment = NULL;
	level.comment_i18n = NULL;
	level.name = NULL;
	level.name_i18n = NULL;
	level.map = NULL;
	level.difficulty = 0;
	level.solution = NULL;

	if (stream_read_string_i18n (stream, &level.name_i18n, &level.name, alloc || meta)) {
		error = TRUE;
	}

  while (!stream->eof && !end && !error) {
		gchar *key;

		if (stream_read_keyword (stream, &key)) {
			error = TRUE;
			break;
		}

    if (strcmp(key, "EndLevel") == 0) {
      end = TRUE;
    } else {
			gboolean found = FALSE;
			if (strcmp(key, "Comment") == 0) {
				found = TRUE;
				if (stream_read_string_i18n (stream, &level.comment_i18n, &level.comment, alloc || meta)) {
					error = TRUE;
				}
			}

      if (strcmp(key, "Map") == 0) {
				found = TRUE;
				switch (parse_map (stream, &level, alloc)) {
				case -1 :
					error = TRUE;
					break;
				case -2 :
					error = TRUE;
					stream->message =
						g_strdup_printf (_("Line %d column %d : malformed map"),
														 stream->row, stream->col);
				}
      }

			if (strcmp(key, "Solution") == 0) {
				found = TRUE;
				if (stream_read_string (stream, &level.solution, alloc)) {
					error = TRUE;
				}
			}

      if (strcmp(key, "Difficulty") == 0) {
				found = TRUE;
				gint n = 0;
				stream_skip_blanks (stream);
				while (!stream->eof) {
					gchar c = stream_get (stream);
					if (!isdigit(c)) {
						stream_put (stream, c);
						break;
					}
					n = n * 10 + (gint) (c - '0');
					if (n <= 0 || n > 10) {
						stream->message =
							g_strdup_printf (_("Line %d column %d : bad difficulty"),
															 stream->row, stream->col);
					}
				}
				level.difficulty = n;
      }
      
      for(int j = 0;j < MATERIALS;j++) {
				if (strcmp(key, materials_assoc[j].name) == 0) {
					found = TRUE;
					gint n = 0;
					stream_skip_blanks (stream);
					while (!stream->eof) {
						gchar c = stream_get (stream);
						if (!isdigit(c)) {
							stream_put (stream, c);
							break;
						}
						n = n * 10 + (gint) (c - '0');
						if (n <= 0) {
							stream->message =
								g_strdup_printf (_("Line %d column %d : bad number of %s"),
																 stream->row, stream->col, key);
						}
						level.materials[materials_assoc[j].index] = n;
					}
				}
      }

			if (!found) {
				stream->message =
					g_strdup_printf (_("Line %d column %d : unknown keyword %s"),
													 stream->row, stream->col, key);
				error = TRUE;
			}

			g_free (key);
    }
  }

	if (error || (!end && stream->eof)) {
		if (alloc) {
			free_i18n (level.name_i18n);
			free_i18n (level.comment_i18n);
			if (level.map != NULL) {
				for (int i = 0;i < level.height;i++) {
					g_free (level.map[i]);
				}
				g_free (level.map);
			}
		}
		if (!end && stream->eof) {
			stream->message =
				g_strdup_printf ("%s", _("File not correctly terminated"));
		}
		return -1;
	}

	if (meta) {
		gpointer *pair;
		if ((pair = malloc (2 * sizeof (gpointer))) == NULL) {
			free_i18n (level.name_i18n);
			free_i18n (level.comment_i18n);
		} else {
			pair[0] = level.name_i18n;
			pair[1] = level.name;
			*names = g_slist_prepend (*names, pair);
		}
	}

	/* expand */
	if (alloc) {
		int nb_materials = 0;
		
		for (int i = 0;i < MATERIALS;i++) {
			nb_materials += level.materials[i];
		}
		
		int nb_extra_rows = nb_materials / level.width + 1;
		
		if (nb_materials % level.width != 0) {
			nb_extra_rows++;
		}
		
		item **new_map =
			realloc (level.map, (level.height + nb_extra_rows) * sizeof (struct item*));
		
		if (new_map == NULL) {
			free_i18n (level.name_i18n);
			free_i18n (level.comment_i18n);
			for (int i = 0;i < level.height;i++) {
				g_free (level.map[i]);
			}
			g_free (level.map);
			level.map = NULL;
			return -1;
		}
		
		level.map = new_map;
		
		for (int i = 0;i < nb_extra_rows;i++) {
			level.map[level.height + i] = malloc (level.width * sizeof (struct item));
			if (level.map[level.height + i] == NULL) {
				free_i18n (level.name_i18n);
				free_i18n (level.comment_i18n);
				for (int j = 0;j < level.height + i;j++) {
					g_free (level.map[j]);
				}
				g_free (level.map);
				level.map = NULL;
				return -1;
			}
		}
		
		for (int i = 1;i < nb_extra_rows;i++) {
			for (int j = 0;j < level.width;j++) {
				level.map[level.height + i][j].type = NONE;
			}
		}
		
		for (int i = 0;i < level.width;i++) {
			level.map[level.height][i].type = MAP_WALL;
		}
		
		int row = level.height + 1;
		int col = 0;
		for (int i = 0;i < MATERIALS;i++) {
			for (int n = 0;n < level.materials[i];n++) {
				level.map[row][col].type = i + 1;
				level.map[row][col].direction = 0;
				level.map[row][col].colour = NONE;
				col++;
				if (col >= level.width) {
					col = 0;
					row++;
				}
			}
		}
		level.real_height = level.height;
		level.height += nb_extra_rows;
		
		if (level_set->levels == NULL) {
			new_levels = malloc (sizeof (struct level));
		} else {
			new_levels = realloc (level_set->levels,
														sizeof (struct level) * (level_set->nb_levels + 1));
		}
		if (new_levels == NULL) {
			free_i18n (level.name_i18n);
			free_i18n (level.comment_i18n);
			for (int j = 0;j < level.height;j++) {
				g_free (level.map[j]);
			}
			g_free (level.map);
			return -1;
		}
		level_set->levels = new_levels;
		memcpy (&level_set->levels[level_set->nb_levels], &level, sizeof (struct level));
		level_set->nb_levels++;
	}

	return 0;
}

GSList* parse_cheat_code(gchar *code, gboolean *err) {
	GSList *result = NULL;
	gchar *s = code;
	gint x, y, sign;
	gboolean error = FALSE;
	while (*s && !error) {
		x = 0;
		sign = 1;
		if (*s == '-') {
			sign = -1;
			s++;
		}
		while (*s && *s != ',') {
			if (*s >= '0' && *s <= '9') {
				x = x * 10 + (*s - '0');
			} else {
				error = TRUE;
			}
			s++;
		}
		x *= sign;
		s++;
		y = 0;
		sign = 1;
		if (*s == '-') {
			sign = TRUE;
			s++;
		}
		while (*s && *s != ';') {
			if (*s >= '0' && *s <= '9') {
				y = y * 10 + (*s - '0');
			} else {
				error = TRUE;
			}
			s++;
		}
		y *= sign;
		s++;
		if (!error) {
			gint *pair = g_malloc (2 * sizeof (gint));
			if (pair == NULL) {
				error = TRUE;
			} else {
				pair[0] = x;
				pair[1] = y;
				result = g_slist_prepend (result, pair);
			}
		}
	}
	*err = error;
	return result;
}
